/* 
 * $RCSfile: gj_agents.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/07/17 18:43:49 $
 * 
 * $Revision: 1.9 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <glib.h>

#include <libxml/tree.h>
#include <libxml/parser.h>

#include "gj_agents.h"
#include "gj_parser.h"
#include "gj_presence.h"
#include "gj_message.h"
#include "gj_iq_requests.h"

/* agent item */
void gj_agent_destroy(agent ag);

/* agent list */
gboolean gj_agent_list_add(agent ag);
gboolean gj_agent_list_del(agent ag);

void gj_agent_list_cb(guint requestID);
GList *gj_agent_list_parse(xmlNodePtr parent);

/* call backs */
gboolean gj_agent_set_callback(agentCallback cb);
agentCallback gj_agent_get_callback();

struct t_agent
{
  gchar *jid;
  gchar *name;
  gchar *description;
 
  gchar *transport;
  gchar *groupchat;
  gchar *service;
  gchar *registers;
  gchar *search;
};

GList *agent_list;
agentCallback agent_cb;

gboolean gj_agent_init()
{
  if(agent_list != NULL)
    gj_agent_term();

  /* free agent response proc */
  agent_cb = NULL;

  /* create mem  - nothing to do, created on g_list_append */
  return TRUE;
}

gboolean gj_agent_term()
{
  gint index = 0;

  /* list */
  if(agent_list == NULL)
    return TRUE;

  /* free all items on the list */
  for(index=0;index<g_list_length(agent_list);index++)
    {
      agent ag = NULL;
      
      if((ag = (agent) g_list_nth_data(agent_list,index)) != NULL)
	gj_agent_free(ag);
    }

  g_list_free(agent_list);
  agent_list = NULL;

  /* free agent response proc */
  agent_cb = NULL;

  return TRUE;
}

gboolean gj_agent_list_add(agent ag)
{
  if(ag == NULL)
    {
      g_warning("%s: agent was NULL", __FUNCTION__);
      return FALSE;
    }

  agent_list = g_list_append(agent_list,ag);

  return TRUE;
}

gboolean gj_agent_list_del(agent ag)
{
  if(ag == NULL)
    {
      g_warning("%s: agent was NULL", __FUNCTION__);
      return FALSE;
    }

  agent_list = g_list_remove(agent_list,ag);
  return TRUE;
}

GList *gj_agent_list_get()
{
  return agent_list;
}

/* 
 * item functions
 */
agent gj_agent_new(gchar *jid)
{
  agent ag;

  if(jid == NULL)
    {
      g_warning("%s: jid was NULL", __FUNCTION__);
      return NULL;
    }
  
  if((ag = g_new(struct t_agent,1)))
    {
      ag->jid = g_strdup(jid);
      ag->name = NULL;
      ag->description = NULL;

      ag->transport = NULL;
      ag->groupchat = NULL;
      ag->service = NULL;
      ag->registers = NULL;
      ag->search = NULL;

      gj_agent_list_add(ag);
    }
  
  return ag;  
}

void gj_agent_free(agent ag)
{
  if(ag == NULL)
    {
      g_warning("%s: jid was NULL", __FUNCTION__);
      return;
    }

  /* free mem */
  gj_agent_list_del(ag);  

  /* destroy members */
  gj_agent_destroy(ag);

  /* free structure */
  g_free(ag);
}

void gj_agent_destroy(agent ag)
{
  if(ag == NULL)
    {
      g_warning("%s: agent was NULL", __FUNCTION__);
      return;
    }
  
  if(ag->jid != NULL)
    {
      g_free(ag->jid);
      ag->jid = NULL;
    }

  if(ag->name != NULL)
    {
      g_free(ag->name);
      ag->name = NULL;
    }

  if(ag->description != NULL)
    {
      g_free(ag->description);
      ag->description = NULL;
    }

  if(ag->transport != NULL)
    {
      g_free(ag->transport);
      ag->transport = NULL;
    }

  if(ag->groupchat != NULL)
    {
      g_free(ag->groupchat);
      ag->groupchat = NULL;
    }

  if(ag->service != NULL)
    {
      g_free(ag->service);
      ag->service = NULL;
    }

  if(ag->registers != NULL)
    {
      g_free(ag->registers);
      ag->registers = NULL;
    }

  if(ag->search != NULL)
    {
      g_free(ag->search);
      ag->search = NULL;
    }
}

/*
 * list functions
 */
gboolean gj_agent_list_request(gchar *server, agentCallback cb)
{
  if(server == NULL)
    {
      g_warning("%s: server was NULL", __FUNCTION__);
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("%s: callback was NULL", __FUNCTION__);
      return FALSE;
    }

  /* set call back proc */
  gj_agent_set_callback(cb);

  /* do request */
  gj_iq_request_agents(server, gj_agent_list_cb);

  return TRUE;
}

void gj_agent_list_cb(guint eventID)
{
  infoquery iq = NULL;
  infoquery iq_related = NULL;

  guint eventIDRelated = 0;
  gint error_code = 0;
  gchar *error_message = NULL;
  agentCallback cb = NULL;

  xmlChar *type = NULL;
  xmlNodePtr node = NULL;

  if((iq = gj_iq_find_by_event_id(eventID)) == NULL)
    {
      g_warning("%s: could not find infoquery from event id:%d", __FUNCTION__, eventID);
      return;
    }

  eventIDRelated = gj_iq_get_related_id(iq);
  if((iq_related = gj_iq_find_by_event_id(eventIDRelated)) == NULL)
    {
      g_warning("%s: could not find infoquery from related id:%d", __FUNCTION__, eventIDRelated);
      return;
    }

  node = gj_iq_get_xml(iq_related);
  type = gj_parser_find_by_attr_and_ret(node, (guchar*)"type");

  if(type != NULL && xmlStrcmp(type, (guchar*)"error") == 0)
    {
      xmlChar *errorCode = NULL;
      xmlChar *errorMessage = NULL;

      if((node = gj_parser_find_by_node(node, (guchar*)"error")) != NULL)
	{
	  if((errorCode = gj_parser_find_by_attr_and_ret(node, (guchar*)"code")) != NULL)
	    error_code = atoi((gchar*)errorCode);
	  
	  if((errorMessage = gj_parser_find_by_node_and_ret(node, (guchar*)"error")) != NULL) 
	    error_message = g_strdup((gchar*)errorMessage);
	}
      else
	g_warning("%s: failed to find node 'error'", __FUNCTION__);
    }

  if((cb = gj_agent_get_callback()) != NULL)
    {
      if(agent_list != NULL)
	{
	  /* free list */
	}
      
      /* parse list */
      agent_list = gj_agent_list_parse(node);
      
      /* return list */
      (cb)(agent_list, error_code, error_message);
    }
  else
    g_warning("%s: callback is NULL", __FUNCTION__);
  
  /* clean up */
  gj_iq_free_all_related_by_event_id(eventID);
  if(error_message != NULL) g_free(error_message);
}

GList *gj_agent_list_parse(xmlNodePtr parent)
{
  gint count = 0;
  xmlNodePtr node = NULL;

  if(parent == NULL)
    {
      g_warning("%s: xml node was NULL", __FUNCTION__);
      return NULL;
    }

  /* create list */
  gj_agent_init();

  /* get item node */
  node = gj_parser_find_by_node(parent, (guchar*)"agent");

  /* get items in agent */
  while(node)
    {
      xmlChar *tmp = NULL;

      count++;
      
      /* get data */
      if((tmp = gj_parser_find_by_attr_and_ret(node, (guchar*)"jid")) != NULL)
	{
	  gchar *jid = NULL;
	  gchar *resource = NULL;
	  gchar **vjid = NULL;

	  agent ag = NULL;

	  vjid = g_strsplit((gchar*)tmp,"/",2);

	  if(vjid != NULL)
	    {
	      jid = g_strdup(vjid[0]);
	      
	      if(vjid[1] != NULL)
		resource = g_strdup(vjid[1]);
	      
	      g_strfreev(vjid);
	    }
	  
	  /* create new agent item */
	  ag = gj_agent_new(jid);
      
	  /* check it... */
	  if(ag == NULL)
	    {
	      g_warning("%s: failed to create new agent, continuing to next item...", __FUNCTION__);
	      continue;
	    }
	  
#if 0
	  /* resource */
	  if(resource != NULL)
	    ag->resource = resource;
#endif

	  /* find other details */
	  if((tmp = gj_parser_find_by_node_and_ret(node, (guchar*)"name")) != NULL)
	    gj_agent_set_name(ag, (gchar*)tmp);
	  if((tmp = gj_parser_find_by_node_and_ret(node, (guchar*)"description")) != NULL)
	    gj_agent_set_description(ag, (gchar*)tmp);
	  if((tmp = gj_parser_find_by_node_and_ret(node, (guchar*)"transport")) != NULL)
	    gj_agent_set_transport(ag, (gchar*)tmp);
	  if((tmp = gj_parser_find_by_node_and_ret(node, (guchar*)"groupchat")) != NULL)
	    gj_agent_set_groupchat(ag, (gchar*)tmp);
	  if((tmp = gj_parser_find_by_node_and_ret(node, (guchar*)"service")) != NULL)
	    gj_agent_set_service(ag, (gchar*)tmp);
	  if((tmp = gj_parser_find_by_node_and_ret(node, (guchar*)"register")) != NULL)
	    gj_agent_set_registers(ag, (gchar*)tmp);
	  if((tmp = gj_parser_find_by_node_and_ret(node, (guchar*)"search")) != NULL)
	    gj_agent_set_search(ag, (gchar*)tmp);
	}

      node = node->next;
    }

  g_message("%s: items:%d", __FUNCTION__, count);

  return agent_list;
}

gboolean gj_agent_set_callback(agentCallback cb)
{
  if(cb == NULL)
    {
      g_warning("%s: callback was NULL", __FUNCTION__);
      return FALSE;
    }

  agent_cb = cb;
  
  return TRUE;
}

agentCallback gj_agent_get_callback()
{
  return agent_cb;
}

/* find functions */
agent gj_agent_find_by_name(gchar *name)
{
  gint index = 0;

  if(name == NULL)
    {
      g_warning("%s: name was NULL", __FUNCTION__);
      return NULL;
    }

  for(index=0;index<g_list_length(agent_list);index++)
    {
      agent ag  = g_list_nth_data(agent_list, index);

      if(ag == NULL) continue;
      if(g_strcasecmp(ag->name,name) == 0) return ag;
    }

  return NULL;
}

agent gj_agent_find_by_jid(gchar *jid)
{
  gint index = 0;

  if(jid == NULL)
    {
      g_warning("%s: jid was NULL", __FUNCTION__);
      return NULL;
    }

  for(index=0;index<g_list_length(agent_list);index++)
    {
      agent ag  = g_list_nth_data(agent_list, index);

      if(ag == NULL) continue;
      if(g_strcasecmp(ag->jid,jid) == 0) return ag;
    }

  return NULL;
}

/*
 * member functions
 */

/* sets */
gboolean gj_agent_set_jid(agent ag, gchar *jid)
{
  if(ag == NULL)
    {
      g_warning("%s: agent was NULL", __FUNCTION__);
      return FALSE;
    }

  if(jid == NULL)
    {
      g_warning("%s: jid was NULL", __FUNCTION__);
      return FALSE;
    }

  if(ag->jid != NULL)
    {
      g_free(ag->jid);
      ag->jid = NULL;
    }
  
  ag->jid = g_strdup(jid);
  return TRUE;
}

gboolean gj_agent_set_name(agent ag, gchar *name)
{
  if(ag == NULL)
    {
      g_warning("%s: agent was NULL", __FUNCTION__);
      return FALSE;
    }

  if(name == NULL)
    {
      g_warning("%s: name was NULL", __FUNCTION__);
      return FALSE;
    }

  if(ag->name != NULL)
    {
      g_free(ag->name);
      ag->name = NULL;
    }
  
  ag->name = g_strdup(name);
  return TRUE;
}

gboolean gj_agent_set_description(agent ag, gchar *description)
{
  if(ag == NULL)
    {
      g_warning("%s: agent was NULL", __FUNCTION__);
      return FALSE;
    }

  if(description == NULL)
    {
      g_warning("%s: description was NULL", __FUNCTION__);
      return FALSE;
    }

  if(ag->description != NULL)
    {
      g_free(ag->description);
      ag->description = NULL;
    }
  
  ag->description = g_strdup(description);
  return TRUE;
}

gboolean gj_agent_set_transport(agent ag, gchar *transport)
{
  if(ag == NULL)
    {
      g_warning("%s: agent was NULL", __FUNCTION__);
      return FALSE;
    }

  if(transport == NULL)
    {
      g_warning("%s: transport was NULL", __FUNCTION__);
      return FALSE;
    }

  if(ag->transport != NULL)
    {
      g_free(ag->transport);
      ag->transport = NULL;
    }
  
  ag->transport = g_strdup(transport);
  return TRUE;
}

gboolean gj_agent_set_groupchat(agent ag, gchar *groupchat)
{
  if(ag == NULL)
    {
      g_warning("%s: agent was NULL", __FUNCTION__);
      return FALSE;
    }

  if(groupchat == NULL)
    {
      g_warning("%s: groupchat was NULL", __FUNCTION__);
      return FALSE;
    }

  if(ag->groupchat != NULL)
    {
      g_free(ag->groupchat);
      ag->groupchat = NULL;
    }
  
  ag->groupchat = g_strdup(groupchat);
  return TRUE;
}

gboolean gj_agent_set_service(agent ag, gchar *service)
{
  if(ag == NULL)
    {
      g_warning("%s: agent was NULL", __FUNCTION__);
      return FALSE;
    }

  if(service == NULL)
    {
      g_warning("%s: service was NULL", __FUNCTION__);
      return FALSE;
    }

  if(ag->service != NULL)
    {
      g_free(ag->service);
      ag->service = NULL;
    }
  
  ag->service = g_strdup(service);
  return TRUE;
}

gboolean gj_agent_set_registers(agent ag, gchar *registers)
{
  if(ag == NULL)
    {
      g_warning("%s: agent was NULL", __FUNCTION__);
      return FALSE;
    }

  if(registers == NULL)
    {
      g_warning("%s: registers was NULL", __FUNCTION__);
      return FALSE;
    }

  if(ag->registers != NULL)
    {
      g_free(ag->registers);
      ag->registers = NULL;
    }
  
  ag->registers = g_strdup(registers);
  return TRUE;
}

gboolean gj_agent_set_search(agent ag, gchar *search)
{
  if(ag == NULL)
    {
      g_warning("%s: agent was NULL", __FUNCTION__);
      return FALSE;
    }

  if(search == NULL)
    {
      g_warning("%s: search was NULL", __FUNCTION__);
      return FALSE;
    }

  if(ag->search != NULL)
    {
      g_free(ag->search);
      ag->search = NULL;
    }
  
  ag->search = g_strdup(search);
  return TRUE;
}

/* gets */
gchar *gj_agent_get_jid(agent ag)
{
  if(ag == NULL)
    {
      g_warning("%s: agent was NULL", __FUNCTION__);
      return NULL;
    }

  return ag->jid;
}

gchar *gj_agent_get_name(agent ag)
{
  if(ag == NULL)
    {
      g_warning("%s: agent was NULL", __FUNCTION__);
      return NULL;
    }

  return ag->name;
}

gchar *gj_agent_get_description(agent ag)
{
  if(ag == NULL)
    {
      g_warning("%s: agent was NULL", __FUNCTION__);
      return NULL;
    }

  return ag->description;
}

gchar *gj_agent_get_transport(agent ag)
{
  if(ag == NULL)
    {
      g_warning("%s: agent was NULL", __FUNCTION__);
      return NULL;
    }

  return ag->transport;
}

gchar *gj_agent_get_groupchat(agent ag)
{
  if(ag == NULL)
    {
      g_warning("%s: agent was NULL", __FUNCTION__);
      return NULL;
    }

  return ag->groupchat;
}

gchar *gj_agent_get_service(agent ag)
{
  if(ag == NULL)
    {
      g_warning("%s: agent was NULL", __FUNCTION__);
      return NULL;
    }

  return ag->service;
}

gchar *gj_agent_get_registers(agent ag)
{
  if(ag == NULL)
    {
      g_warning("%s: agent was NULL", __FUNCTION__);
      return NULL;
    }

  return ag->registers;
}

gchar *gj_agent_get_search(agent ag)
{
  if(ag == NULL)
    {
      g_warning("%s: agent was NULL", __FUNCTION__);
      return NULL;
    }

  return ag->search;
}

/* debugging */
void gj_agent_list_print()
{
  agent ag = NULL;
  gint index = 0;

  g_message("++++++++++++++++++++++ Agent - Print All, Begin. +++++++++++++++++++++");

  for(index=0;index<g_list_length(agent_list);index++)
    {
      ag = (agent) g_list_nth_data(agent_list,index);
      gj_agent_print(ag);
    }

  g_message("++++++++++++++++++++++ Agent - Print All, End.   +++++++++++++++++++++");
  
  return;  
}

void gj_agent_print(agent ag)
{
  if(ag == NULL)
    return;

  g_message("+++++++++++++++++++++++++++ Agent +++++++++++++++++++++++++");
  g_message("++++");
  g_message("++++ jid:'%s'",ag->jid);
  g_message("++++ name:'%s'",ag->name);
  g_message("++++ description:'%s'",ag->description);
  g_message("++++");
  g_message("++++ transport:'%s'",ag->transport);
  g_message("++++ groupchat:'%s'",ag->groupchat);
  g_message("++++ service:'%s'",ag->service);
  g_message("++++ registers:'%s'",ag->registers);
  g_message("++++ search:'%s'",ag->search);
  g_message("++++");
  g_message("+++++++++++++++++++++++++++ Agent +++++++++++++++++++++++++");
}

#ifdef __cplusplus
}
#endif
