/* 
 * $RCSfile: gj_gatherdata.h,v $
 *
 * $Author: ginxd $
 * $Date: 2003/07/13 17:50:15 $
 * 
 * $Revision: 1.4 $ 
 *
 */

#ifndef __gj_gatherdata_h
#define __gj_gatherdata_h

#include <glib.h>

typedef struct t_gatherdata *gatherdata;
typedef struct t_gatherdata_field *gatherdata_field;
typedef struct t_gatherdata_field_option *gatherdata_field_option;

typedef void (*gatherdataCallback)(gatherdata gd, gint error_code, gchar *error_reason);
typedef void (*gatherdataRequestCallback)(GList *list, gint error_code, gchar *error_reason);

typedef enum gatherdataFormType {
  gatherdataFormTypeForm = 1,        /* This packet contains a form to fill out. Display it to the user */    
  gatherdataFormTypeSubmit,          /* The form is filled out, and this is the data that is being returned from the form. */
  gatherdataFormTypeCancel,          /* The form was cancelled. Tell the asker that piece of information. */
  gatherdataFormTypeResult,          /* Data results being returned from a search, or some other query. */

  gatherdataFormTypeEnd

} gatherdataFormType;

typedef enum gatherdataFieldType {
  
  gatherdataFieldTypeTextSingle = 1, /* single line or word of text */
  gatherdataFieldTypeTextPrivate,    /* instead of showing the user what they typed, you show ***** to protect it */
  gatherdataFieldTypeTextMulti,      /* multiple lines of text entry */
  gatherdataFieldTypeListSingle,     /* given a list of choices, pick one */
  gatherdataFieldTypeListMulti,      /* given a list of choices, pick one or more */
  gatherdataFieldTypeBoolean,        /* 0 or 1, true or false, yes or no. Default value is 0 */
  gatherdataFieldTypeFixed,          /* fixed for putting in text to show sections, 
					or just advertise your web site in the middle of the form */
  gatherdataFieldTypeHidden,         /* is not given to the user at all, but returned with the questionnaire
					(e.g., a unique number at the top of a sheet of paper to identify the form) */
  gatherdataFieldTypeJIDSingle,      /* Jabber ID - choosing a JID from your roster, and entering one based on the rules for a JID. */
  gatherdataFieldTypeJIDMulti,       /* multiple entries for JIDs */

  gatherdataFieldTypeEnd
  
} gatherdataFieldType;

/* list functions */
gboolean gj_gatherdata_init();
gboolean gj_gatherdata_term();

/* list functions */
GList *gj_gatherdata_list_get();

/* item functions */
gatherdata gj_gatherdata_new();
void gj_gatherdata_free(gatherdata gd);

/* find */
gatherdata gj_gatherdata_find_by_id(gchar *id);
gatherdata gj_gatherdata_find_by_to(gchar *to);
gatherdata gj_gatherdata_find_by_from(gchar *from);
gatherdata gj_gatherdata_find_by_related_id(gchar *id);

/* sets */
gboolean gj_gatherdata_set_id(gatherdata gd, gchar *id);
gboolean gj_gatherdata_set_to(gatherdata gd, gchar *to);
gboolean gj_gatherdata_set_from(gatherdata gd, gchar *to);
gboolean gj_gatherdata_set_type(gatherdata gd, gatherdataFormType type);
gboolean gj_gatherdata_set_instructions(gatherdata gd, gchar *instructions);
gboolean gj_gatherdata_set_fields(gatherdata gd, GList *fields);
gboolean gj_gatherdata_set_callback(gatherdata gd, gatherdataCallback cb);
gboolean gj_gatherdata_set_related_id(gatherdata gd, gchar *id);

/* gets */
gchar *gj_gatherdata_get_id(gatherdata gd);
gchar *gj_gatherdata_get_to(gatherdata gd);
gchar *gj_gatherdata_get_from(gatherdata gd);
gatherdataFormType gj_gatherdata_get_type(gatherdata gd);
gchar *gj_gatherdata_get_instructions(gatherdata gd);
GList *gj_gatherdata_get_fields(gatherdata gd);
gatherdataCallback gj_gatherdata_get_callback(gatherdata gd);
gchar *gj_gatherdata_get_related_id(gatherdata gd);

/* callbacks */
/* gboolean gatherdataListRequest(gchar *server, gatherdataResponseProc proc); */

/* debugging */
void gj_gatherdata_print(gatherdata gd);
void gj_gatherdata_print_all();

/* 
 * gatherdata field
 */

/* list functions */
GList *gj_gatherdata_field_list_get(gatherdata gd);

/* item functions */
gatherdata_field gj_gatherdata_field_new(gatherdata gd, gatherdataFieldType type);
void gj_gatherdata_field_free(gatherdata gd, gatherdata_field gdf);

/* sets */
gboolean gj_gatherdata_field_set_type(gatherdata_field gdf, gatherdataFieldType type);
gboolean gj_gatherdata_field_set_label(gatherdata_field gdf, gchar *label);
gboolean gj_gatherdata_field_set_var(gatherdata_field gdf, gchar *var);
gboolean gj_gatherdata_field_set_value(gatherdata_field gdf, gchar *value);
gboolean gj_gatherdata_field_set_required(gatherdata_field gdf, gboolean required);
gboolean gj_gatherdata_field_set_options(gatherdata_field gdf, GList *options);

/* gets */
gatherdataFieldType gj_gatherdata_field_get_type(gatherdata_field gdf);
gchar *gj_gatherdata_field_get_label(gatherdata_field gdf);
gchar *gj_gatherdata_field_get_var(gatherdata_field gdf);
gchar *gj_gatherdata_field_get_value(gatherdata_field gdf);
gboolean gj_gatherdata_field_get_required(gatherdata_field gdf);
GList *gj_gatherdata_field_get_options(gatherdata_field gdf);

/* 
 * gatherdata field option
 */

/* list functions */
GList *gj_gatherdata_field_option_list_get(gatherdata_field gdf);

/* item functions */
gatherdata_field_option gj_gatherdata_field_option_new(gatherdata_field gdf, gchar *label);
void gj_gatherdata_field_option_free(gatherdata_field gdf, gatherdata_field_option gdfo);

/* sets */
gboolean gj_gatherdata_field_option_set_label(gatherdata_field_option gdfo, gchar *label);
gboolean gj_gatherdata_field_option_set_value(gatherdata_field_option gdfo, gchar *value);

/* gets */
gchar *gj_gatherdata_field_option_get_label(gatherdata_field_option gdfo);
gchar *gj_gatherdata_field_option_get_value(gatherdata_field_option gdfo);

/*
 * parsing functions  
 */
gboolean gj_gatherdata_request(gchar *server, gatherdataRequestCallback cb);

#endif

