/* 
 * $RCSfile: gj_gtk_chat.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/08/23 11:17:26 $
 * 
 * $Revision: 1.33 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <glade/glade.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_GNOME
#include <gnome.h>
#endif

#include "common.h"

#include "gj_main.h"
#include "gj_roster.h"
#include "gj_presence.h"
#include "gj_message.h"
#include "gj_translate.h"
#include "gj_history.h"
#include "gj_support.h"
#include "gj_config.h"

#include "gj_gtk_roster.h"
#include "gj_gtk_chat.h"
#include "gj_gtk_connection_settings.h"
#include "gj_gtk_my_preferences.h"
#include "gj_gtk_user_information.h"
#include "gj_gtk_support.h"
#include "gj_gtk_add_contact.h"

struct t_controls
{
  GtkWidget *window;

  GtkWidget *textview_send;
  GtkWidget *textview_receive;

  GtkWidget *image_composing;
  GtkWidget *image_status;
  GtkWidget *label_status;

  GtkWidget *button_send;
  GtkWidget *button_information;
  GtkWidget *button_add;
  GtkWidget *button_close;
  GtkWidget *button_emote;
  GtkWidget *button_history;

};

struct t_ChatWindow
{
  struct t_controls controls;

  gint id;

  gint ri_id;

  gchar *nickname;
  gchar *groups;

  gchar *thread;
 
  gboolean reverse_text_direction;

  gint last_recv_x_event_type;
  gchar *last_recv_x_event_id;
  
  gint last_sent_x_event_type;
  gchar *last_sent_x_event_id;

  gint composing_timeout_id;
  gboolean composing_timeout_stop;
  gboolean composing_timeout_active;
    
  gboolean composing_typing;
  guint composing_duration;
  guint composing_stopped;
    
};


GList *chat_window_list = NULL;

gboolean gj_gtk_cw_load(ChatWindow cw);
gboolean gj_gtk_cw_unload_by_widget(GtkWidget *widget);

void gj_gtk_cw_present(ChatWindow cw);
void gj_gtk_cw_raise(ChatWindow cw);

/* list functions */
gboolean gj_gtk_cw_list_add(ChatWindow cw);
gboolean gj_gtk_cw_list_del(ChatWindow cw);
GList *gj_gtk_cw_list_get();

/* window functions */
ChatWindow gj_gtk_cw_new(rosterItem ri);
gboolean gj_gtk_cw_free(ChatWindow cw);
void gj_gtk_cw_destroy(ChatWindow cw);

/* utilities */
gboolean gj_gtk_cw_setup(ChatWindow cw);
gboolean gj_gtk_cw_send_set(ChatWindow cw, gchar *text);
gboolean gj_gtk_cw_receive_set(ChatWindow cw, message msg);

gboolean gj_gtk_cw_print_offline(ChatWindow cw, message msg);

void gj_gtk_cw_emote_icon_create_popup_menu(GtkWidget *widget);
void gj_gtk_cw_emote_icon_get_as_text(gchar *emote_icon, gpointer user_data);
GdkPixbuf *gj_gtk_cw_emote_icon_get_as_pixbuf(msgEmote icon);

/* utils */
gboolean gj_gtk_cw_check_for_url(GtkTextBuffer *buffer, gchar *text, guint from_pos);

/* composing */
gboolean gj_gtk_cw_composing(ChatWindow cw);
gboolean gj_gtk_cw_cancelled(ChatWindow cw);

gboolean gj_gtk_cw_composing_start(ChatWindow cw);
gboolean gj_gtk_cw_composing_stop(ChatWindow cw);
gboolean gj_gtk_cw_composing_timeout_cb(gpointer data);

/* gui callbacks */
void on_textbuffer_cw_receive_changed(GtkTextBuffer *textbuffer, gpointer user_data);
gboolean on_textview_cw_receive_motion_notify_event(GtkWidget *widget, GdkEventMotion *event, gpointer user_data);
gboolean on_textview_cw_receive_find_space(gunichar ch, gpointer user_data);
gboolean on_textview_cw_receive_find_alpha_numeric(gunichar ch, gpointer user_data);
gboolean on_textview_cw_receive_button_press_event(GtkWidget *widget, GdkEventButton *event, gpointer user_data);
gboolean on_textview_cw_send_key_release_event(GtkWidget *widget, GdkEventKey *event, gpointer user_data);
gboolean on_textview_cw_send_key_press_event(GtkWidget *widget, GdkEventKey *event, gpointer user_data);
gboolean on_textview_cw_send_focus_in_event(GtkWidget *widget, GdkEventFocus *event, gpointer user_data);
gboolean on_textview_cw_send_focus_out_event(GtkWidget *widget, GdkEventFocus *event, gpointer user_data);

void on_button_cw_close_clicked(GtkButton *button, gpointer user_data);
void on_button_cw_send_clicked(GtkButton *button, gpointer user_data);
void on_button_cw_information_clicked(GtkButton *button, gpointer user_data);
void on_button_cw_history_clicked(GtkButton *button, gpointer user_data);
void on_button_cw_add_clicked(GtkButton *button, gpointer user_data);
void on_button_cw_emote_clicked(GtkButton *button, gpointer user_data);

void on_mnu_emoteicon_big_grin_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_mnu_emoteicon_confused_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_mnu_emoteicon_cool_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_mnu_emoteicon_eek_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_mnu_emoteicon_frown_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_mnu_emoteicon_mad_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_mnu_emoteicon_red_face_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_mnu_emoteicon_roll_eyes_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_mnu_emoteicon_smile_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_mnu_emoteicon_tongue_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_mnu_emoteicon_wink_activate(GtkMenuItem *menuitem, gpointer user_data);

gboolean on_chat_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data);

/*
 * code
 */

/* list handling */
gboolean gj_gtk_cw_list_add(ChatWindow cw)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  chat_window_list = g_list_append(chat_window_list,cw);
  return TRUE;
}

gboolean gj_gtk_cw_list_del(ChatWindow cw)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  chat_window_list = g_list_remove(chat_window_list,cw);
  return TRUE;
}

GList *gj_gtk_cw_list_get()
{
  return chat_window_list;
}

ChatWindow gj_gtk_cw_new(rosterItem ri)
{
  ChatWindow cw = NULL;

  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return NULL;
    }

  if((cw = (ChatWindow) g_new(struct t_ChatWindow,1)) != NULL)
    {
      cw->id = gj_support_get_unique_id();

      cw->ri_id = gj_roster_item_get_id(ri);

      cw->nickname = g_strdup(gj_roster_item_get_name(ri));
      cw->groups = gj_roster_item_get_groups_as_string(ri);

      cw->thread = NULL;

      cw->last_recv_x_event_type = 0;
      cw->last_recv_x_event_id = NULL;
      
      cw->last_sent_x_event_type = 0;
      cw->last_sent_x_event_id = NULL;

      cw->composing_timeout_id = -1;
      cw->composing_timeout_stop = FALSE;
      cw->composing_typing = FALSE;
      cw->composing_duration = 0;
      cw->composing_stopped = 0;

      gj_gtk_cw_list_add(cw);
    }

  return cw;
}

gboolean gj_gtk_cw_free(ChatWindow cw)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  /* free from list */
  gj_gtk_cw_list_del(cw);
  
  /* destroy members */
  gj_gtk_cw_destroy(cw);

  /* free struct */
  g_free(cw);
  
  return TRUE;
}

void gj_gtk_cw_destroy(ChatWindow cw)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return;
    }

  /* stops the composing timeout */
  gj_gtk_cw_composing_stop(cw);
	
  cw->id = -1;

  cw->ri_id = -1;

  if(cw->nickname != NULL)
    {
      g_free(cw->nickname);
      cw->nickname = NULL;
    }

  if(cw->groups != NULL)
    {
      g_free(cw->groups);
      cw->groups = NULL;
    }

  if(cw->thread != NULL)
    {
      g_free(cw->thread);
      cw->thread = NULL;
    }

  cw->last_recv_x_event_type = 0;

  if(cw->last_recv_x_event_id != NULL)
    {
      g_free(cw->last_recv_x_event_id);
      cw->last_recv_x_event_id = NULL;
    }

  cw->last_sent_x_event_type = 0;

  if(cw->last_sent_x_event_id != NULL)
    {
      g_free(cw->last_sent_x_event_id);
      cw->last_sent_x_event_id = NULL;
    }

  cw->composing_timeout_id = -1;
  cw->composing_timeout_stop = FALSE;
  cw->composing_typing = FALSE;
  cw->composing_duration = 0;
  cw->composing_stopped = 0;
}

ChatWindow gj_gtk_cw_find_by_id(gint id)
{
  gint index = 0;

  if(id < 0)
    {
      g_warning("%s: roster item id:%d was < 0", __FUNCTION__, id);
      return NULL;
    }

  for(index=0;index<g_list_length(chat_window_list);index++)
    {
      ChatWindow cw = g_list_nth_data(chat_window_list, index);

      if(cw == NULL) continue;
      if(gj_gtk_cw_get_id(cw) == id) return cw;
    }

  return NULL;  
}

ChatWindow gj_gtk_cw_find_by_ri_id(gint id)
{
  gint index = 0;

  if(id < 0)
    {
      g_warning("%s: roster item id:%d was < 0", __FUNCTION__, id);
      return NULL;
    }

  for(index=0;index<g_list_length(chat_window_list);index++)
    {
      ChatWindow cw = g_list_nth_data(chat_window_list, index);

      if(cw == NULL) continue;
      if(gj_gtk_cw_get_ri_id(cw) == id) return cw;
    }

  return NULL;  
}

ChatWindow gj_gtk_cw_find_by_widget(GtkWidget *widget)
{
  gint index = 0;

  GtkWidget *toplevel = NULL;

  if(widget == NULL)
    {
      g_warning("%s: widget was NULL", __FUNCTION__);
      return NULL;
    }

  /* get top level widget */
  toplevel = gtk_widget_get_toplevel(widget);

  if(toplevel == NULL)
    {
      g_warning("%s: top level widget was NULL", __FUNCTION__);
      return NULL;
    }

  for(index=0;index<g_list_length(chat_window_list);index++)
    {
      ChatWindow cw = g_list_nth_data(chat_window_list, index);

      if(cw == NULL) continue;
      if(cw->controls.window == toplevel) return cw;
    }

  return NULL;  
}

ChatWindow gj_gtk_cw_find_by_last_recv_x_event_id(gchar *message_id)
{
  gint index = 0;

  if(message_id == NULL)
    {
      g_warning("%s: message id was NULL", __FUNCTION__);
      return NULL;
    }

  for(index=0;index<g_list_length(chat_window_list);index++)
    {
      ChatWindow cw = g_list_nth_data(chat_window_list, index);

      if(cw == NULL) continue;
      if(gj_gtk_cw_get_last_recv_x_event_id(cw) != NULL && 
	 strcmp(gj_gtk_cw_get_last_recv_x_event_id(cw), message_id) == 0)
	return cw;
    }

  return NULL;  
}

ChatWindow gj_gtk_cw_find_by_last_sent_x_event_id(gchar *message_id)
{
  gint index = 0;

  if(message_id == NULL)
    {
      g_warning("%s: message id was NULL", __FUNCTION__);
      return NULL;
    }

  for(index=0;index<g_list_length(chat_window_list);index++)
    {
      ChatWindow cw = g_list_nth_data(chat_window_list, index);

      if(cw == NULL) continue;
      if(gj_gtk_cw_get_last_sent_x_event_id(cw) != NULL && 
	 strcmp(gj_gtk_cw_get_last_sent_x_event_id(cw), message_id) == 0)
	return cw;
    }

  return NULL;  
}

ChatWindow gj_gtk_cw_find_by_thread(gchar *thread)
{
  gint index = 0;

  if(thread == NULL)
    {
      g_warning("%s: thread was NULL", __FUNCTION__);
      return NULL;
    }

  for(index=0;index<g_list_length(chat_window_list);index++)
    {
      ChatWindow cw = g_list_nth_data(chat_window_list, index);

      if(cw == NULL) continue;
      if(strcmp(gj_gtk_cw_get_thread(cw), thread) == 0) return cw;
    }

  return NULL;  
}

/* 
 * member functions 
 */

/* gets */ 
gint gj_gtk_cw_get_id(ChatWindow cw)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return -1;
    }

  return cw->id;
}

gint gj_gtk_cw_get_ri_id(ChatWindow cw)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return -1;
    }

  return cw->ri_id;
}

gchar *gj_gtk_cw_get_nickname(ChatWindow cw)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return NULL;
    }

  return cw->nickname;
}

gchar *gj_gtk_cw_get_groups(ChatWindow cw)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return NULL;
    }

  return cw->groups;
}

gchar *gj_gtk_cw_get_thread(ChatWindow cw)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return NULL;
    }

  return cw->thread;
}

gboolean gj_gtk_cw_get_reverse_text_direction(ChatWindow cw)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  return cw->reverse_text_direction;
}

gint gj_gtk_cw_get_last_recv_x_event_type(ChatWindow cw)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return 0;
    }

  return cw->last_recv_x_event_type;
}

gchar *gj_gtk_cw_get_last_recv_x_event_id(ChatWindow cw)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return NULL;
    }

  return cw->last_recv_x_event_id;
}

gint gj_gtk_cw_get_last_sent_x_event_type(ChatWindow cw)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return 0;
    }

  return cw->last_sent_x_event_type;
}

gchar *gj_gtk_cw_get_last_sent_x_event_id(ChatWindow cw)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return NULL;
    }

  return cw->last_sent_x_event_id;
}


guint gj_gtk_cw_get_composing_timeout_id(ChatWindow cw)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  return cw->composing_timeout_id;
}

gboolean gj_gtk_cw_get_composing_timeout_stop(ChatWindow cw)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  return cw->composing_timeout_stop;
}

gboolean gj_gtk_cw_get_composing_timeout_active(ChatWindow cw)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  return cw->composing_timeout_active;
}

gboolean gj_gtk_cw_get_composing_typing(ChatWindow cw)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  return cw->composing_typing;
}

gint gj_gtk_cw_get_composing_duration(ChatWindow cw)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  return cw->composing_duration;

}

gint gj_gtk_cw_get_composing_stopped(ChatWindow cw)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  return cw->composing_stopped;

}


/* sets */
gboolean gj_gtk_cw_set_ri_id(ChatWindow cw, gint id)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }  
  
  if(id < 0)
    {
      g_warning("%s: roster item id:%d was < 0", __FUNCTION__, id);
      return FALSE;
    }    

  cw->ri_id = id;
  return TRUE;
}

gboolean gj_gtk_cw_set_nickname(ChatWindow cw, gchar *nickname)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }  
  
  if(nickname == NULL)
    {
      g_warning("%s: nickname was NULL", __FUNCTION__);
      return FALSE;
    }    

  if(cw->nickname != NULL)
    {
      g_free(cw->nickname);
      cw->nickname = NULL;
    }

  cw->nickname = g_strdup(nickname);
  return TRUE;
}

gboolean gj_gtk_cw_set_groups(ChatWindow cw, gchar *groups)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }  
  
  if(groups == NULL)
    {
      g_warning("%s: groups was NULL", __FUNCTION__);
      return FALSE;
    }    

  if(cw->groups != NULL)
    {
      g_free(cw->groups);
      cw->groups = NULL;
    }

  cw->groups = g_strdup(groups);
  return TRUE;
}

gboolean gj_gtk_cw_set_thread(ChatWindow cw, gchar *thread)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }  
  
  if(thread == NULL)
    {
      g_warning("%s: thread was NULL", __FUNCTION__);
      return FALSE;
    }    

  if(cw->thread != NULL)
    {
      g_free(cw->thread);
      cw->thread = NULL;
    }

  cw->thread = g_strdup(thread);
  return TRUE;
}

gboolean gj_gtk_cw_set_reverse_text_direction(ChatWindow cw, gboolean reverse_text_direction)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }  
  
  cw->reverse_text_direction = reverse_text_direction;
  return TRUE;
}

gboolean gj_gtk_cw_set_last_recv_x_event_type(ChatWindow cw, gint type)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }  
  
  cw->last_recv_x_event_type = type;
  return TRUE;
}

gboolean gj_gtk_cw_set_last_recv_x_event_id(ChatWindow cw, gchar *id)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }  
  
  if(id == NULL)
    {
      g_warning("%s: id was NULL", __FUNCTION__);
      return FALSE;
    }    

  if(cw->last_recv_x_event_id != NULL)
    {
      g_free(cw->last_recv_x_event_id);
      cw->last_recv_x_event_id = NULL;
    }

  g_message("%s: ChatWindow:0x%.8x, id:'%s'", __FUNCTION__, (gint)cw, id);
  cw->last_recv_x_event_id = g_strdup(id);
  return TRUE;
}

gboolean gj_gtk_cw_set_last_sent_x_event_type(ChatWindow cw, gint type)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }  
  
  cw->last_sent_x_event_type = type;
  return TRUE;
}

gboolean gj_gtk_cw_set_last_sent_x_event_id(ChatWindow cw, gchar *id)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }  
  
  if(id == NULL)
    {
      g_warning("%s: id was NULL", __FUNCTION__);
      return FALSE;
    }    

  if(cw->last_sent_x_event_id != NULL)
    {
      g_free(cw->last_sent_x_event_id);
      cw->last_sent_x_event_id = NULL;
    }

  g_message("%s: ChatWindow:0x%.8x id:'%s'", __FUNCTION__, (gint)cw, id);
  cw->last_sent_x_event_id = g_strdup(id);
  return TRUE;
}

gboolean gj_gtk_cw_set_composing_timeout_id(ChatWindow cw, guint id)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  cw->composing_timeout_id = id;
  return TRUE;
}

gboolean gj_gtk_cw_set_composing_timeout_stop(ChatWindow cw, gboolean stop)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  cw->composing_timeout_stop = stop;
  return TRUE;
}

gboolean gj_gtk_cw_set_composing_timeout_active(ChatWindow cw, gboolean active)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  cw->composing_timeout_active = active;
  return TRUE;
}

gboolean gj_gtk_cw_set_composing_typing(ChatWindow cw, gboolean composing)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  cw->composing_typing = composing;
  return TRUE;
}

gboolean gj_gtk_cw_set_composing_duration(ChatWindow cw, gint duration)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }
    
  if(duration < 0)
    {
      g_warning("%s: duration value:%d is < 0", __FUNCTION__, duration);
      return FALSE;
    }
    
  cw->composing_duration = duration;
  return TRUE;
}

gboolean gj_gtk_cw_set_composing_stopped(ChatWindow cw, gint stopped)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }
    
  if(stopped < 0)
    {
      g_warning("%s: stopped value:%d is < 0", __FUNCTION__, stopped);
      return FALSE;
    }
    
  cw->composing_stopped = stopped;
  return TRUE;
}



void gj_gtk_cw_print_all()
{
  gint index = 0;

  /* loop through all contacts on roster */
  for(index=0;index<g_list_length(chat_window_list);index++)
    {
      ChatWindow cw = g_list_nth_data(chat_window_list, index);
      if(cw == NULL) continue;

      gj_gtk_cw_print(cw);
    }  
}

void gj_gtk_cw_print(ChatWindow cw)
{
  if(cw == NULL)
    return;

  g_message("++++++++++++++++++++++++++++++++++++++++++++++++");
  g_message("++++");
  g_message("++++ id:%d", cw->id);
  g_message("++++");
  g_message("++++ roster item id:%d", cw->ri_id);
  g_message("++++");
  g_message("++++ nickname:'%s'", cw->nickname);
  g_message("++++ groups:'%s'", cw->groups);
  g_message("++++ thread:'%s'", cw->thread);
  g_message("++++");
  g_message("++++ reverse_text_direction:%s", cw->reverse_text_direction?"true":"false");
  g_message("++++ last_recv_x_event_type:%d", cw->last_recv_x_event_type);
  g_message("++++ last_recv_x_event_id:%s", cw->last_recv_x_event_id);
  g_message("++++ last_sent_x_event_type:%d", cw->last_sent_x_event_type);
  g_message("++++ last_sent_x_event_id:%s", cw->last_sent_x_event_id);
  g_message("++++");
  g_message("++++++++++++++++++++++++++++++++++++++++++++++++");
}




/*
 * window functions 
 */
gboolean gj_gtk_cw_show(ChatWindow cw, gboolean show)
{
  if(cw == NULL)
    {
      g_warning("gj_gtk_cw_show: cw was NULL");
      return FALSE;
    }
  
  if(show == TRUE)
    gtk_widget_show(cw->controls.window);
  else
    gtk_widget_hide(cw->controls.window);

  return TRUE;
}

void gj_gtk_cw_present(ChatWindow cw)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return;
    }

  if(cw->controls.window == NULL || cw->controls.window->window == NULL)
    {
      g_warning("%s: window was NULL or GDK window was NULL", __FUNCTION__);
      return;
    }

  gtk_window_present(GTK_WINDOW(cw->controls.window));
}

void gj_gtk_cw_raise(ChatWindow cw)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return;
    }

  if(cw->controls.window == NULL || cw->controls.window->window == NULL)
    {
      g_warning("%s: window was NULL or GDK window was NULL", __FUNCTION__);
      return;
    }

  gdk_window_raise(GDK_WINDOW(cw->controls.window->window)); 
}

gboolean gj_gtk_cw_load(ChatWindow cw)
{
  GladeXML *xml = NULL;
  GdkPixbuf *pb = NULL;

  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  if((xml = glade_xml_new(gnome_jabber_get_glade_file(),
			  "chat",
			  NULL)) == NULL)
    {
      g_warning("%s: couldn't open glade file at root 'chat'", __FUNCTION__);
      return FALSE;
    }
  
  cw->controls.window = glade_xml_get_widget(xml,"chat");

  cw->controls.textview_send = glade_xml_get_widget(xml, "textview_cw_send");
  cw->controls.textview_receive = glade_xml_get_widget(xml, "textview_cw_receive");

  cw->controls.image_composing = glade_xml_get_widget(xml, "image_cw_x_event_composing");
  cw->controls.image_status = glade_xml_get_widget(xml, "image_cw_status");
  cw->controls.label_status = glade_xml_get_widget(xml, "label_cw_status");

  cw->controls.button_send = glade_xml_get_widget(xml, "button_cw_send");
  cw->controls.button_information = glade_xml_get_widget(xml, "button_cw_information");
  cw->controls.button_history = glade_xml_get_widget(xml, "button_cw_history");
  cw->controls.button_close = glade_xml_get_widget(xml, "button_cw_close");
  cw->controls.button_add = glade_xml_get_widget(xml, "button_cw_add");
  cw->controls.button_emote = glade_xml_get_widget(xml, "button_cw_emote");

  gdk_window_set_decorations (GDK_WINDOW(cw->controls.window->window), 
			      GDK_DECOR_ALL);

#ifndef WIN32
  glade_xml_signal_autoconnect(xml);
#else
  gtk_widget_set_events(GTK_WIDGET(cw->controls.textview_receive),
			GDK_EXPOSURE_MASK | 
			GDK_POINTER_MOTION_MASK | 
			GDK_POINTER_MOTION_HINT_MASK | 
			GDK_BUTTON_RELEASE_MASK | 
			GDK_STRUCTURE_MASK | 
			GDK_PROPERTY_CHANGE_MASK);

  g_signal_connect ((gpointer) cw->controls.window, "delete_event",
		    G_CALLBACK (on_chat_delete_event),
		    NULL);
  g_signal_connect ((gpointer) cw->controls.button_close, "clicked",
		    G_CALLBACK (on_button_cw_close_clicked),
		    NULL);
  g_signal_connect ((gpointer) cw->controls.button_send, "clicked",
		    G_CALLBACK (on_button_cw_send_clicked),
		    NULL);
  g_signal_connect ((gpointer) cw->controls.button_information, "clicked",
		    G_CALLBACK (on_button_cw_information_clicked),
		    NULL);
  g_signal_connect ((gpointer) cw->controls.button_history, "clicked",
		    G_CALLBACK (on_button_cw_history_clicked),
		    NULL);
  g_signal_connect ((gpointer) cw->controls.button_add, "clicked",
		    G_CALLBACK (on_button_cw_add_clicked),
		    NULL);
  g_signal_connect ((gpointer) cw->controls.button_emote, "clicked",
		    G_CALLBACK (on_button_cw_emote_clicked),
		    NULL);
  g_signal_connect ((gpointer) cw->controls.textview_receive, "button_press_event",
		    G_CALLBACK (on_textview_cw_receive_button_press_event),
		    NULL);
  g_signal_connect ((gpointer) cw->controls.textview_receive, "motion_notify_event",
		    G_CALLBACK (on_textview_cw_receive_motion_notify_event),
		    NULL);
  g_signal_connect ((gpointer) cw->controls.textview_send, "key_press_event",
		    G_CALLBACK (on_textview_cw_send_key_press_event),
		    NULL);
  g_signal_connect ((gpointer) cw->controls.textview_send, "key_release_event",
		    G_CALLBACK (on_textview_cw_send_key_release_event),
		    NULL);
#endif

  /* set icon */
  pb = gj_gtk_support_pixbuf_new_from_file(FILE_MESSAGE_CHAT);
  gtk_window_set_icon(GTK_WINDOW(cw->controls.window), pb);
  g_object_unref(G_OBJECT(pb));

  /* clean up */
  if(xml != NULL)
    {
      g_object_unref(G_OBJECT(xml));
      xml = NULL;
    }

  return TRUE;
}

gboolean gj_gtk_cw_unload_by_widget(GtkWidget *widget)
{
  ChatWindow cw = NULL;

  if(widget == NULL)
    {
      g_warning("%s: widget was NULL", __FUNCTION__);
      return FALSE;
    }

  if((cw = gj_gtk_cw_find_by_widget(widget)) != NULL)
    {
      /* if no more msgs reset icon */
      gtk_widget_destroy(cw->controls.window);
      gj_gtk_cw_free(cw);
      
      g_message("%s: unloaded...", __FUNCTION__);
    }
  else
    g_warning("%s: could not find ChatWindow from widget", __FUNCTION__);
  
  return TRUE;
}

gboolean gj_gtk_cw_setup(ChatWindow cw)
{
  gchar *title = NULL;
  
  GtkTextBuffer *buffer;

  rosterItem ri = NULL;
  presence pres = NULL;

  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  /* set title */
  title = g_strdup_printf("%s %s %s%s%s",
			  _("Chatting with"),
			  gj_gtk_cw_get_nickname(cw),
			  gj_gtk_cw_get_groups(cw)?"[":"",
			  gj_gtk_cw_get_groups(cw)?gj_gtk_cw_get_groups(cw):"",
			  gj_gtk_cw_get_groups(cw)?"]":"");
  gtk_window_set_title(GTK_WINDOW(cw->controls.window), title);
  if(title != NULL) g_free(title);

  /* set buffer iterator at position 0 */
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(cw->controls.textview_receive));
  
  /* set function callbacks */
  g_signal_connect((gpointer) buffer, "changed", 
		   G_CALLBACK(on_textbuffer_cw_receive_changed), 
		   NULL);
 
  /* set styles */
  gtk_text_buffer_create_tag(buffer, "bold", 
			     "weight", PANGO_WEIGHT_BOLD, 
			     NULL);
  gtk_text_buffer_create_tag(buffer, "italic", 
			     "style", PANGO_STYLE_ITALIC, 
			     NULL);  
  gtk_text_buffer_create_tag(buffer, "blue_foreground", 
			     "foreground", "darkblue", 
			     NULL);  
  gtk_text_buffer_create_tag(buffer, "red_foreground", 
			     "foreground", "darkred", 
			     NULL);  
  gtk_text_buffer_create_tag(buffer, "green_foreground", 
			     "foreground", "darkgreen", 
			     NULL);  
  gtk_text_buffer_create_tag(buffer, "black_foreground", 
			     "foreground", "black", 
			     NULL); 
  gtk_text_buffer_create_tag(buffer, "url", 
			     "foreground", "blue", 
			     NULL); 

  /* find roster item by JID */
  if((ri = gj_roster_item_find_by_id(gj_gtk_cw_get_ri_id(cw))) == NULL)
    {
      g_message("%s: could not find roster item from id:%d", __FUNCTION__, gj_gtk_cw_get_ri_id(cw)); 
      return FALSE;
    }

  if((pres = gj_roster_item_get_presence(ri)) == NULL)
    {
      g_warning("%s: could not get presence from roster item with jid:'%s'", __FUNCTION__, gj_roster_item_get_jid(ri)); 
      return FALSE;
    }  

  /* set toolbar */
  if(gj_roster_item_get_is_permanent(ri) == FALSE)
    gtk_widget_set_sensitive(GTK_WIDGET(cw->controls.button_add), TRUE);
  else
    gtk_widget_set_sensitive(GTK_WIDGET(cw->controls.button_add), FALSE);

  /* set text direction */
  gj_gtk_cw_set_reverse_text_direction(cw, gj_config_get_mpf_reverse_text_direction());

  /* set presence */
  gj_gtk_cw_presence_update(ri, pres);

  return TRUE;
}

ChatWindow gj_gtk_cw_receive_new(rosterItem ri)
{
  ChatWindow cw = NULL;

  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return NULL;
    }

  /* check for current chat window for this JID first */
  if((cw = gj_gtk_cw_find_by_ri_id(gj_roster_item_get_id(ri))) != NULL)
    {
      g_message("%s: not creating new window, re-using window already open for jid:'%s'", 
		__FUNCTION__, gj_roster_item_get_jid(ri));
      gtk_window_present(GTK_WINDOW(cw->controls.window));
      return cw;
    }

  cw = gj_gtk_cw_new(ri);
  gj_gtk_cw_load(cw);

  /* setup ... */
  gj_gtk_cw_setup(cw);

  return cw;
}

ChatWindow gj_gtk_cw_send_new(rosterItem ri)
{
  ChatWindow cw = NULL;

  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return NULL;
    }

  /* check for current chat window for this JID first */
  if((cw = gj_gtk_cw_find_by_ri_id(gj_roster_item_get_id(ri))) != NULL)
    {
      g_message("%s: not creating new window, re-using window already open for jid:'%s'", 
		__FUNCTION__, gj_roster_item_get_jid(ri));
      gtk_window_present(GTK_WINDOW(cw->controls.window));
      return cw;
    }

  cw = gj_gtk_cw_new(ri);
  gj_gtk_cw_load(cw);

  /* setup ... */
  gj_gtk_cw_setup(cw);

  return cw;
}

gboolean gj_gtk_cw_receive_set(ChatWindow cw, message msg)
{
  GtkTextView *textview = NULL;
  GtkTextBuffer *buffer = NULL;
  GtkTextMark *mark = NULL;
  GtkTextIter iter;

  GdkPixbuf *pb = NULL;

  gchar *message = NULL;
  gchar *time = NULL;
  glong i = 0;

  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  if(msg == NULL)
    {
      g_warning("%s: msg was NULL", __FUNCTION__);
      return FALSE;
    }

  if((textview = GTK_TEXT_VIEW(cw->controls.textview_receive)) == NULL)
    {
      g_warning("%s: textview was NULL", __FUNCTION__);
      return FALSE;
    }

  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(cw->controls.textview_receive));
  time = g_strdup_printf("[%s]", gj_support_get_timestamp_from_timet(gj_message_get_time(msg)));

  /* set details */
  message = g_strdup_printf(": %s\n", gj_message_get_body(msg));

  /* get text direction */
  if(gj_gtk_cw_get_reverse_text_direction(cw) == TRUE)
    gtk_text_buffer_get_start_iter(buffer, &iter);
  else
    gtk_text_buffer_get_end_iter(buffer, &iter);

  if(gj_config_get_mpf_hide_time() == FALSE)
    {
      /* insert at iter */
      gtk_text_buffer_insert_with_tags_by_name(buffer, &iter,
					       time, -1,
					       "blue_foreground",
					       NULL);
      gtk_text_buffer_insert_with_tags_by_name(buffer, &iter,
					       " ", -1,
					       "blue_foreground",
					       NULL);
    }

  gtk_text_buffer_insert_with_tags_by_name(buffer, &iter,
					   gj_gtk_cw_get_nickname(cw), -1,
					   "bold", 
					   "blue_foreground",
					   NULL);
  
  /* search for occurences of emote icons */
  for(i=0;i<g_utf8_strlen(message, -1);i++)
    {
      gchar *p = NULL;
      p = g_utf8_offset_to_pointer(message, i);

      if(strncmp(p, EMOTE_BIGGRIN, g_utf8_strlen(EMOTE_BIGGRIN,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_biggrin)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_BIGGRIN,-1)-1);
	}
      else if(strncmp(p, EMOTE_CONFUSED, g_utf8_strlen(EMOTE_CONFUSED,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_confused)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_CONFUSED,-1)-1);
	}
      else if(strncmp(p, EMOTE_COOL, g_utf8_strlen(EMOTE_COOL,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_cool)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_COOL,-1)-1);
	}
      else if(strncmp(p, EMOTE_EEK, g_utf8_strlen(EMOTE_EEK,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_eek)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_EEK,-1)-1);
	}
      else if(strncmp(p, EMOTE_FROWN, g_utf8_strlen(EMOTE_FROWN,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_frown)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_FROWN,-1)-1);
	}
      else if(strncmp(p, EMOTE_MAD, g_utf8_strlen(EMOTE_MAD,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_mad)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_MAD,-1)-1);
	}
      else if(strncmp(p, EMOTE_REDFACE, g_utf8_strlen(EMOTE_REDFACE,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_redface)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_REDFACE,-1)-1);
	}
      else if(strncmp(p, EMOTE_ROLLEYES, g_utf8_strlen(EMOTE_ROLLEYES,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_rolleyes)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_ROLLEYES,-1)-1);
	}
      else if(strncmp(p, EMOTE_SMILE, g_utf8_strlen(EMOTE_SMILE,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_smile)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_SMILE,-1)-1);
	}
      else if(strncmp(p, EMOTE_TONGUE, g_utf8_strlen(EMOTE_TONGUE,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_tongue)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_TONGUE,-1)-1);
	}
      else if(strncmp(p, EMOTE_WINK, g_utf8_strlen(EMOTE_WINK,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_wink)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_WINK,-1)-1);
	}
      else
	{
	  gchar *str = NULL;

	  /* create mem - utf should be no longer than 6 bytes */
	  str = g_new0(char, 10);
	  
	  /* copy char */
	  str = g_utf8_strncpy(str, p, 1);
	  
	  /* insert text */
	  gtk_text_buffer_insert_with_tags_by_name(buffer, &iter, 
						   str, -1,
						   "black_foreground",
						   NULL);

	  if(str != NULL) g_free(str);
	}
    }

  /* log to history */
  gj_history_write_received_by_ri_id(gj_gtk_cw_get_ri_id(cw), 
				     (gchar*)gj_support_get_timestamp_nice_from_timet(gj_message_get_time(msg)), 
				     NULL, 
				     gj_message_get_body(msg));

  if(gj_gtk_cw_get_reverse_text_direction(cw) == FALSE)
    {
      mark = gtk_text_buffer_create_mark(buffer, NULL, &iter, FALSE);
      gtk_text_view_scroll_mark_onscreen(textview, mark);
      gtk_text_buffer_delete_mark(buffer, mark);
    }

  /* save details */
  gj_gtk_cw_set_last_recv_x_event_type(cw, gj_message_get_x_event_type(msg));
  gj_gtk_cw_set_last_recv_x_event_id(cw, gj_message_get_id(msg)?gj_message_get_id(msg):gj_message_get_thread(msg));
  gj_gtk_cw_set_thread(cw, gj_message_get_thread(msg));

  /* clean up */
  if(time != NULL) g_free(time);
  if(message != NULL) g_free(message);

  return TRUE;
}

gboolean gj_gtk_cw_send_set(ChatWindow cw, gchar *text)
{
  GtkTextView *textview = NULL;
  GtkTextBuffer *buffer = NULL;
  GtkTextMark *mark = NULL;
  GtkTextIter iter;

  GdkPixbuf *pb = NULL;

  gchar *message = NULL;
  gchar *message_id = NULL;
  gint message_event = 0;
  gchar *time = NULL;
  guint i = 0;

  rosterItem ri = NULL;

  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  if(text == NULL)
    {
      g_warning("%s: text was NULL, can not send NULL", __FUNCTION__);
      return FALSE;
    }

  if((textview = GTK_TEXT_VIEW(cw->controls.textview_receive)) == NULL)
    {
      g_warning("%s: textview was NULL", __FUNCTION__);
      return FALSE;
    }

  if((ri = gj_roster_item_find_by_id(gj_gtk_cw_get_ri_id(cw))) == NULL)
    {
      g_warning("%s: could not find roster item by id:%d", __FUNCTION__, gj_gtk_cw_get_ri_id(cw));
      return FALSE;
    }

  buffer = gtk_text_view_get_buffer(textview);
  time = g_strdup_printf("[%s]", gj_support_get_timestamp());
  
  /* set details */
  message = g_strdup_printf(": %s\n", text);

  /* get text direction */
  if(gj_gtk_cw_get_reverse_text_direction(cw) == TRUE)
    gtk_text_buffer_get_start_iter(buffer, &iter);
  else
    gtk_text_buffer_get_end_iter(buffer, &iter);

  /* insert at iter */
  if(gj_config_get_mpf_hide_time() == FALSE)
    {
      gtk_text_buffer_insert_with_tags_by_name(buffer, &iter,
					       time, -1,
					       "green_foreground",
					       NULL);
      gtk_text_buffer_insert_with_tags_by_name(buffer, &iter,
					       " ", -1,
					       "green_foreground",
					       NULL);
    }

  gtk_text_buffer_insert_with_tags_by_name(buffer, &iter,
					   gj_gtk_cs_get_nickname(), -1,
					   "bold", 
					   "green_foreground",
					   NULL);

  /* search for occurences of emote icons */
  for(i=0;i<g_utf8_strlen(message,-1);i++)
    {
      gchar *p = NULL;
      p = g_utf8_offset_to_pointer(message, i);

      if(strncmp(p, EMOTE_BIGGRIN, g_utf8_strlen(EMOTE_BIGGRIN,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_biggrin)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_BIGGRIN,-1)-1);
	}
      else if(strncmp(p, EMOTE_CONFUSED, g_utf8_strlen(EMOTE_CONFUSED,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_confused)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_CONFUSED,-1)-1);
	}
      else if(strncmp(p, EMOTE_COOL, g_utf8_strlen(EMOTE_COOL,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_cool)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_COOL,-1)-1);
	}
      else if(strncmp(p, EMOTE_EEK, g_utf8_strlen(EMOTE_EEK,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_eek)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_EEK,-1)-1);
	}
      else if(strncmp(p, EMOTE_FROWN, g_utf8_strlen(EMOTE_FROWN,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_frown)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_FROWN,-1)-1);
	}
      else if(strncmp(p, EMOTE_MAD, g_utf8_strlen(EMOTE_MAD,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_mad)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_MAD,-1)-1);
	}
      else if(strncmp(p, EMOTE_REDFACE, g_utf8_strlen(EMOTE_REDFACE,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_redface)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_REDFACE,-1)-1);
	}
      else if(strncmp(p, EMOTE_ROLLEYES, g_utf8_strlen(EMOTE_ROLLEYES,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_rolleyes)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_ROLLEYES,-1)-1);
	}
      else if(strncmp(p, EMOTE_SMILE, g_utf8_strlen(EMOTE_SMILE,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_smile)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_SMILE,-1)-1);
	}
      else if(strncmp(p, EMOTE_TONGUE, g_utf8_strlen(EMOTE_TONGUE,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_tongue)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_TONGUE,-1)-1);
	}
      else if(strncmp(p, EMOTE_WINK, g_utf8_strlen(EMOTE_WINK,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_wink)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_WINK,-1)-1);
	}
      else
	{
	  gchar *str = NULL;

	  /* create mem - utf should be no longer than 6 bytes */
	  str = g_new0(char, 10);
	  
	  /* copy char */
	  str = g_utf8_strncpy(str, p, 1);
	  
	  /* insert text */
	  gtk_text_buffer_insert_with_tags_by_name(buffer, &iter, 
						   str, -1,
						   "black_foreground",
						   NULL);

	  if(str != NULL) g_free(str);
	}
    }

  /* scroll to end if not reversing text direction */
  if(gj_gtk_cw_get_reverse_text_direction(cw) == FALSE)
    {
      mark = gtk_text_buffer_create_mark(buffer, NULL, &iter, FALSE);
      gtk_text_view_scroll_mark_onscreen(textview, mark);
      gtk_text_buffer_delete_mark(buffer, mark);
    }

  /* set message */
  message_event = messageXEventTypeOffline | 
    messageXEventTypeDelivered | 
    messageXEventTypeDisplayed | 
    messageXEventTypeComposing;

  message_id = gj_message_send_with_x_event_request(messageTypeChat, 
						   gj_roster_item_get_jid(ri), 
						   NULL, 
						   text, 
						   gj_gtk_cw_get_thread(cw), 
						   message_event);

  /* send cancelled event */
  gj_gtk_cw_cancelled(cw);

  /* log to history */
  gj_history_write_sent_by_ri_id(gj_gtk_cw_get_ri_id(cw), 
				 (gchar*)gj_support_get_timestamp_nice(), 
				 NULL, 
				 text);

  /* save details */
  gj_gtk_cw_set_last_sent_x_event_id(cw, message_id);
  gj_gtk_cw_set_last_sent_x_event_type(cw, message_event);
  /* should we use a thread? */

  /* clean up */
  if(time != NULL) g_free(time);
  if(message != NULL) g_free(message);
  if(message_id != NULL) g_free(message_id);

  return TRUE;
}

ChatWindow gj_gtk_cw_notification_new(gint ri_id, message msg)
{
  ChatWindow cw = NULL;
  rosterItem ri = NULL;

  if(ri_id < 0)
    {
      g_warning("%s: roster item id:%d was < 0", __FUNCTION__, ri_id);
      return NULL;
    }

  if(msg == NULL)
    {
      g_warning("%s: msg was NULL", __FUNCTION__);
      return NULL;
    }

  if((ri = gj_roster_item_find_by_id(ri_id)) == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return NULL;
    }

  /* look for chat window now... */
  cw = gj_gtk_cw_find_by_ri_id(ri_id);

  if(cw == NULL)
    {
      g_message("%s: no previous window, creating new for jid:'%s'", __FUNCTION__, gj_message_get_from_jid(msg)); 

      /* window doesnt exist - create it */
      cw = gj_gtk_cw_receive_new(ri);

      /* NOTE !!!!! this should perhaps be sent using map/unmap events 
	 so we know WHEN and WHEN NOT the window is displayed!!!! */

      /* if XEvents ask to be notified when the message has 
	 been displayed, this is the time to send that notification */
      if(gj_message_get_x_event_type(msg) & messageXEventTypeDisplayed)
	gj_message_send_x_event(messageXEventTypeDisplayed, gj_message_get_from_jid(msg), gj_message_get_id(msg));
    }
  else
    {
      g_message("%s: found window for jid:'%s'", __FUNCTION__, gj_message_get_from_jid(msg)); 

      /* should we show it? */
      if(gj_config_get_mpf_raise_window_on_new_message() == TRUE)
	{
	  gj_gtk_cw_raise(cw);

	  /* if XEvents ask to be notified when the message has 
	     been displayed, this is the time to send that notification */
	  if(gj_message_get_x_event_type(msg) & messageXEventTypeDisplayed)
	    gj_message_send_x_event(messageXEventTypeDisplayed, gj_message_get_from_jid(msg), gj_message_get_id(msg));
	}

#if 0
      /* if XEvents ask to be notified when the message has 
	 been displayed, this is the time to send that notification */
      if(gj_message_get_x_event_type(msg) & messageXEventTypeOffline)
	gj_gtk_cw_print_offline(cw, msg);
#endif
    }

  gj_gtk_cw_receive_set(cw,msg);

  return cw;
}

void gj_gtk_cw_notification_of_x_event_new(ChatWindow cw, message msg)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return;
    }

  if(msg == NULL)
    {
      g_warning("%s: msg was NULL", __FUNCTION__);
      return;
    }
   
  switch(gj_message_get_x_event_type(msg))
    {
    case messageXEventTypeOffline:
      gj_gtk_cw_print_offline(cw, msg);
      break;
    case messageXEventTypeDelivered:
      break;
    case messageXEventTypeDisplayed: 
      break;
    case messageXEventTypeComposing:
      gtk_widget_show(GTK_WIDGET(cw->controls.image_composing));
      break;
    case messageXEventTypeCancelled:
      gtk_widget_hide(GTK_WIDGET(cw->controls.image_composing));
      break;
    }
}

void gj_gtk_cw_default_action(rosterItem ri)
{
  ChatWindow cw = NULL;

  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return;
    }
  
  if((cw = gj_gtk_cw_find_by_ri_id(gj_roster_item_get_id(ri))) == NULL)
    {
      g_message("%s: no previous window, creating new for jid:'%s'", __FUNCTION__, gj_roster_item_get_jid(ri)); 

      /* make conversation */
      cw = gj_gtk_cw_new(ri);

      /* set up window */
      gj_gtk_cw_load(cw);
      gj_gtk_cw_setup(cw);
      gj_gtk_cw_show(cw,TRUE);
    }
  else
    {
      g_message("%s: found window, showing window for jid:'%s'", __FUNCTION__, gj_roster_item_get_jid(ri)); 

      /* receive conversation */
      gj_gtk_cw_present(cw); 
    }


  /* if XEvents ask to be notified when the message has 
     been displayed, this is the time to send that notification */
  if(gj_gtk_cw_get_last_recv_x_event_type(cw) & messageXEventTypeDisplayed)
    gj_message_send_x_event(messageXEventTypeDisplayed, gj_roster_item_get_jid(ri), gj_gtk_cw_get_last_recv_x_event_id(cw));
}

void gj_gtk_cw_emote_icon_create_popup_menu(GtkWidget *widget)
{
  GtkWidget *emote_icons_menu;
  GtkWidget *image;
  GtkWidget *mnu_emoteicon_big_grin;
  GtkWidget *mnu_emoteicon_confused;
  GtkWidget *mnu_emoteicon_cool;
  GtkWidget *mnu_emoteicon_eek;
  GtkWidget *mnu_emoteicon_frown;
  GtkWidget *mnu_emoteicon_mad;
  GtkWidget *mnu_emoteicon_red_face;
  GtkWidget *mnu_emoteicon_roll_eyes;
  GtkWidget *mnu_emoteicon_smile;
  GtkWidget *mnu_emoteicon_tongue;
  GtkWidget *mnu_emoteicon_wink;

  if(widget == NULL)
    {
      g_warning("%s: widget was NULL, no widget to relate menu to chat window", __FUNCTION__);
      return;
    }

  emote_icons_menu = gtk_menu_new ();
  gtk_widget_set_name (emote_icons_menu, "emote_icons_menu");

  mnu_emoteicon_big_grin = gtk_image_menu_item_new_with_mnemonic ("Big Grin");
  gtk_widget_set_name (mnu_emoteicon_big_grin, "mnu_emoteicon_big_grin");
  gtk_widget_show (mnu_emoteicon_big_grin);
  gtk_container_add (GTK_CONTAINER (emote_icons_menu), mnu_emoteicon_big_grin);

  image = gtk_image_new_from_file (FILE_EMOTE_BIGGRIN);
  gtk_widget_set_name (image, "image");
  gtk_widget_show (image);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (mnu_emoteicon_big_grin), image);

  mnu_emoteicon_confused = gtk_image_menu_item_new_with_mnemonic ("Confused");
  gtk_widget_set_name (mnu_emoteicon_confused, "mnu_emoteicon_confused");
  gtk_widget_show (mnu_emoteicon_confused);
  gtk_container_add (GTK_CONTAINER (emote_icons_menu), mnu_emoteicon_confused);

  image = gtk_image_new_from_file (FILE_EMOTE_CONFUSED);
  gtk_widget_set_name (image, "image");
  gtk_widget_show (image);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (mnu_emoteicon_confused), image);

  mnu_emoteicon_cool = gtk_image_menu_item_new_with_mnemonic ("Cool");
  gtk_widget_set_name (mnu_emoteicon_cool, "mnu_emoteicon_cool");
  gtk_widget_show (mnu_emoteicon_cool);
  gtk_container_add (GTK_CONTAINER (emote_icons_menu), mnu_emoteicon_cool);

  image = gtk_image_new_from_file (FILE_EMOTE_COOL);
  gtk_widget_set_name (image, "image");
  gtk_widget_show (image);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (mnu_emoteicon_cool), image);

  mnu_emoteicon_eek = gtk_image_menu_item_new_with_mnemonic ("EEk");
  gtk_widget_set_name (mnu_emoteicon_eek, "mnu_emoteicon_eek");
  gtk_widget_show (mnu_emoteicon_eek);
  gtk_container_add (GTK_CONTAINER (emote_icons_menu), mnu_emoteicon_eek);

  image = gtk_image_new_from_file (FILE_EMOTE_EEK);
  gtk_widget_set_name (image, "image");
  gtk_widget_show (image);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (mnu_emoteicon_eek), image);

  mnu_emoteicon_frown = gtk_image_menu_item_new_with_mnemonic ("Frown");
  gtk_widget_set_name (mnu_emoteicon_frown, "mnu_emoteicon_frown");
  gtk_widget_show (mnu_emoteicon_frown);
  gtk_container_add (GTK_CONTAINER (emote_icons_menu), mnu_emoteicon_frown);

  image = gtk_image_new_from_file (FILE_EMOTE_FROWN);
  gtk_widget_set_name (image, "image");
  gtk_widget_show (image);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (mnu_emoteicon_frown), image);

  mnu_emoteicon_mad = gtk_image_menu_item_new_with_mnemonic ("Mad");
  gtk_widget_set_name (mnu_emoteicon_mad, "mnu_emoteicon_mad");
  gtk_widget_show (mnu_emoteicon_mad);
  gtk_container_add (GTK_CONTAINER (emote_icons_menu), mnu_emoteicon_mad);

  image = gtk_image_new_from_file (FILE_EMOTE_MAD);
  gtk_widget_set_name (image, "image");
  gtk_widget_show (image);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (mnu_emoteicon_mad), image);

  mnu_emoteicon_red_face = gtk_image_menu_item_new_with_mnemonic ("Red Face");
  gtk_widget_set_name (mnu_emoteicon_red_face, "mnu_emoteicon_red_face");
  gtk_widget_show (mnu_emoteicon_red_face);
  gtk_container_add (GTK_CONTAINER (emote_icons_menu), mnu_emoteicon_red_face);

  image = gtk_image_new_from_file (FILE_EMOTE_REDFACE);
  gtk_widget_set_name (image, "image");
  gtk_widget_show (image);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (mnu_emoteicon_red_face), image);

  mnu_emoteicon_roll_eyes = gtk_image_menu_item_new_with_mnemonic ("Roll Eyes");
  gtk_widget_set_name (mnu_emoteicon_roll_eyes, "mnu_emoteicon_roll_eyes");
  gtk_widget_show (mnu_emoteicon_roll_eyes);
  gtk_container_add (GTK_CONTAINER (emote_icons_menu), mnu_emoteicon_roll_eyes);

  image = gtk_image_new_from_file (FILE_EMOTE_ROLLEYES);
  gtk_widget_set_name (image, "image");
  gtk_widget_show (image);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (mnu_emoteicon_roll_eyes), image);

  mnu_emoteicon_smile = gtk_image_menu_item_new_with_mnemonic ("Smile");
  gtk_widget_set_name (mnu_emoteicon_smile, "mnu_emoteicon_smile");
  gtk_widget_show (mnu_emoteicon_smile);
  gtk_container_add (GTK_CONTAINER (emote_icons_menu), mnu_emoteicon_smile);

  image = gtk_image_new_from_file (FILE_EMOTE_SMILE);
  gtk_widget_set_name (image, "image");
  gtk_widget_show (image);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (mnu_emoteicon_smile), image);

  mnu_emoteicon_tongue = gtk_image_menu_item_new_with_mnemonic ("Tongue");
  gtk_widget_set_name (mnu_emoteicon_tongue, "mnu_emoteicon_tongue");
  gtk_widget_show (mnu_emoteicon_tongue);
  gtk_container_add (GTK_CONTAINER (emote_icons_menu), mnu_emoteicon_tongue);

  image = gtk_image_new_from_file (FILE_EMOTE_TONGUE);
  gtk_widget_set_name (image, "image");
  gtk_widget_show (image);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (mnu_emoteicon_tongue), image);

  mnu_emoteicon_wink = gtk_image_menu_item_new_with_mnemonic ("Wink");
  gtk_widget_set_name (mnu_emoteicon_wink, "mnu_emoteicon_wink");
  gtk_widget_show (mnu_emoteicon_wink);
  gtk_container_add (GTK_CONTAINER (emote_icons_menu), mnu_emoteicon_wink);

  image = gtk_image_new_from_file (FILE_EMOTE_WINK);
  gtk_widget_set_name (image, "image");
  gtk_widget_show (image);
  gtk_image_menu_item_set_image (GTK_IMAGE_MENU_ITEM (mnu_emoteicon_wink), image);

  /* hook up signals */
  g_signal_connect ((gpointer) mnu_emoteicon_big_grin, "activate",
		    G_CALLBACK (on_mnu_emoteicon_big_grin_activate),
		    (gpointer) widget);
  g_signal_connect ((gpointer) mnu_emoteicon_confused, "activate",
		    G_CALLBACK (on_mnu_emoteicon_confused_activate),
		    (gpointer) widget);
  g_signal_connect ((gpointer) mnu_emoteicon_cool, "activate",
		    G_CALLBACK (on_mnu_emoteicon_cool_activate),
		    (gpointer)widget);
  g_signal_connect ((gpointer) mnu_emoteicon_eek, "activate",
		    G_CALLBACK (on_mnu_emoteicon_eek_activate),
		    (gpointer)widget);
  g_signal_connect ((gpointer) mnu_emoteicon_frown, "activate",
		    G_CALLBACK (on_mnu_emoteicon_frown_activate),
		    (gpointer)widget);
  g_signal_connect ((gpointer) mnu_emoteicon_mad, "activate",
		    G_CALLBACK (on_mnu_emoteicon_mad_activate),
		    (gpointer)widget);
  g_signal_connect ((gpointer) mnu_emoteicon_red_face, "activate",
		    G_CALLBACK (on_mnu_emoteicon_red_face_activate),
		    (gpointer)widget);
  g_signal_connect ((gpointer) mnu_emoteicon_roll_eyes, "activate",
		    G_CALLBACK (on_mnu_emoteicon_roll_eyes_activate),
		    (gpointer)widget);
  g_signal_connect ((gpointer) mnu_emoteicon_smile, "activate",
		    G_CALLBACK (on_mnu_emoteicon_smile_activate),
		    (gpointer)widget);
  g_signal_connect ((gpointer) mnu_emoteicon_tongue, "activate",
		    G_CALLBACK (on_mnu_emoteicon_tongue_activate),
		    (gpointer)widget);
  g_signal_connect ((gpointer) mnu_emoteicon_wink, "activate",
		    G_CALLBACK (on_mnu_emoteicon_wink_activate),
		    (gpointer)widget);
  
  /* show menu */
  gtk_menu_popup(GTK_MENU(emote_icons_menu),
		 NULL,
		 NULL, 
		 NULL, 
		 NULL,
		 1, 
		 gtk_get_current_event_time()); 
}

GdkPixbuf *gj_gtk_cw_emote_icon_get_as_pixbuf(msgEmote icon)
{
  GdkPixbuf *pixbuf = NULL;
  GdkPixbuf *scaled = NULL;
  gchar *filename = NULL;

  switch(icon)
    {
    case emote_biggrin:
      filename = FILE_EMOTE_BIGGRIN;
      break;
    case emote_confused:
      filename = FILE_EMOTE_CONFUSED;
      break;
    case emote_cool:
      filename = FILE_EMOTE_COOL;
      break;
    case emote_eek:
      filename = FILE_EMOTE_EEK;
      break;
    case emote_frown:
      filename = FILE_EMOTE_FROWN;
      break;
    case emote_mad:
      filename = FILE_EMOTE_MAD;
      break;
    case emote_redface:
      filename = FILE_EMOTE_REDFACE;
      break;
    case emote_rolleyes:
      filename = FILE_EMOTE_ROLLEYES;
      break;
    case emote_smile:
      filename = FILE_EMOTE_SMILE;
      break;
    case emote_tongue:
      filename = FILE_EMOTE_TONGUE;
      break;
    case emote_wink:
      filename = FILE_EMOTE_WINK;
      break;
    default:
      g_warning("%s: failed to determin which icon to use from %d", __FUNCTION__, icon);
      return NULL;
      break;
    }

  if((pixbuf = gdk_pixbuf_new_from_file(filename, NULL)) == NULL)
    {
      g_warning("%s: failed to load image file:'%s'", __FUNCTION__, filename);
      return NULL;
    }

  scaled = gdk_pixbuf_scale_simple(pixbuf, 15, 15, GDK_INTERP_BILINEAR);
  g_object_unref(G_OBJECT(pixbuf));
  pixbuf = scaled;

  return pixbuf;
}

void gj_gtk_cw_emote_icon_get_as_text(gchar *emote_icon, gpointer user_data)
{
  GtkTextBuffer *buffer = NULL;
  GtkTextIter iter;

  ChatWindow cw = NULL;

  if(user_data == NULL)
    {
      g_warning("%s: user data is NULL", __FUNCTION__);
      return;
    }
  
  if((cw = gj_gtk_cw_find_by_widget(GTK_WIDGET(user_data))) == NULL)
    {
      g_warning("%s: failed to find window based on button widget", __FUNCTION__);
      return;
    }
  
  /* get buffer and end iterator */
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(cw->controls.textview_send));
  gtk_text_buffer_get_end_iter(GTK_TEXT_BUFFER(buffer), &iter);
  gtk_text_buffer_insert(buffer, &iter,
			 emote_icon, -1);

  /* set focus back to the chat text box */
  gtk_widget_grab_focus(GTK_WIDGET(cw->controls.textview_send));
}

gboolean gj_gtk_cw_presence_update(rosterItem ri, presence pres)
{
  gchar *file = NULL;
  gchar *text = NULL;
  GdkPixbuf *pixbuf = NULL;

  ChatWindow cw = NULL;

  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }

  if(pres == NULL)
    {
      g_warning("%s: presence was NULL", __FUNCTION__);
      return FALSE;
    }

  if((cw = gj_gtk_cw_find_by_ri_id(gj_roster_item_get_id(ri))) == NULL)
    {
      g_message("%s: no chat window for presence update for jid:'%s'", __FUNCTION__, gj_roster_item_get_jid(ri));
      return FALSE;
    }

  switch(gj_presence_get_type(pres))
    {
    case presenceTypeAvailable:
      {
	switch(gj_presence_get_show(pres))
	  {
	  case presenceShowAway:
	    file = FILE_PRESENCE_AWAY;
	    text = g_strdup_printf(_("Away"));
	    break;
	  case presenceShowChat:
	    file = FILE_PRESENCE_CHAT;
	    text = g_strdup_printf(_("Chat"));
	    break;
	  case presenceShowDND:
	    file = FILE_PRESENCE_DND;
	    text = g_strdup_printf(_("Do Not Disturb"));
	    break;
	  case presenceShowNormal:
	    file = FILE_PRESENCE_NORMAL;
	    text = g_strdup_printf(_("Normal"));
	    break;
	  case presenceShowXA:
	    file = FILE_PRESENCE_XA;
	    text = g_strdup_printf(_("eXtended Away"));
	    break;
	  default:
	    file = FILE_PRESENCE_AVAILABLE;
	    text = g_strdup_printf(_("Available"));
	    g_warning("%s: failed to determin presence 'show' from %d", __FUNCTION__, gj_presence_get_show(pres));
	    return FALSE;
	  }
      }
      break;
      
    default:
    case presenceTypeUnAvailable:
      {
	file = FILE_PRESENCE_UNAVAILABLE;
	text = g_strdup_printf(_("Unavailable"));	
      }
      break;
    }

  /* update image and label */
  pixbuf = gj_gtk_support_pixbuf_new_from_file(file);
  gtk_image_set_from_pixbuf(GTK_IMAGE(cw->controls.image_status), pixbuf);
  gtk_label_set_text(GTK_LABEL(cw->controls.label_status), text);

  /* clean up */
  if(text != NULL) g_free(text);
  if(pixbuf != NULL) g_object_unref(G_OBJECT(pixbuf));

  return TRUE;
}

gboolean gj_gtk_cw_print_offline(ChatWindow cw, message msg)
{
  GtkTextView *textview = NULL;
  GtkTextBuffer *buffer = NULL;
  GtkTextMark *mark = NULL;
  GtkTextIter iter;

  gchar *message = NULL;

  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  if(msg == NULL)
    {
      g_warning("%s: msg was NULL", __FUNCTION__);
      return FALSE;
    }

  if((textview = GTK_TEXT_VIEW(cw->controls.textview_receive)) == NULL)
    {
      g_warning("%s: textview was NULL", __FUNCTION__);
      return FALSE;
    }

  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(cw->controls.textview_receive));

  /* set details */
  message = g_strdup_printf(": User Offline\n");

  /* get text direction */
  if(gj_gtk_cw_get_reverse_text_direction(cw) == TRUE)
    gtk_text_buffer_get_start_iter(buffer, &iter);
  else
    gtk_text_buffer_get_end_iter(buffer, &iter);

  if(gj_config_get_mpf_hide_time() == FALSE)
    {
      gchar *time = NULL;
      time = g_strdup_printf("[%s]", gj_support_get_timestamp_from_timet(gj_message_get_time(msg)));

      /* insert at iter */
      gtk_text_buffer_insert_with_tags_by_name(buffer, &iter,
					       time, -1,
					       "blue_foreground",
					       NULL);
      gtk_text_buffer_insert_with_tags_by_name(buffer, &iter,
					       " ", -1,
					       "blue_foreground",
					       NULL);

      /* clean up */
      if(time != NULL) g_free(time);
    }

  gtk_text_buffer_insert_with_tags_by_name(buffer, &iter,
					   message, -1,
					   "bold", 
					   "red_foreground",
					   NULL);

  if(gj_gtk_cw_get_reverse_text_direction(cw) == FALSE)
    {
      mark = gtk_text_buffer_create_mark(buffer, NULL, &iter, FALSE);
      gtk_text_view_scroll_mark_onscreen(textview, mark);
      gtk_text_buffer_delete_mark(buffer, mark);
    }

  return TRUE;
}

gboolean gj_gtk_cw_composing_start(ChatWindow cw)
{
  guint id = 0;

  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  id = g_timeout_add(1000, gj_gtk_cw_composing_timeout_cb, GINT_TO_POINTER(gj_gtk_cw_get_id(cw)));

  gj_gtk_cw_set_composing_timeout_id(cw, id);
  gj_gtk_cw_set_composing_timeout_stop(cw, FALSE);

  return TRUE;
}

gboolean gj_gtk_cw_composing_stop(ChatWindow cw)
{
  guint id = 0;

  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  id = gj_gtk_cw_get_composing_timeout_id(cw);
  gj_gtk_cw_set_composing_timeout_id(cw, -1);
  gj_gtk_cw_set_composing_timeout_stop(cw, TRUE);

  return TRUE;
}

gboolean gj_gtk_cw_composing(ChatWindow cw)
{
  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }
    
  if(gj_gtk_cw_get_composing_timeout_id(cw) == -1)
    {
      rosterItem ri = NULL;

      /* it is stopped so we start it */
      gj_gtk_cw_composing_start(cw);

      if((ri = gj_roster_item_find_by_id(gj_gtk_cw_get_ri_id(cw))) == NULL)
	{
	  g_warning("%s: could not find roster item by id:%d", __FUNCTION__, gj_gtk_cw_get_ri_id(cw));
	  return FALSE;
	}
	
      /* send the composing event */
      if(gj_gtk_cw_get_last_recv_x_event_type(cw) & messageXEventTypeComposing)
	gj_message_send_x_event(messageXEventTypeComposing, gj_roster_item_get_jid(ri), gj_gtk_cw_get_last_recv_x_event_id(cw));
    }

  gj_gtk_cw_set_composing_typing(cw, TRUE);

  return TRUE;
}

gboolean gj_gtk_cw_cancelled(ChatWindow cw)
{
  rosterItem ri = NULL;

  if(cw == NULL)
    {
      g_warning("%s: ChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }
    
  if(gj_gtk_cw_get_composing_timeout_id(cw) != -1)
    {
      /* it is stopped so we start it */
      gj_gtk_cw_composing_stop(cw);
    }


  /* send the composing event */
  if((ri = gj_roster_item_find_by_id(gj_gtk_cw_get_ri_id(cw))) == NULL)
    {
      g_warning("%s: could not find roster item by id:%d", __FUNCTION__, gj_gtk_cw_get_ri_id(cw));
      return FALSE;
    }
    
  if(gj_gtk_cw_get_last_recv_x_event_type(cw) & messageXEventTypeComposing)
    gj_message_send_x_event(messageXEventTypeCancelled, gj_roster_item_get_jid(ri), gj_gtk_cw_get_last_recv_x_event_id(cw));

  gj_gtk_cw_set_composing_typing(cw, FALSE);

  return TRUE;
}

gboolean gj_gtk_cw_composing_timeout_cb(gpointer data)
{
  gint id = 0;

  ChatWindow cw = NULL;

  if(data == NULL)
    {
      g_warning("%s: data was NULL", __FUNCTION__);
      return FALSE;
    }

  id = GPOINTER_TO_INT(data);

  if((cw = gj_gtk_cw_find_by_id(id)) == NULL)
    {
      g_warning("%s: ChatWindow was NULL from id:%d", __FUNCTION__, id);
      return FALSE;
    }
    
  if(gj_gtk_cw_get_composing_timeout_stop(cw) == TRUE)
    return FALSE;

  if(gj_gtk_cw_get_composing_typing(cw) == TRUE)
    {
      gj_gtk_cw_set_composing_duration(cw, gj_gtk_cw_get_composing_duration(cw) + 1);
      gj_gtk_cw_set_composing_stopped(cw, 0); 
    }
  else
    {
      gj_gtk_cw_set_composing_duration(cw, gj_gtk_cw_get_composing_duration(cw) + 1);
      gj_gtk_cw_set_composing_stopped(cw, gj_gtk_cw_get_composing_stopped(cw) + 1);
    }

  /*     g_message("%s: composing duration:%d seconds, composing_stopped:%d seconds",  */
  /* 	      __FUNCTION__, (gint)gj_gtk_cw_get_composing_duration(cw), (gint)gj_gtk_cw_get_composing_stopped(cw)); */
    
  if(gj_gtk_cw_get_composing_duration(cw) >= 16)
    {
      rosterItem ri = NULL;


      if((ri = gj_roster_item_find_by_id(gj_gtk_cw_get_ri_id(cw))) == NULL)
	{
	  g_warning("%s: could not find roster item by id:%d", __FUNCTION__, gj_gtk_cw_get_ri_id(cw));
	  return FALSE;
	}

      /* send another composing event */
      if(gj_gtk_cw_get_last_recv_x_event_type(cw) & messageXEventTypeComposing)
	gj_message_send_x_event(messageXEventTypeComposing, gj_roster_item_get_jid(ri), gj_gtk_cw_get_last_recv_x_event_id(cw));
	
      /* reset all */
      gj_gtk_cw_set_composing_duration(cw, 0);
      gj_gtk_cw_set_composing_stopped(cw, 0);
    }
  else if(gj_gtk_cw_get_composing_stopped(cw) >= 5)
    {
      rosterItem ri = NULL;

      if((ri = gj_roster_item_find_by_id(gj_gtk_cw_get_ri_id(cw))) == NULL)
	{
	  g_warning("%s: could not find roster item by id:%d", __FUNCTION__, gj_gtk_cw_get_ri_id(cw));
	  return FALSE;
	}

      /* send cancelled event */
      if(gj_gtk_cw_get_last_recv_x_event_type(cw) & messageXEventTypeComposing)
	gj_message_send_x_event(messageXEventTypeCancelled, gj_roster_item_get_jid(ri), gj_gtk_cw_get_last_recv_x_event_id(cw));

      gj_gtk_cw_composing_stop(cw);
	
      /* reset all */
      gj_gtk_cw_set_composing_duration(cw, 0);
      gj_gtk_cw_set_composing_stopped(cw, 0);
    }

  gj_gtk_cw_set_composing_typing(cw, FALSE);
    
  return TRUE;
}

/*
 * GUI events
 */
void on_textbuffer_cw_receive_changed(GtkTextBuffer *textbuffer, gpointer user_data)
{
  GtkTextMark *mark = NULL;
  GtkTextIter iter_space;
  GtkTextIter iter_word_start;
  GtkTextIter iter_word_end;
  GtkTextIter iter_start;
  GtkTextIter iter_end;

  gint length = 0;

  /* check WHOLE buffer for urls */
  gtk_text_buffer_get_bounds(textbuffer, &iter_start, &iter_end);
  length = gtk_text_buffer_get_char_count(textbuffer);

  if((mark = gtk_text_buffer_get_mark(textbuffer, "url")) != NULL)
    gtk_text_buffer_get_iter_at_mark(textbuffer, &iter_start, mark);

  /* save last iter */
  iter_word_start = iter_start;
  iter_word_end = iter_start;
  iter_space = iter_start;

  while(TRUE)
    {
      gboolean found = FALSE;
      gchar *buf = NULL;

      /* search for next space */
      found = gtk_text_iter_forward_find_char(&iter_space, on_textview_cw_receive_find_space, NULL, NULL);

      if(found == FALSE)
	break;
      
      /* set word end to the space */
      iter_word_start = iter_word_end;
      iter_word_end = iter_space;

      /* move word end iter forward 1 */
      gtk_text_iter_forward_char(&iter_word_start);

      /* get */
      buf = gtk_text_iter_get_text(&iter_word_start, &iter_word_end);

      if(buf != NULL) 
	{
	  if(gj_support_is_uri(buf) != uriTypeUnknown) 
	    {
	      /* highlight as uri */
	      gtk_text_buffer_apply_tag_by_name(textbuffer, "url", &iter_word_start, &iter_word_end);
	    } 
	  else
	    {
	      /* none highlighted */
	      gtk_text_buffer_remove_tag_by_name(textbuffer, "url", &iter_word_start, &iter_word_end);
	    }

	  g_free(buf);
	}

      /* set mark, then we can check next time from 
	 the last mark and it is more efficient */
      if((mark = gtk_text_buffer_get_mark(textbuffer, "url")) == NULL)
	{
	  if((mark = gtk_text_buffer_create_mark(textbuffer, "url", &iter_space, FALSE)) == NULL)
	    {
	      g_warning("%s: could not create mark!", __FUNCTION__);
	      return;
	    }

	  /* no need to move it, just continue */
	  continue;
	}

      /* move mark to latest iter */
      gtk_text_buffer_move_mark_by_name(textbuffer, "url", &iter_space);
    }
}

gboolean on_textview_cw_receive_motion_notify_event(GtkWidget *widget, GdkEventMotion *event, gpointer user_data)
{
  GdkModifierType state = 0;
  GdkWindow *window = NULL;

  GtkTextView *textview = NULL;
  GtkTextBuffer *buffer = NULL;
  GtkTextIter iter;
  GtkTextIter iter_start;
  GtkTextIter iter_end;
  GtkTextWindowType text_window_type = 0;

  gboolean found_start = FALSE;
  gboolean found_end = FALSE;

  gint pointer_x = 0;
  gint pointer_y = 0;
  gint x = 0;
  gint y = 0;

  gchar *buf = NULL;

  ChatWindow cw = NULL;

  if(event->is_hint)
    {
      window = gdk_window_get_pointer(event->window, &pointer_x, &pointer_y, &state);
    }
  else
    {
      pointer_x = event->x;
      pointer_y = event->y;
      state = event->state;
    }

  if((cw = gj_gtk_cw_find_by_widget(GTK_WIDGET(widget))) == NULL)
    {
      g_warning("%s: ChatWindow was NULL fom widget:0x%.8x", __FUNCTION__, (gint) widget);
      return FALSE;
    }
  
  if((textview = GTK_TEXT_VIEW(widget)) == NULL)
    {
      g_warning("%s: textview was NULL", __FUNCTION__);
      return FALSE;
    }

  if((buffer = GTK_TEXT_BUFFER(gtk_text_view_get_buffer(textview))) == NULL)
    {
      g_warning("%s: buffer was NULL", __FUNCTION__);
      return FALSE;
    }

  text_window_type = gtk_text_view_get_window_type(textview, event->window); 
  gtk_text_view_window_to_buffer_coords(textview, text_window_type, pointer_x, pointer_y, &x, &y);
  gtk_text_view_get_iter_at_location(textview, &iter, x, y);

  if(g_unichar_isspace(gtk_text_iter_get_char(&iter)) == TRUE)
    {
      gdk_window_set_cursor(event->window, NULL);
      return FALSE;
    }

  iter_start = iter;
  iter_end = iter;

  /* search for next space */
  found_start = gtk_text_iter_backward_find_char(&iter_start, on_textview_cw_receive_find_space, NULL, NULL);
  found_end = gtk_text_iter_forward_find_char(&iter_end, on_textview_cw_receive_find_space, NULL, NULL);

  if(found_start == FALSE && found_end == FALSE)
    {
      gdk_window_set_cursor(event->window, NULL);
      return FALSE;
    }

  if(gtk_text_iter_is_start(&iter_start) == FALSE)
    gtk_text_iter_forward_char(&iter_start);
  
  if((buf = gtk_text_buffer_get_text(buffer, &iter_start, &iter_end, FALSE)) == NULL)
    {
      g_warning("%s: buffer was NULL", __FUNCTION__);
      return FALSE;
    }
  
  if(gj_support_is_uri(buf) != uriTypeUnknown)
    {
      /* width wise it works... */
      GdkCursor *cursor = NULL;
      cursor = gdk_cursor_new(GDK_HAND2);
      gdk_window_set_cursor (event->window, cursor);
      
      /* clean up */
      gdk_cursor_unref(cursor);
    }  
  else
    {
      /* return to parent window's cursor */
      gdk_window_set_cursor(event->window, NULL);
    }

      
  /* clean up */
  if(buf != NULL) g_free(buf);

  return FALSE;
}

gboolean on_textview_cw_receive_find_space(gunichar ch, gpointer user_data)
{
  if(g_unichar_isspace(ch) == TRUE)
    return TRUE;

  return FALSE;
}

gboolean on_textview_cw_receive_find_alpha_numeric(gunichar ch, gpointer user_data)
{
  if(g_unichar_isalnum(ch) == TRUE)
    return TRUE;

  return FALSE;
}

gboolean on_textview_cw_receive_button_press_event(GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{
  GdkModifierType state;
  GtkTextBuffer *buffer = NULL;
  GtkTextIter iter;
  GtkTextIter iter_start;
  GtkTextIter iter_end;
  GtkTextWindowType text_window_type = 0;

  gboolean found_start = FALSE;
  gboolean found_end = FALSE;
    
  gint pointer_x = 0;
  gint pointer_y = 0;
  gint x = 0;
  gint y = 0;

  gchar *buf = NULL;

  /* if not mouse 1 return */
  if(event->button != 1)
    return FALSE;

  gdk_window_get_pointer(event->window, &pointer_x, &pointer_y, &state);
    
  if((buffer = GTK_TEXT_BUFFER(gtk_text_view_get_buffer(GTK_TEXT_VIEW(widget)))) == NULL)
    {
      g_warning("%s: buffer was NULL", __FUNCTION__);
      return FALSE;
    }
    
  text_window_type = gtk_text_view_get_window_type(GTK_TEXT_VIEW(widget), event->window); 
  gtk_text_view_window_to_buffer_coords(GTK_TEXT_VIEW(widget), text_window_type, pointer_x, pointer_y, &x, &y);
  gtk_text_view_get_iter_at_location(GTK_TEXT_VIEW(widget), &iter, x, y);
 
  iter_start = iter;
  iter_end = iter;
    
  /* search for next space */
  found_start = gtk_text_iter_backward_find_char(&iter_start, on_textview_cw_receive_find_space, NULL, NULL);
  found_end = gtk_text_iter_forward_find_char(&iter_end, on_textview_cw_receive_find_space, NULL, NULL);
    
  if(found_start == FALSE && found_end == FALSE)
    return FALSE;
    
  if(gtk_text_iter_is_start(&iter_start) == FALSE)
    gtk_text_iter_forward_char(&iter_start);
    
  if((buf = gtk_text_buffer_get_text(buffer, &iter_start, &iter_end, FALSE)) == NULL)
    {
      g_warning("%s: buffer was NULL", __FUNCTION__);
      return FALSE;
    }
  
  if(gj_support_is_uri(buf) != uriTypeUnknown)
    {
      gj_gtk_support_open_url(buf);
    }  
    
  /* clean up */
  if(buf != NULL) g_free(buf);
   
  return FALSE;
}

gboolean on_textview_cw_send_key_release_event(GtkWidget *widget, GdkEventKey *event, gpointer user_data)
{
  GtkTextBuffer *buffer = NULL;
  ChatWindow cw = NULL;

  if((buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(widget))) == NULL)
    {
      g_warning("%s: buffer was NULL", __FUNCTION__);
      return FALSE;
    }

  if((cw = gj_gtk_cw_find_by_widget(GTK_WIDGET(widget))) == NULL)
    {
      g_warning("%s: failed to find window based on widget:0x%.8x", __FUNCTION__, (gint)widget);
      return FALSE;
    }

  /* if was enter then clear box */
  if(event != NULL && event->keyval == GDK_Return && (event->state & GDK_SHIFT_MASK) == FALSE)
    {
      GtkTextIter iter_start;
      GtkTextIter iter_end; 

      gtk_text_buffer_get_bounds(GTK_TEXT_BUFFER(buffer), &iter_start, &iter_end);
      gtk_text_buffer_delete(GTK_TEXT_BUFFER(buffer), &iter_start, &iter_end);

      gtk_widget_set_sensitive(cw->controls.button_send, FALSE);
    }
  else
    {
      gint length = 0;

      /* set toolbar icon */  
      length = gtk_text_buffer_get_char_count(buffer);
      if(length < 1)
	gtk_widget_set_sensitive(cw->controls.button_send, FALSE);
      else
	gtk_widget_set_sensitive(cw->controls.button_send, TRUE);
    }

  return FALSE;
}

gboolean on_textview_cw_send_key_press_event(GtkWidget *widget, GdkEventKey *event, gpointer user_data)
{
  GtkTextBuffer *buffer = NULL;
  GtkTextIter iter_start;
  GtkTextIter iter_end;  

  gchar *text = NULL;
  gint length = 0;

  ChatWindow cw = NULL;
    
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(widget));
  length = gtk_text_buffer_get_char_count(buffer);

  if((cw = gj_gtk_cw_find_by_widget(GTK_WIDGET(widget))) == NULL)
    {
      g_warning("%s: failed to find window based on button widget", __FUNCTION__);
      return FALSE;
    }
  
  gj_gtk_cw_composing(cw);

  if(length < 1)
    {
      /* if enter was pressed... warn user that we can not do this */
      if(event != NULL && event->keyval == GDK_Return && (event->state & GDK_SHIFT_MASK) == FALSE)
	g_warning("%s: can't send NULL or 0 length message", __FUNCTION__);
      
      return FALSE;
    }

  gtk_text_buffer_get_bounds(GTK_TEXT_BUFFER(buffer), &iter_start, &iter_end);
  text = gtk_text_buffer_get_text(GTK_TEXT_BUFFER(buffer), &iter_start, &iter_end, TRUE);
   
  /* send message and clear text box */
  if(event->keyval == GDK_Return && (event->state & GDK_SHIFT_MASK) == FALSE)
    gj_gtk_cw_send_set(cw,text);
    
  /* cleanup */
  if(text != NULL) g_free(text);
    
  return FALSE;
}

void on_button_cw_close_clicked(GtkButton *button, gpointer user_data)
{
  gj_gtk_cw_unload_by_widget(GTK_WIDGET(button));
}

void on_button_cw_send_clicked(GtkButton *button, gpointer user_data)
{
  gchar *text = NULL;
  ChatWindow cw = NULL;

  GtkTextBuffer *buffer = NULL;
  GtkTextIter iter_start;
  GtkTextIter iter_end;

  if((cw = gj_gtk_cw_find_by_widget(GTK_WIDGET(button))) == NULL)
    {
      g_warning("%s: failed to find window based on button widget", __FUNCTION__);
      return;
    }
  
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(cw->controls.textview_send));
  gtk_text_buffer_get_bounds(GTK_TEXT_BUFFER(buffer), &iter_start, &iter_end);
  text = gtk_text_buffer_get_text(GTK_TEXT_BUFFER(buffer), &iter_start, &iter_end, TRUE);

  if(text == NULL || g_utf8_strlen(text, -1) < 1)
    {
      if(text != NULL) g_free(text);

      g_warning("%s: cant send NULL or 0 length message", __FUNCTION__);
      return;
    }

  gtk_text_buffer_delete(GTK_TEXT_BUFFER(buffer), &iter_start, &iter_end);
  gj_gtk_cw_send_set(cw,text);

  g_free(text);
}

void on_button_cw_information_clicked(GtkButton *button, gpointer user_data)
{
  ChatWindow cw = NULL;
  rosterItem ri = NULL;

  if((cw = gj_gtk_cw_find_by_widget(GTK_WIDGET(button))) == NULL)
    {
      g_warning("%s: failed to find window based on button widget", __FUNCTION__);
      return;
    }

  if((ri = gj_roster_item_find_by_id(gj_gtk_cw_get_ri_id(cw))) == NULL)
    {
      g_warning("%s: roster item was NULL, can not load user information", __FUNCTION__);
      return;
    }

  gj_gtk_ui_load(gj_roster_item_get_jid(ri));
}

void on_button_cw_history_clicked(GtkButton *button, gpointer user_data)
{
  ChatWindow cw = NULL;
  history hist = NULL;

  /*   gchar *url = NULL; */

  if((cw = gj_gtk_cw_find_by_widget(GTK_WIDGET(button))) == NULL)
    {
      g_warning("%s: failed to find window based on button widget", __FUNCTION__);
      return;
    }

  if((hist = gj_history_find_by_ri_id(gj_gtk_cw_get_ri_id(cw))) == NULL)
    {
      g_warning("%s: could not find history based on roster item id:%d", __FUNCTION__, gj_gtk_cw_get_ri_id(cw));
      return;
    }

  gj_gtk_support_open_url(gj_history_get_filename(hist));
}

void on_button_cw_add_clicked(GtkButton *button, gpointer user_data)
{
  ChatWindow cw = NULL;
  rosterItem ri = NULL;

  if((cw = gj_gtk_cw_find_by_widget(GTK_WIDGET(button))) == NULL)
    {
      g_warning("%s: failed to find window based on button widget", __FUNCTION__);
      return;
    }

  if((ri = gj_roster_item_find_by_id(gj_gtk_cw_get_ri_id(cw))) == NULL)
    {
      g_warning("%s: roster item was NULL, can not add contact", __FUNCTION__);
      return;
    }

  gj_gtk_ac_load(gj_roster_item_get_jid(ri));
}

void on_button_cw_emote_clicked(GtkButton *button, gpointer user_data)
{
  gj_gtk_cw_emote_icon_create_popup_menu(GTK_WIDGET(button));
}

void on_mnu_emoteicon_big_grin_activate(GtkMenuItem *menuitem, gpointer user_data)
{
  gj_gtk_cw_emote_icon_get_as_text(EMOTE_BIGGRIN, user_data);
}

void on_mnu_emoteicon_confused_activate(GtkMenuItem *menuitem, gpointer user_data)
{
  gj_gtk_cw_emote_icon_get_as_text(EMOTE_CONFUSED, user_data);
}

void on_mnu_emoteicon_cool_activate(GtkMenuItem *menuitem, gpointer user_data)
{
  gj_gtk_cw_emote_icon_get_as_text(EMOTE_COOL, user_data);
}

void on_mnu_emoteicon_eek_activate(GtkMenuItem *menuitem, gpointer user_data)
{
  gj_gtk_cw_emote_icon_get_as_text(EMOTE_EEK, user_data);
}

void on_mnu_emoteicon_frown_activate(GtkMenuItem *menuitem, gpointer user_data)
{
  gj_gtk_cw_emote_icon_get_as_text(EMOTE_FROWN, user_data);
}

void on_mnu_emoteicon_mad_activate(GtkMenuItem *menuitem, gpointer user_data)
{
  gj_gtk_cw_emote_icon_get_as_text(EMOTE_MAD, user_data);
}

void on_mnu_emoteicon_red_face_activate(GtkMenuItem *menuitem, gpointer user_data)
{
  gj_gtk_cw_emote_icon_get_as_text(EMOTE_REDFACE, user_data);
}

void on_mnu_emoteicon_roll_eyes_activate(GtkMenuItem *menuitem, gpointer user_data)
{
  gj_gtk_cw_emote_icon_get_as_text(EMOTE_ROLLEYES, user_data);
}

void on_mnu_emoteicon_smile_activate(GtkMenuItem *menuitem, gpointer user_data)
{
  gj_gtk_cw_emote_icon_get_as_text(EMOTE_SMILE, user_data);
}

void on_mnu_emoteicon_tongue_activate(GtkMenuItem *menuitem, gpointer user_data)
{
  gj_gtk_cw_emote_icon_get_as_text(EMOTE_TONGUE, user_data);
}

void on_mnu_emoteicon_wink_activate(GtkMenuItem *menuitem, gpointer user_data)
{
  gj_gtk_cw_emote_icon_get_as_text(EMOTE_WINK, user_data);
}

gboolean on_chat_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data)
{
  gj_gtk_cw_unload_by_widget(widget);

  return FALSE;
}

#ifdef __cplusplus
}
#endif
