/* 
 * $RCSfile: gj_gtk_debug.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/07/31 16:43:54 $
 * 
 * $Revision: 1.11 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <string.h>

#include <glib.h>
#include <gtk/gtk.h>
#include <glade/glade.h>

#include "common.h"

#include "gj_main.h"
#include "gj_parser.h"

#include "gj_gtk_debug.h"

struct 
{
  GtkWidget *window;

  GtkWidget *textview_input;
  GtkWidget *textview_output;

  GtkWidget *button_input;
  GtkWidget *button_output;
  
} ctlDebug;


/*
 * GUI functions
 */
void on_button_debug_input_clicked(GtkButton *button, gpointer user_data);
void on_button_debug_output_clicked(GtkButton *button, gpointer user_data);

gboolean on_debug_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data);

/*
 * code
 */
gboolean gj_gtk_debug_load()
{
  GladeXML *xml = NULL;

  if(ctlDebug.window != NULL)
    {
	gtk_window_present(GTK_WINDOW(ctlDebug.window));
	return TRUE;
    }

  if((xml = glade_xml_new(gnome_jabber_get_glade_file(),
			  "debug",
			  NULL)) == NULL)
    {
      g_warning("gj_gtk_debug_load: couldn't open glade file at root 'debug'");
      return FALSE;
    }
  
  ctlDebug.window = glade_xml_get_widget(xml,"debug");

  ctlDebug.textview_input = glade_xml_get_widget(xml,"textview_debug_input");
  ctlDebug.textview_output = glade_xml_get_widget(xml,"textview_debug_output");

  ctlDebug.button_input = glade_xml_get_widget(xml,"button_debug_input");
  ctlDebug.button_output = glade_xml_get_widget(xml,"button_debug_output");
  
  gdk_window_set_decorations (GDK_WINDOW(ctlDebug.window->window), 
			      GDK_DECOR_ALL);

#ifndef WIN32
  glade_xml_signal_autoconnect(xml);
#else
  g_signal_connect ((gpointer) ctlDebug.window, "delete_event",
		    G_CALLBACK (on_debug_delete_event),
		    NULL);
  g_signal_connect ((gpointer) ctlDebug.button_output, "clicked",
		    G_CALLBACK (on_button_debug_output_clicked),
		    NULL);
  g_signal_connect ((gpointer) ctlDebug.button_input, "clicked",
		    G_CALLBACK (on_button_debug_input_clicked),
		    NULL);
#endif

  /* clean up */
  if(xml != NULL)
    {
      g_object_unref(G_OBJECT(xml));
      xml = NULL;
    }

  return TRUE;
}

gboolean gj_gtk_debug_unload()
{
  gtk_widget_destroy(ctlDebug.window);
  ctlDebug.window = NULL;

  return TRUE;
}

#if 0
gboolean gj_gtk_debug_clear()
{
  GtkTextBuffer *buffer = NULL;
  GtkTextIter iter_start;
  GtkTextIter iter_end;

  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(ctlDebug.textview_log));
  gtk_text_buffer_get_bounds(GTK_TEXT_BUFFER(buffer), &iter_start, &iter_end);
  
  /* clear buffer */
  gtk_text_buffer_delete(GTK_TEXT_BUFFER(buffer), &iter_start, &iter_end);
  
  return TRUE;
}
#endif

/*
 * GUI events
 */
void on_button_debug_input_clicked(GtkButton *button, gpointer user_data)
{
    GtkTextBuffer *buffer = NULL;
    GtkTextIter iter_start;
    GtkTextIter iter_end; 

    gchar *text = NULL;

    buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(ctlDebug.textview_input));
    gtk_text_buffer_get_bounds(GTK_TEXT_BUFFER(buffer), &iter_start, &iter_end);
    text = gtk_text_buffer_get_text(GTK_TEXT_BUFFER(buffer), &iter_start, &iter_end, TRUE);

    g_message("on_button_debug_input_clicked: inputing buffer to XML parser: '%s'", text);

    if(g_utf8_validate(text, -1, NULL) == FALSE)
	g_warning("%s: text is not valid utf8", __FUNCTION__);
    
    gj_parser_push_input(NULL, text, g_utf8_strlen(text, -1));

    /* clean up */
    g_free(text);
}

void on_button_debug_output_clicked(GtkButton *button, gpointer user_data)
{
    GtkTextBuffer *buffer = NULL;
    GtkTextIter iter_start;
    GtkTextIter iter_end; 

    gchar *text = NULL;
    
    buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(ctlDebug.textview_output));
    gtk_text_buffer_get_bounds(GTK_TEXT_BUFFER(buffer), &iter_start, &iter_end);
    text = gtk_text_buffer_get_text(GTK_TEXT_BUFFER(buffer), &iter_start, &iter_end, TRUE);

    g_message("on_button_debug_input_clicked: outputing buffer to server: '%s'", text);

    gj_parser_push_output_from_text(text);

    /* clean up */
    g_free(text);
}

gboolean on_debug_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data)
{
    gj_gtk_debug_unload();
  return TRUE;
}

#ifdef __cplusplus
}
#endif
