/* 
 * $RCSfile: gj_gtk_group_chat.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/08/23 12:06:20 $
 * 
 * $Revision: 1.7 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <string.h>
#include <stdlib.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <glade/glade.h>

#include "common.h"

#include "gj_main.h"
#include "gj_roster.h"
#include "gj_message.h"
#include "gj_presence.h"
#include "gj_event.h"
#include "gj_support.h"
#include "gj_config.h"

#include "gj_gtk_group_chat.h"
#include "gj_gtk_msg.h"
#include "gj_gtk_support.h"


/* columns */
enum
  {
    COL_GROUP_CHAT_IMAGE,
    COL_GROUP_CHAT_JID,
    COL_GROUP_CHAT_NAME,
    COL_GROUP_CHAT_ENABLED,
    COL_GROUP_CHAT_COUNT
  };


struct t_controls
{
  GtkWidget *window;

  GtkWidget *entry;
  GtkWidget *textview_receive;
  GtkWidget *textview_send;
  GtkWidget *treeview;

};

struct t_GroupChatWindow
{
  struct t_controls controls;

  gchar *host;
  gchar *room;
  gchar *nickname;

  gint ri_id;
};

GList *group_chat_window_list = NULL;

/* list functions */
gboolean gj_gtk_gcw_list_add(GroupChatWindow gcw);
gboolean gj_gtk_gcw_list_del(GroupChatWindow gcw);
GList *gj_gtk_gcw_list_get();

/* item functions */
GroupChatWindow gj_gtk_gcw_new(const gchar *host, const gchar *room, const gchar *nickname);
gboolean gj_gtk_gcw_free(GroupChatWindow gcw);
void gj_gtk_gcw_destroy(GroupChatWindow gcw);

/* find functions */
GroupChatWindow gj_gtk_gcw_find_by_host_and_room(const gchar *host, const gchar *room);
GroupChatWindow gj_gtk_gcw_find_by_widget(GtkWidget *widget);
GroupChatWindow gj_gtk_gcw_find_by_ri_id(gint id);


/*
 * member functions
 */
     
/* gets */
const gint gj_gtk_gcw_get_ri_id();

const gchar *gj_gtk_gcw_get_host();
const gchar *gj_gtk_gcw_get_room();
const gchar *gj_gtk_gcw_get_nickname();

/* sets */
gboolean gj_gtk_gcw_set_ri_id(GroupChatWindow gcw, const gint ri_id);

gboolean gj_gtk_gcw_set_host(GroupChatWindow gcw, const gchar *host);
gboolean gj_gtk_gcw_set_room(GroupChatWindow gcw, const gchar *room);
gboolean gj_gtk_gcw_set_nickname(GroupChatWindow gcw, const gchar *nickname);


/*
 * window functions 
 */
gboolean gj_gtk_gcw_load();
gboolean gj_gtk_gcw_unload(GroupChatWindow gcw);

gboolean gj_gtk_gcw_setup(GroupChatWindow gcw);
gboolean gj_gtk_gcw_send();

gboolean gj_gtk_gcw_disconnect(GroupChatWindow gcw);

gboolean gj_gtk_gcw_handle_presence(GroupChatWindow gcw, presence pres);
gboolean gj_gtk_gcw_handle_message(GroupChatWindow gcw, message msg);

gboolean gj_gtk_gcw_send(GroupChatWindow gcw, gchar *text);
gboolean gj_gtk_gcw_receive(GroupChatWindow gcw, message msg);


/* 
 * model functions 
 */
gboolean gj_gtk_gcw_model_unload(GroupChatWindow gcw);

gboolean gj_gtk_gcw_model_setup(GroupChatWindow gcw);
gboolean gj_gtk_gcw_model_populate_columns(GtkTreeView *view);

gboolean gj_gtk_gcw_model_find_iter_by_jid(GroupChatWindow gcw, const gchar *jid, GtkTreeIter *ret_iter);

gboolean gj_gtk_gcw_model_add(GroupChatWindow gcw, presence pres);
gboolean gj_gtk_gcw_model_remove(GroupChatWindow gcw, presence pres);
gboolean gj_gtk_gcw_model_update(GroupChatWindow gcw, presence pres);


/*
 * debugging
 */
void gj_gtk_gcw_print_all();
void gj_gtk_gcw_print(GroupChatWindow gcw);


/*
 * gui functions 
 */
void on_textbuffer_gcw_receive_changed(GtkTextBuffer *textbuffer, gpointer user_data);
gboolean on_textview_gcw_receive_motion_notify_event(GtkWidget *widget, GdkEventMotion *event, gpointer user_data);
gboolean on_textview_gcw_receive_find_space(gunichar ch, gpointer user_data);
gboolean on_textview_gcw_receive_find_alpha_numeric(gunichar ch, gpointer user_data);
gboolean on_textview_gcw_receive_button_press_event(GtkWidget *widget, GdkEventButton *event, gpointer user_data);
gboolean on_textview_gcw_send_key_release_event(GtkWidget *widget, GdkEventKey *event, gpointer user_data);
gboolean on_textview_gcw_send_key_press_event(GtkWidget *widget, GdkEventKey *event, gpointer user_data);

gboolean on_group_chat_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data);


/*
 * code
 */

/* list handling */
gboolean gj_gtk_gcw_list_add(GroupChatWindow gcw)
{
  if(gcw == NULL)
    {
      g_warning("%s: chatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  group_chat_window_list = g_list_append(group_chat_window_list, gcw);
  return TRUE;
}

gboolean gj_gtk_gcw_list_del(GroupChatWindow gcw)
{
  if(gcw == NULL)
    {
      g_warning("%s: chatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  group_chat_window_list = g_list_remove(group_chat_window_list, gcw);
  return TRUE;
}

GList *gj_gtk_gcw_list_get()
{
  return group_chat_window_list;
}

GroupChatWindow gj_gtk_gcw_new(const gchar *host, const gchar *room, const gchar *nickname)
{
  GroupChatWindow gcw = NULL;

  if(host == NULL)
    {
      g_warning("%s: host was NULL", __FUNCTION__);
      return NULL;
    }

  if(room == NULL)
    {
      g_warning("%s: room was NULL", __FUNCTION__);
      return NULL;
    }

  if(nickname == NULL)
    {
      g_warning("%s: nickname was NULL", __FUNCTION__);
      return NULL;
    }

  if((gcw = (GroupChatWindow) g_new(struct t_GroupChatWindow,1)) != NULL)
    {
      gcw->ri_id = -1;

      gcw->host = g_strdup(host);
      gcw->room = g_strdup(room);
      gcw->nickname = g_strdup(nickname);

      gj_gtk_gcw_list_add(gcw);
    }

  return gcw;
}

gboolean gj_gtk_gcw_free(GroupChatWindow gcw)
{
  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  /* free from list */
  gj_gtk_gcw_list_del(gcw);
  
  /* destroy members */
  gj_gtk_gcw_destroy(gcw);

  /* free struct */
  g_free(gcw);
  
  return TRUE;
}

void gj_gtk_gcw_destroy(GroupChatWindow gcw)
{
  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return;
    }

  gcw->ri_id = -1;

  if(gcw->host != NULL)
    {
      g_free(gcw->host);
      gcw->host = NULL;
    }

  if(gcw->room != NULL)
    {
      g_free(gcw->room);
      gcw->room = NULL;
    }

  if(gcw->nickname != NULL)
    {
      g_free(gcw->nickname);
      gcw->nickname = NULL;
    }
}

GroupChatWindow gj_gtk_gcw_find_by_host_and_room(const gchar *host, const gchar *room)
{
  gint index = 0;

  if(host == NULL)
    {
      g_warning("%s: host was NULL", __FUNCTION__);
      return NULL;
    }

  if(room == NULL)
    {
      g_warning("%s: room was NULL", __FUNCTION__);
      return NULL;
    }

  for(index=0;index<g_list_length(group_chat_window_list);index++)
    {
      GroupChatWindow gcw = g_list_nth_data(group_chat_window_list, index);

      if(gcw == NULL) continue;
      if(gj_gtk_gcw_get_host(gcw) == NULL) continue;
      if(gj_gtk_gcw_get_room(gcw) == NULL) continue;

      if(strcmp(gj_gtk_gcw_get_host(gcw), host) == 0 &&
	 strcmp(gj_gtk_gcw_get_room(gcw), room) == 0)
	return gcw;
    }

  return NULL;  
}

GroupChatWindow gj_gtk_gcw_find_by_widget(GtkWidget *widget)
{
  gint index = 0;

  GtkWidget *toplevel = NULL;

  if(widget == NULL)
    {
      g_warning("%s: widget was NULL", __FUNCTION__);
      return NULL;
    }

  /* get top level widget */
  toplevel = gtk_widget_get_toplevel(widget);

  if(toplevel == NULL)
    {
      g_warning("%s: top level widget was NULL", __FUNCTION__);
      return NULL;
    }

  for(index=0;index<g_list_length(group_chat_window_list);index++)
    {
      GroupChatWindow gcw = g_list_nth_data(group_chat_window_list, index);

      if(gcw == NULL) continue;
      if(gcw->controls.window == toplevel) return gcw;
    }

  return NULL;  
}

GroupChatWindow gj_gtk_gcw_find_by_ri_id(gint id)
{
  gint index = 0;

  if(id < 0)
    {
      g_warning("%s: id:%d was NULL", __FUNCTION__, id);
      return NULL;
    }

  for(index=0;index<g_list_length(group_chat_window_list);index++)
    {
      GroupChatWindow gcw = g_list_nth_data(group_chat_window_list, index);

      if(gcw == NULL) continue;
      if(gcw->ri_id == id) return gcw;
    }

  return NULL;  
}

/* 
 * member functions 
 */

/* gets */ 
const gint gj_gtk_gcw_get_ri_id(GroupChatWindow gcw)
{
  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return -1;
    }

  return gcw->ri_id;
}

const gchar *gj_gtk_gcw_get_host(GroupChatWindow gcw)
{
  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return NULL;
    }

  return gcw->host;
}

const gchar *gj_gtk_gcw_get_room(GroupChatWindow gcw)
{
  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return NULL;
    }

  return gcw->room;
}

const gchar *gj_gtk_gcw_get_nickname(GroupChatWindow gcw)
{
  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return NULL;
    }

  return gcw->nickname;
}

/* sets */
gboolean gj_gtk_gcw_set_ri_id(GroupChatWindow gcw, const gint ri_id)
{
  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }  
  
  if(ri_id < 0)
    {
      g_warning("%s: id:%d was < 0", __FUNCTION__, ri_id);
      return FALSE;
    }    

  gcw->ri_id = ri_id;
  return TRUE;
}

gboolean gj_gtk_gcw_set_host(GroupChatWindow gcw, const gchar *host)
{
  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }  
  
  if(host == NULL)
    {
      g_warning("%s: host was NULL", __FUNCTION__);
      return FALSE;
    }    

  if(gcw->host != NULL)
    {
      g_free(gcw->host);
      gcw->host = NULL;
    }

  gcw->host = g_strdup(host);
  return TRUE;
}

gboolean gj_gtk_gcw_set_room(GroupChatWindow gcw, const gchar *room)
{
  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }  
  
  if(room == NULL)
    {
      g_warning("%s: room was NULL", __FUNCTION__);
      return FALSE;
    }    

  if(gcw->room != NULL)
    {
      g_free(gcw->room);
      gcw->room = NULL;
    }

  gcw->room = g_strdup(room);
  return TRUE;
}

gboolean gj_gtk_gcw_set_nickname(GroupChatWindow gcw, const gchar *nickname)
{
  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }  
  
  if(nickname == NULL)
    {
      g_warning("%s: nickname was NULL", __FUNCTION__);
      return FALSE;
    }    

  if(gcw->nickname != NULL)
    {
      g_free(gcw->nickname);
      gcw->nickname = NULL;
    }

  gcw->nickname = g_strdup(nickname);
  return TRUE;
}

void gj_gtk_gcw_print_all()
{
  gint index = 0;

  /* loop through all contacts on roster */
  for(index=0;index<g_list_length(group_chat_window_list);index++)
    {
      GroupChatWindow gcw = g_list_nth_data(group_chat_window_list, index);
      if(gcw == NULL) continue;

      gj_gtk_gcw_print(gcw);
    }  
}

void gj_gtk_gcw_print(GroupChatWindow gcw)
{
  if(gcw == NULL)
    return;

  g_message("++++++++++++++++++++++++++++++++++++++++++++++++");
  g_message("++++");
  g_message("++++ ri id:'%d'", gcw->ri_id);
  g_message("++++");
  g_message("++++ host:'%s'", gcw->host);
  g_message("++++ room:'%s'", gcw->room);
  g_message("++++ nickname:'%s'", gcw->nickname);
  g_message("++++");
  g_message("++++++++++++++++++++++++++++++++++++++++++++++++");
}

/*
 * window functions
 */
gboolean gj_gtk_gcw_load(GroupChatWindow gcw)
{
  GladeXML *xml = NULL;

  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  if((xml = glade_xml_new(gnome_jabber_get_glade_file(),
			  "group_chat",
			  NULL)) == NULL)
    {
      g_warning("gj_gtk_gcw_load: couldn't open glade file at root 'group_chat'");
      return FALSE;
    }
  
  gcw->controls.window = glade_xml_get_widget(xml,"group_chat");

  gcw->controls.entry = glade_xml_get_widget(xml,"entry_gcw");
  gcw->controls.textview_receive = glade_xml_get_widget(xml,"textview_gcw_receive");
  gcw->controls.textview_send = glade_xml_get_widget(xml,"textview_gcw_send");
  gcw->controls.treeview = glade_xml_get_widget(xml,"treeview_gcw");

  gdk_window_set_decorations (GDK_WINDOW(gcw->controls.window->window), 
			      GDK_DECOR_ALL);

#ifndef WIN32
  glade_xml_signal_autoconnect(xml);
#else
  gtk_signal_connect (GTK_OBJECT (gcw->controls.window), "delete_event",
                      GTK_SIGNAL_FUNC (on_group_chat_delete_event),
                      NULL);
  g_signal_connect ((gpointer) gcw->controls.textview_receive, "button_press_event",
		    G_CALLBACK (on_textview_gcw_receive_button_press_event),
		    NULL);
  g_signal_connect ((gpointer) gcw->controls.textview_receive, "motion_notify_event",
		    G_CALLBACK (on_textview_gcw_receive_motion_notify_event),
		    NULL);
  g_signal_connect ((gpointer) gcw->controls.textview_send, "key_press_event",
		    G_CALLBACK (on_textview_gcw_send_key_press_event),
		    NULL);
  g_signal_connect ((gpointer) gcw->controls.textview_send, "key_release_event",
		    G_CALLBACK (on_textview_gcw_send_key_release_event),
		    NULL);
#endif

  gj_gtk_gcw_setup(gcw);

  /* clean up */
  if(xml != NULL)
    {
      g_object_unref(G_OBJECT(xml));
      xml = NULL;
    }

  return TRUE;
}

gboolean gj_gtk_gcw_unload(GroupChatWindow gcw)
{
  rosterItem ri = NULL;

  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  /* remove roster item */
  if((ri = gj_roster_item_find_by_id(gj_gtk_gcw_get_ri_id(gcw))) != NULL)
    gj_roster_item_free(ri);
  
  /* kill the window */
  gtk_widget_destroy(gcw->controls.window);
  
  /* free the mem */
  gj_gtk_gcw_free(gcw);
  
  g_message("%s: unloaded...", __FUNCTION__);
  
  return TRUE;
}

gboolean gj_gtk_gcw_setup(GroupChatWindow gcw)
{
  GtkTextBuffer *buffer = NULL;

  gchar *text = NULL;

  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  /* set title */
  text = g_strdup_printf("%s %s@%s", 
			 _("Chatting in"), 
			 gj_gtk_gcw_get_room(gcw),
			 gj_gtk_gcw_get_host(gcw));

  gtk_window_set_title(GTK_WINDOW(gcw->controls.window), text);
  g_free(text);

  /* set buffer changed callback */
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gcw->controls.textview_receive));

  g_signal_connect(G_OBJECT(buffer), "changed",  
 		   G_CALLBACK(on_textbuffer_gcw_receive_changed),  
 		   NULL); 

  /* set styles */
  gtk_text_buffer_create_tag(buffer, "bold", 
			     "weight", PANGO_WEIGHT_BOLD, 
			     NULL);
  gtk_text_buffer_create_tag(buffer, "italic", 
			     "style", PANGO_STYLE_ITALIC, 
			     NULL);  
  gtk_text_buffer_create_tag(buffer, "blue_foreground", 
			     "foreground", "darkblue", 
			     NULL);  
  gtk_text_buffer_create_tag(buffer, "red_foreground", 
			     "foreground", "darkred", 
			     NULL);  
  gtk_text_buffer_create_tag(buffer, "green_foreground", 
			     "foreground", "darkgreen", 
			     NULL);  
  gtk_text_buffer_create_tag(buffer, "black_foreground", 
			     "foreground", "black", 
			     NULL); 
  gtk_text_buffer_create_tag(buffer, "url", 
			     "foreground", "blue", 
			     NULL); 

  /* set up model */
  gj_gtk_gcw_model_setup(gcw);

  return TRUE;
}

gboolean gj_gtk_gcw_join(const gchar *host, const gchar *room, const gchar *nickname)
{
  GroupChatWindow gcw = NULL;
  rosterItem ri = NULL;
  gchar *jid = NULL;

  if(nickname == NULL || g_utf8_strlen(nickname, -1) < 1)
    {
      g_warning("%s: nickname was NULL", __FUNCTION__);
      return FALSE;
    }

  if(host == NULL || g_utf8_strlen(host, -1) < 1)
    {
      g_warning("%s: host was NULL", __FUNCTION__);
      return FALSE;
    }

  if(room == NULL || g_utf8_strlen(room, -1) < 1)
    {
      g_warning("%s: room was NULL", __FUNCTION__);
      return FALSE;
    }

  if((gcw = gj_gtk_gcw_find_by_host_and_room(host, room)) != NULL)
    {
      g_message("%s: there is already a GroupChatWindow open to host:'%s', room:'%s'", __FUNCTION__, host, room);
      gtk_window_present(GTK_WINDOW(gcw->controls.window));

      return TRUE;
    }

  if((gcw = gj_gtk_gcw_new(host, room, nickname)) == NULL)
    return FALSE;

  if(gj_gtk_gcw_load(gcw) == FALSE)
    return FALSE;

  /* set up jid */
  jid = g_strdup_printf("%s@%s/%s", room, host, nickname);

  if((ri = gj_roster_item_new(jid)) == NULL)
    {
      if(jid != NULL) g_free(jid);
      return FALSE;
    }

  /* remember so we can free it up after */
  gj_gtk_gcw_set_ri_id(gcw, gj_roster_item_get_id(ri));
  
  /* set up roster item */
  gj_roster_item_set_type(ri, rosterItemTypeGroupChat);

  /* send presence... and effectively join */
  gj_presence_send_to(jid, presenceTypeAvailable, presenceShowNormal, "Online", 0);

  /* clean up */
  if(jid != NULL) g_free(jid);

  return TRUE;
}

gboolean gj_gtk_gcw_disconnect(GroupChatWindow gcw)
{
  gchar *jid = NULL;
  
  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  /* set up jid */
  jid = g_strdup_printf("%s@%s/%s", gj_gtk_gcw_get_room(gcw), gj_gtk_gcw_get_host(gcw), gj_gtk_gcw_get_nickname(gcw));
  
  /* send unavailable presence back */
  gj_presence_send_to(jid, presenceTypeUnAvailable, presenceShowNormal, NULL, 0);

  /* clean up */
  if(jid != NULL) g_free(jid);
  
  return TRUE;
}

gboolean gj_gtk_gcw_post_by_event(event ev)
{
  GroupChatWindow gcw = NULL;

  if(ev == NULL)
    {
      g_warning("%s: event was NULL", __FUNCTION__);
      return FALSE;
    }
  
  if((gcw = gj_gtk_gcw_find_by_ri_id(gj_event_get_ri_id(ev))) == NULL)
    {
      g_warning("%s: could not find GroupChatWindow by roster item id:%d", __FUNCTION__, gj_event_get_ri_id(ev));
      return FALSE;
    }

  switch(gj_event_get_type(ev))
    {
    case eventTypePresence:
      gj_gtk_gcw_handle_presence(gcw, (presence)gj_event_get_data(ev));
      break;

    case eventTypeMessage:
      gj_gtk_gcw_handle_message(gcw, (message)gj_event_get_data(ev));
      break;

    default:
      break;
    }

  return TRUE;
}

gboolean gj_gtk_gcw_handle_presence(GroupChatWindow gcw, presence pres)
{
  GtkTreeIter iter;

  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  if(pres == NULL)
    {
      g_warning("%s: presence was NULL", __FUNCTION__);
      return FALSE;
    }

  if(gj_presence_get_type(pres) == presenceTypeAvailable)
    {
      if(gj_gtk_gcw_model_find_iter_by_jid(gcw, gj_presence_get_from_jid(pres), &iter) == FALSE)
	{
	  /* not found, add contact */
	  gj_gtk_gcw_model_add(gcw, pres);
	}
      else
	{
	  /* update presence */
	  gj_gtk_gcw_model_update(gcw, pres);
	}
    }
  else
    {
      gchar *jid = NULL;

      /* set up jid */
      jid = g_strdup_printf("%s@%s/%s", gj_gtk_gcw_get_room(gcw), gj_gtk_gcw_get_host(gcw), gj_gtk_gcw_get_nickname(gcw));
      
      /* remove */
      if(strcmp(gj_presence_get_from_jid(pres), jid) == 0)
	gj_gtk_gcw_unload(gcw);
      else
	gj_gtk_gcw_model_remove(gcw, pres);

      /* clean up */
      if(jid != NULL) g_free(jid);
    }

  return TRUE;
}

gboolean gj_gtk_gcw_handle_message(GroupChatWindow gcw, message msg)
{
  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return FALSE;
    }

  if(gj_message_get_type(msg) != messageTypeGroupChat)
    return TRUE;
  
  gj_gtk_gcw_receive(gcw, msg);
  return TRUE;
}

gboolean gj_gtk_gcw_send(GroupChatWindow gcw, gchar *text)
{
  gchar *jid = NULL;

  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  if(text == NULL)
    {
      g_warning("%s: text was NULL, can not send NULL", __FUNCTION__);
      return FALSE;
    }

  /* set up jid */
  jid = g_strdup_printf("%s@%s", gj_gtk_gcw_get_room(gcw), gj_gtk_gcw_get_host(gcw));

  /* send message */
  gj_message_send(messageTypeGroupChat, 
		  jid, 
		  NULL, 
		  text, 
		  NULL);
  
#if 0
  /* log to history */
  gj_history_write_sent_by_ri_id(gj_gtk_cw_get_ri_id(cw), 
				 (gchar*)gj_support_get_timestamp_nice(), 
				 NULL, 
				 text);
#endif

  /* clean up */
  if(jid != NULL) g_free(jid);

  return TRUE;
}

gboolean gj_gtk_gcw_receive(GroupChatWindow gcw, message msg)
{
  GtkTextView *textview = NULL;
  GtkTextBuffer *buffer = NULL;
  GtkTextMark *mark = NULL;
  GtkTextIter iter;

  gchar *message = NULL;
  gchar *time = NULL;
  gchar *colour_time_and_name = NULL;
  gchar *colour_content = NULL;
  glong i = 0;

  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  if(msg == NULL)
    {
      g_warning("%s: msg was NULL", __FUNCTION__);
      return FALSE;
    }

  if((textview = GTK_TEXT_VIEW(gcw->controls.textview_receive)) == NULL)
    {
      g_warning("%s: textview was NULL", __FUNCTION__);
      return FALSE;
    }

  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gcw->controls.textview_receive));
  time = g_strdup_printf("[%s]", gj_support_get_timestamp_from_timet(gj_message_get_time(msg)));

  /* set details */
  message = g_strdup_printf("%s%s\n", gj_message_get_from_resource(msg) == NULL ? "" : ": ", gj_message_get_body(msg));

  /* set colours */
  if(gj_message_get_from_resource(msg) != NULL)
    {
      if(strcmp(gj_message_get_from_resource(msg), gj_gtk_gcw_get_nickname(gcw)) == 0)
	colour_time_and_name = g_strdup_printf("green_foreground");
      else
	colour_time_and_name = g_strdup_printf("blue_foreground");

      colour_content = g_strdup_printf("black_foreground");
    }
  else
    {
      colour_time_and_name = g_strdup_printf("red_foreground");
      colour_content = g_strdup_printf("red_foreground");
    }


#if 0
  /* get text direction */
  if(gj_gtk_cw_get_reverse_text_direction(cw) == TRUE)
    gtk_text_buffer_get_start_iter(buffer, &iter);
  else
    gtk_text_buffer_get_end_iter(buffer, &iter);
#else
    gtk_text_buffer_get_end_iter(buffer, &iter);
#endif

    if(gj_config_get_mpf_hide_time() == FALSE)
      {
	/* insert at iter */
	gtk_text_buffer_insert_with_tags_by_name(buffer, &iter,
						 time, -1,
						 colour_time_and_name,
						 NULL);
	gtk_text_buffer_insert_with_tags_by_name(buffer, &iter,
						 " ", -1,
						 colour_time_and_name,
						 NULL);
      }

    if(gj_message_get_from_resource(msg) != NULL)
      {
	gtk_text_buffer_insert_with_tags_by_name(buffer, &iter,
						 gj_message_get_from_resource(msg), -1,
						 "bold", 
						 colour_time_and_name,
						 NULL);
      }
  
  /* search for occurences of emote icons */
  for(i=0;i<g_utf8_strlen(message, -1);i++)
    {
      gchar *p = NULL;
      p = g_utf8_offset_to_pointer(message, i);

#if 0
      if(strncmp(p, EMOTE_BIGGRIN, g_utf8_strlen(EMOTE_BIGGRIN,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_biggrin)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_BIGGRIN,-1)-1);
	}
      else if(strncmp(p, EMOTE_CONFUSED, g_utf8_strlen(EMOTE_CONFUSED,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_confused)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_CONFUSED,-1)-1);
	}
      else if(strncmp(p, EMOTE_COOL, g_utf8_strlen(EMOTE_COOL,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_cool)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_COOL,-1)-1);
	}
      else if(strncmp(p, EMOTE_EEK, g_utf8_strlen(EMOTE_EEK,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_eek)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_EEK,-1)-1);
	}
      else if(strncmp(p, EMOTE_FROWN, g_utf8_strlen(EMOTE_FROWN,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_frown)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_FROWN,-1)-1);
	}
      else if(strncmp(p, EMOTE_MAD, g_utf8_strlen(EMOTE_MAD,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_mad)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_MAD,-1)-1);
	}
      else if(strncmp(p, EMOTE_REDFACE, g_utf8_strlen(EMOTE_REDFACE,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_redface)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_REDFACE,-1)-1);
	}
      else if(strncmp(p, EMOTE_ROLLEYES, g_utf8_strlen(EMOTE_ROLLEYES,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_rolleyes)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_ROLLEYES,-1)-1);
	}
      else if(strncmp(p, EMOTE_SMILE, g_utf8_strlen(EMOTE_SMILE,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_smile)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_SMILE,-1)-1);
	}
      else if(strncmp(p, EMOTE_TONGUE, g_utf8_strlen(EMOTE_TONGUE,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_tongue)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_TONGUE,-1)-1);
	}
      else if(strncmp(p, EMOTE_WINK, g_utf8_strlen(EMOTE_WINK,-1)) == 0)
	{
	  if((pb = gj_gtk_cw_emote_icon_get_as_pixbuf(emote_wink)) != NULL)
	    {
	      gtk_text_buffer_insert_pixbuf(buffer, &iter, pb);
	      g_object_unref(G_OBJECT(pb));
	    }
	  
	  i += (g_utf8_strlen(EMOTE_WINK,-1)-1);
	}
      else
#endif
	{
	  gchar *str = NULL;

	  /* create mem - utf should be no longer than 6 bytes */
	  str = g_new0(char, 10);
	  
	  /* copy char */
	  str = g_utf8_strncpy(str, p, 1);
	  
	  /* insert text */
	  gtk_text_buffer_insert_with_tags_by_name(buffer, &iter, 
						   str, -1,
						   colour_content,
						   NULL);

	  if(str != NULL) g_free(str);
	}
    }

  /* set subject */
  if(gj_message_get_subject(msg) != NULL)
    gtk_entry_set_text(GTK_ENTRY(gcw->controls.entry), gj_message_get_subject(msg));

#if 0
  /* log to history */
  gj_history_write_received_by_ri_id(gj_gtk_cw_get_ri_id(cw), 
				     (gchar*)gj_support_get_timestamp_nice_from_timet(gj_message_get_time(msg)), 
				     NULL, 
				     gj_message_get_body(msg));

  if(gj_gtk_cw_get_reverse_text_direction(cw) == FALSE)
    {
      mark = gtk_text_buffer_create_mark(buffer, NULL, &iter, FALSE);
      gtk_text_view_scroll_mark_onscreen(textview, mark);
      gtk_text_buffer_delete_mark(buffer, mark);
    }

#endif

  mark = gtk_text_buffer_create_mark(buffer, NULL, &iter, FALSE);
  gtk_text_view_scroll_mark_onscreen(textview, mark);
  gtk_text_buffer_delete_mark(buffer, mark);

  /* clean up */
  if(colour_time_and_name != NULL) g_free(colour_time_and_name);
  if(colour_content != NULL) g_free(colour_content);
  if(time != NULL) g_free(time);
  if(message != NULL) g_free(message);

  return TRUE;
}


/*
 * model functions
 */
gboolean gj_gtk_gcw_model_unload(GroupChatWindow gcw)
{
  GtkTreeView *view = NULL;
  GtkListStore *store = NULL;

  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  if((view = GTK_TREE_VIEW(gcw->controls.treeview)) == NULL)
    {
      g_warning("%s: view was NULL", __FUNCTION__);
      return FALSE;
    }

  if((store = GTK_LIST_STORE(gtk_tree_view_get_model(view))) == NULL)
    {
      g_warning("%s: store was NULL", __FUNCTION__);
      return FALSE;
    }

  /* clear roster */
  gtk_list_store_clear(store);
  
  return TRUE;
}

gboolean gj_gtk_gcw_model_setup(GroupChatWindow gcw)
{
  GtkTreeView *view = NULL;
  GtkTreeModel *model = NULL;
  GtkListStore *store = NULL;
  GtkTreeSelection *selection = NULL;

  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  if((view = GTK_TREE_VIEW(gcw->controls.treeview)) == NULL)
    {
      g_warning("%s: view was NULL", __FUNCTION__);
      return FALSE;
    }

  if((selection = GTK_TREE_SELECTION(gtk_tree_view_get_selection(view))) == NULL)
    {
      g_warning("%s: selection was NULL", __FUNCTION__);
      return FALSE;
    }

  /* store */
  store = gtk_list_store_new(COL_GROUP_CHAT_COUNT,
			     GDK_TYPE_PIXBUF,   /* presence */
			     G_TYPE_STRING,     /* jid */
			     G_TYPE_STRING,     /* name */
			     G_TYPE_BOOLEAN);   /* include */
			     
  /* model */
  model = GTK_TREE_MODEL(store);
  gtk_tree_view_set_model(view, model);

  /* selection */
  gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);

  gj_gtk_gcw_model_populate_columns(view);

  /* properties */
  gtk_tree_view_set_enable_search(view, TRUE);
  gtk_tree_view_set_search_column(view, COL_GROUP_CHAT_NAME); 
  gtk_tree_view_set_rules_hint(view, FALSE);
  gtk_tree_view_set_headers_visible(view, FALSE);
  gtk_tree_view_set_headers_clickable(view, FALSE);
  gtk_tree_view_columns_autosize(view);

  /* clean up */
  g_object_unref(G_OBJECT(store));

  return TRUE;
}

gboolean gj_gtk_gcw_model_populate_columns(GtkTreeView *view)
{
  GtkTreeViewColumn *column = NULL;
  GtkCellRenderer *renderer = NULL;

  if(view == NULL)
    {
      g_warning("%s: view was NULL", __FUNCTION__);
      return FALSE;
    }

  /* cell renderer's */
  column = gtk_tree_view_column_new();

  /* COL_ROSTER_IMAGE */  
  renderer = gtk_cell_renderer_pixbuf_new();
  gtk_tree_view_column_pack_start(column, renderer, FALSE);
  gtk_tree_view_column_set_attributes(column,
				      renderer,
				      "pixbuf", COL_GROUP_CHAT_IMAGE,
				      NULL);

  /* COL_ROSTER_TEXT */
  renderer = gtk_cell_renderer_text_new();
  gtk_tree_view_column_pack_start(column, renderer, TRUE);
  gtk_tree_view_column_set_attributes(column,
				      renderer,
				      "text", COL_GROUP_CHAT_NAME,
				      NULL);

  /* insert column */
  gtk_tree_view_append_column(GTK_TREE_VIEW(view), column);

  return TRUE;
}

gboolean gj_gtk_gcw_model_find_iter_by_jid(GroupChatWindow gcw, const gchar *jid, GtkTreeIter *ret_iter)
{
  GtkTreeView *view = NULL;
  GtkTreeModel *model = NULL;
  GtkTreeIter iter;
  
  gboolean valid = FALSE;
  gboolean found = FALSE;
  gchar *text = NULL;

  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  if((view = GTK_TREE_VIEW(gcw->controls.treeview)) == NULL)
    {
      g_warning("%s: view was NULL", __FUNCTION__);
      return FALSE;
    }
  
  if((model = GTK_TREE_MODEL(gtk_tree_view_get_model(view))) == NULL)
    {
      g_warning("%s: model was NULL", __FUNCTION__);
      return FALSE;
    }

  if(jid == NULL)
    {
      g_warning("%s: jid was NULL", __FUNCTION__);
      return FALSE;
    }

  if(ret_iter == NULL)
    {
      g_warning("%s: ret_iter pointer was NULL", __FUNCTION__);
      return FALSE;
    }

  /* Get the first iter in the list */
  valid = gtk_tree_model_get_iter_first(model, &iter);

  for(found = FALSE; valid == TRUE && found == FALSE; valid = gtk_tree_model_iter_next(model, &iter))
    {
      gtk_tree_model_get(model, &iter, 
			 COL_GROUP_CHAT_JID, &text,
			 -1);

      if(text != NULL && strcmp(jid, text) == 0)
	{
	  *ret_iter = iter;
	  found = TRUE;
	}

      if(text == NULL) g_free(text);
    }

  return found;
}


gboolean gj_gtk_gcw_model_add(GroupChatWindow gcw, presence pres)
{
  GtkTreeView *view = NULL;
  GtkListStore *store = NULL;
  GtkTreeIter iter;

  gchar *file = NULL;
  GdkPixbuf *pb = NULL;

  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  if(pres == NULL)
    {
      g_warning("%s: presence was NULL", __FUNCTION__);
      return FALSE;
    } 

  if((view = GTK_TREE_VIEW(gcw->controls.treeview)) == NULL)
    {
      g_warning("%s: view was NULL", __FUNCTION__);
      return FALSE;
    }
  
  if((store = GTK_LIST_STORE(gtk_tree_view_get_model(view))) == NULL)
    {
      g_warning("%s: model was NULL", __FUNCTION__);
      return FALSE;
    }

  if(gj_presence_get_type(pres) != presenceTypeAvailable)
    {
      switch(gj_presence_get_type(pres))
	{
	case presenceTypeUnAvailable:
	  file = FILE_PRESENCE_UNAVAILABLE;
	  break;
	case presenceTypeSubscribe:
	  file = FILE_PRESENCE_SUBSCRIBE;
	  break;
	case presenceTypeUnSubscribe:
	  file = FILE_PRESENCE_UNSUBSCRIBE;
	  break;
	case presenceTypeSubscribed:
	  file = FILE_PRESENCE_SUBSCRIBED;
	  break;
	case presenceTypeUnSubscribed:
	  file = FILE_PRESENCE_UNSUBSCRIBED;
	  break;
	case presenceTypeError:
	  file = FILE_PRESENCE_ERROR;
	  break;
	default:
	  g_warning("%s: failed to determin presence 'type'", __FUNCTION__);
	  return FALSE;
	}
    }
  else
    {
      switch(gj_presence_get_show(pres))
	{
	case presenceShowAway:
	  file = FILE_PRESENCE_AWAY;
	  break;
	case presenceShowChat:
	  file = FILE_PRESENCE_CHAT;
	  break;
	case presenceShowDND:
	  file = FILE_PRESENCE_DND;
	  break;
	case presenceShowNormal:
	  file = FILE_PRESENCE_NORMAL;
	  break;
	case presenceShowXA:
	  file = FILE_PRESENCE_XA;
	  break;
	default:
	      g_warning("%s: failed to determin presence 'show'", __FUNCTION__);
	      return FALSE;
	}
    }

  pb = gj_gtk_support_pixbuf_new_from_file(file);

  gtk_list_store_append(store, &iter);
  gtk_list_store_set(store, &iter, 
		     COL_GROUP_CHAT_IMAGE, pb,
		     COL_GROUP_CHAT_JID, gj_presence_get_from_jid(pres),
		     COL_GROUP_CHAT_NAME, gj_presence_get_from_resource(pres),
		     -1);

  /* clean up */
  if(pb != NULL) g_object_unref(G_OBJECT(pb));

  return TRUE;
}

gboolean gj_gtk_gcw_model_remove(GroupChatWindow gcw, presence pres)
{
  GtkTreeView *view = NULL;
  GtkListStore *store = NULL;
  GtkTreeIter iter;

  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  if(pres == NULL)
    {
      g_warning("%s: presence was NULL", __FUNCTION__);
      return FALSE;
    } 

  if((view = GTK_TREE_VIEW(gcw->controls.treeview)) == NULL)
    {
      g_warning("%s: view was NULL", __FUNCTION__);
      return FALSE;
    }
  
  if((store = GTK_LIST_STORE(gtk_tree_view_get_model(view))) == NULL)
    {
      g_warning("%s: model was NULL", __FUNCTION__);
      return FALSE;
    }

  if(gj_gtk_gcw_model_find_iter_by_jid(gcw, gj_presence_get_from_jid(pres), &iter) == FALSE)
    {
      g_warning("%s: user not found in model with jid:'%s'", __FUNCTION__, gj_presence_get_from_jid(pres));
      return TRUE;
    }

  gtk_list_store_remove(store, &iter);
  return TRUE;
}

gboolean gj_gtk_gcw_model_update(GroupChatWindow gcw, presence pres)
{
  GtkTreeView *view = NULL;
  GtkListStore *store = NULL;
  GtkTreeIter iter;

  gchar *file = NULL;
  GdkPixbuf *pb = NULL;

  if(gcw == NULL)
    {
      g_warning("%s: GroupChatWindow was NULL", __FUNCTION__);
      return FALSE;
    }

  if(pres == NULL)
    {
      g_warning("%s: presence was NULL", __FUNCTION__);
      return FALSE;
    } 

  if((view = GTK_TREE_VIEW(gcw->controls.treeview)) == NULL)
    {
      g_warning("%s: view was NULL", __FUNCTION__);
      return FALSE;
    }
  
  if((store = GTK_LIST_STORE(gtk_tree_view_get_model(view))) == NULL)
    {
      g_warning("%s: model was NULL", __FUNCTION__);
      return FALSE;
    }

  if(gj_gtk_gcw_model_find_iter_by_jid(gcw, gj_presence_get_from_jid(pres), &iter) == FALSE)
    {
      g_warning("%s: user not found in model with jid:'%s'", __FUNCTION__, gj_presence_get_from_jid(pres));
      return TRUE;
    }

  if(gj_presence_get_type(pres) != presenceTypeAvailable)
    {
      switch(gj_presence_get_type(pres))
	{
	case presenceTypeUnAvailable:
	  file = FILE_PRESENCE_UNAVAILABLE;
	  break;
	case presenceTypeSubscribe:
	  file = FILE_PRESENCE_SUBSCRIBE;
	  break;
	case presenceTypeUnSubscribe:
	  file = FILE_PRESENCE_UNSUBSCRIBE;
	  break;
	case presenceTypeSubscribed:
	  file = FILE_PRESENCE_SUBSCRIBED;
	  break;
	case presenceTypeUnSubscribed:
	  file = FILE_PRESENCE_UNSUBSCRIBED;
	  break;
	case presenceTypeError:
	  file = FILE_PRESENCE_ERROR;
	  break;
	default:
	  g_warning("%s: failed to determin presence 'type'", __FUNCTION__);
	  return FALSE;
	}
    }
  else
    {
      switch(gj_presence_get_show(pres))
	{
	case presenceShowAway:
	  file = FILE_PRESENCE_AWAY;
	  break;
	case presenceShowChat:
	  file = FILE_PRESENCE_CHAT;
	  break;
	case presenceShowDND:
	  file = FILE_PRESENCE_DND;
	  break;
	case presenceShowNormal:
	  file = FILE_PRESENCE_NORMAL;
	  break;
	case presenceShowXA:
	  file = FILE_PRESENCE_XA;
	  break;
	default:
	      g_warning("%s: failed to determin presence 'show'", __FUNCTION__);
	      return FALSE;
	}
    }

  pb = gj_gtk_support_pixbuf_new_from_file(file);
  
  gtk_list_store_set(store, &iter, 
		     COL_GROUP_CHAT_IMAGE, pb,
		     COL_GROUP_CHAT_JID, gj_presence_get_from_jid(pres),
		     COL_GROUP_CHAT_NAME, gj_presence_get_from_resource(pres),
		     -1);

  /* clean up */
  if(pb != NULL) g_object_unref(G_OBJECT(pb));

  return TRUE;
}

/*
 * GUI events
 */
void on_textbuffer_gcw_receive_changed(GtkTextBuffer *textbuffer, gpointer user_data)
{
  GtkTextMark *mark = NULL;
  GtkTextIter iter_space;
  GtkTextIter iter_word_start;
  GtkTextIter iter_word_end;
  GtkTextIter iter_start;
  GtkTextIter iter_end;

  gint length = 0;

  /* check WHOLE buffer for urls */
  gtk_text_buffer_get_bounds(textbuffer, &iter_start, &iter_end);
  length = gtk_text_buffer_get_char_count(textbuffer);

  if((mark = gtk_text_buffer_get_mark(textbuffer, "url")) != NULL)
    gtk_text_buffer_get_iter_at_mark(textbuffer, &iter_start, mark);

  /* save last iter */
  iter_word_start = iter_start;
  iter_word_end = iter_start;
  iter_space = iter_start;

  while(TRUE)
    {
      gboolean found = FALSE;
      gchar *buf = NULL;

      /* search for next space */
      found = gtk_text_iter_forward_find_char(&iter_space, on_textview_gcw_receive_find_space, NULL, NULL);

      if(found == FALSE)
	break;
      
      /* set word end to the space */
      iter_word_start = iter_word_end;
      iter_word_end = iter_space;

      /* move word end iter forward 1 */
      gtk_text_iter_forward_char(&iter_word_start);

      /* get */
      buf = gtk_text_iter_get_text(&iter_word_start, &iter_word_end);

      if(buf != NULL) 
	{
	  if(gj_support_is_uri(buf) != uriTypeUnknown) 
	    {
	      /* highlight as uri */
	      gtk_text_buffer_apply_tag_by_name(textbuffer, "url", &iter_word_start, &iter_word_end);
	    } 
	  else
	    {
	      /* none highlighted */
	      gtk_text_buffer_remove_tag_by_name(textbuffer, "url", &iter_word_start, &iter_word_end);
	    }

	  g_free(buf);
	}

      /* set mark, then we can check next time from 
	 the last mark and it is more efficient */
      if((mark = gtk_text_buffer_get_mark(textbuffer, "url")) == NULL)
	{
	  if((mark = gtk_text_buffer_create_mark(textbuffer, "url", &iter_space, FALSE)) == NULL)
	    {
	      g_warning("%s: could not create mark!", __FUNCTION__);
	      return;
	    }

	  /* no need to move it, just continue */
	  continue;
	}

      /* move mark to latest iter */
      gtk_text_buffer_move_mark_by_name(textbuffer, "url", &iter_space);
    }
}

gboolean on_textview_gcw_receive_motion_notify_event(GtkWidget *widget, GdkEventMotion *event, gpointer user_data)
{
  GdkModifierType state = 0;
  GdkWindow *window = NULL;

  GtkTextView *textview = NULL;
  GtkTextBuffer *buffer = NULL;
  GtkTextIter iter;
  GtkTextIter iter_start;
  GtkTextIter iter_end;
  GtkTextWindowType text_window_type = 0;

  gboolean found_start = FALSE;
  gboolean found_end = FALSE;

  gint pointer_x = 0;
  gint pointer_y = 0;
  gint x = 0;
  gint y = 0;

  gchar *buf = NULL;

  GroupChatWindow gcw = NULL;

  if(event->is_hint)
    {
      window = gdk_window_get_pointer(event->window, &pointer_x, &pointer_y, &state);
    }
  else
    {
      pointer_x = event->x;
      pointer_y = event->y;
      state = event->state;
    }

  if((gcw = gj_gtk_gcw_find_by_widget(GTK_WIDGET(widget))) == NULL)
    {
      g_warning("%s: ChatWindow was NULL fom widget:0x%.8x", __FUNCTION__, (gint) widget);
      return FALSE;
    }
  
  if((textview = GTK_TEXT_VIEW(widget)) == NULL)
    {
      g_warning("%s: textview was NULL", __FUNCTION__);
      return FALSE;
    }

  if((buffer = GTK_TEXT_BUFFER(gtk_text_view_get_buffer(textview))) == NULL)
    {
      g_warning("%s: buffer was NULL", __FUNCTION__);
      return FALSE;
    }

  text_window_type = gtk_text_view_get_window_type(textview, event->window); 
  gtk_text_view_window_to_buffer_coords(textview, text_window_type, pointer_x, pointer_y, &x, &y);
  gtk_text_view_get_iter_at_location(textview, &iter, x, y);

  if(g_unichar_isspace(gtk_text_iter_get_char(&iter)) == TRUE)
    {
      gdk_window_set_cursor(event->window, NULL);
      return FALSE;
    }

  iter_start = iter;
  iter_end = iter;

  /* search for next space */
  found_start = gtk_text_iter_backward_find_char(&iter_start, on_textview_gcw_receive_find_space, NULL, NULL);
  found_end = gtk_text_iter_forward_find_char(&iter_end, on_textview_gcw_receive_find_space, NULL, NULL);

  if(found_start == FALSE && found_end == FALSE)
    {
      gdk_window_set_cursor(event->window, NULL);
      return FALSE;
    }

  if(gtk_text_iter_is_start(&iter_start) == FALSE)
    gtk_text_iter_forward_char(&iter_start);
  
  if((buf = gtk_text_buffer_get_text(buffer, &iter_start, &iter_end, FALSE)) == NULL)
    {
      g_warning("%s: buffer was NULL", __FUNCTION__);
      return FALSE;
    }
  
  if(gj_support_is_uri(buf) != uriTypeUnknown)
    {
      /* width wise it works... */
      GdkCursor *cursor = NULL;
      cursor = gdk_cursor_new(GDK_HAND2);
      gdk_window_set_cursor (event->window, cursor);
      
      /* clean up */
      gdk_cursor_unref(cursor);
    }  
  else
    {
      /* return to parent window's cursor */
      gdk_window_set_cursor(event->window, NULL);
    }

      
  /* clean up */
  if(buf != NULL) g_free(buf);

  return FALSE;
}

gboolean on_textview_gcw_receive_find_space(gunichar ch, gpointer user_data)
{
  if(g_unichar_isspace(ch) == TRUE)
    return TRUE;

  return FALSE;
}

gboolean on_textview_gcw_receive_find_alpha_numeric(gunichar ch, gpointer user_data)
{
  if(g_unichar_isalnum(ch) == TRUE)
    return TRUE;

  return FALSE;
}

gboolean on_textview_gcw_receive_button_press_event(GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{
  GdkModifierType state;
  GtkTextBuffer *buffer = NULL;
  GtkTextIter iter;
  GtkTextIter iter_start;
  GtkTextIter iter_end;
  GtkTextWindowType text_window_type = 0;

  gboolean found_start = FALSE;
  gboolean found_end = FALSE;
    
  gint pointer_x = 0;
  gint pointer_y = 0;
  gint x = 0;
  gint y = 0;

  gchar *buf = NULL;

  /* if not mouse 1 return */
  if(event->button != 1)
    return FALSE;

  gdk_window_get_pointer(event->window, &pointer_x, &pointer_y, &state);
    
  if((buffer = GTK_TEXT_BUFFER(gtk_text_view_get_buffer(GTK_TEXT_VIEW(widget)))) == NULL)
    {
      g_warning("%s: buffer was NULL", __FUNCTION__);
      return FALSE;
    }
    
  text_window_type = gtk_text_view_get_window_type(GTK_TEXT_VIEW(widget), event->window); 
  gtk_text_view_window_to_buffer_coords(GTK_TEXT_VIEW(widget), text_window_type, pointer_x, pointer_y, &x, &y);
  gtk_text_view_get_iter_at_location(GTK_TEXT_VIEW(widget), &iter, x, y);
 
  iter_start = iter;
  iter_end = iter;
    
  /* search for next space */
  found_start = gtk_text_iter_backward_find_char(&iter_start, on_textview_gcw_receive_find_space, NULL, NULL);
  found_end = gtk_text_iter_forward_find_char(&iter_end, on_textview_gcw_receive_find_space, NULL, NULL);
    
  if(found_start == FALSE && found_end == FALSE)
    return FALSE;
    
  if(gtk_text_iter_is_start(&iter_start) == FALSE)
    gtk_text_iter_forward_char(&iter_start);
    
  if((buf = gtk_text_buffer_get_text(buffer, &iter_start, &iter_end, FALSE)) == NULL)
    {
      g_warning("%s: buffer was NULL", __FUNCTION__);
      return FALSE;
    }
  
  if(gj_support_is_uri(buf) != uriTypeUnknown)
    {
      gj_gtk_support_open_url(buf);
    }  
    
  /* clean up */
  if(buf != NULL) g_free(buf);
   
  return FALSE;
}

gboolean on_textview_gcw_send_key_release_event(GtkWidget *widget, GdkEventKey *event, gpointer user_data)
{
  GtkTextBuffer *buffer = NULL;
  GroupChatWindow gcw = NULL;

  if((buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(widget))) == NULL)
    {
      g_warning("%s: buffer was NULL", __FUNCTION__);
      return FALSE;
    }

  if((gcw = gj_gtk_gcw_find_by_widget(GTK_WIDGET(widget))) == NULL)
    {
      g_warning("%s: failed to find window based on widget:0x%.8x", __FUNCTION__, (gint)widget);
      return FALSE;
    }

  /* if was enter then clear box */
  if(event != NULL && event->keyval == GDK_Return && (event->state & GDK_SHIFT_MASK) == FALSE)
    {
      GtkTextIter iter_start;
      GtkTextIter iter_end; 

      gtk_text_buffer_get_bounds(GTK_TEXT_BUFFER(buffer), &iter_start, &iter_end);
      gtk_text_buffer_delete(GTK_TEXT_BUFFER(buffer), &iter_start, &iter_end);

/*       gtk_widget_set_sensitive(gcw->controls.button_send, FALSE); */
    }
  else
    {
      gint length = 0;

      /* set toolbar icon */  
      length = gtk_text_buffer_get_char_count(buffer);

/*       if(length < 1) */
/* 	gtk_widget_set_sensitive(cw->controls.button_send, FALSE); */
/*       else */
/* 	gtk_widget_set_sensitive(cw->controls.button_send, TRUE); */
    }

  return FALSE;
}

gboolean on_textview_gcw_send_key_press_event(GtkWidget *widget, GdkEventKey *event, gpointer user_data)
{
  GtkTextBuffer *buffer = NULL;
  GtkTextIter iter_start;
  GtkTextIter iter_end;  

  gchar *text = NULL;
  gint length = 0;

  GroupChatWindow gcw = NULL;
    
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(widget));
  length = gtk_text_buffer_get_char_count(buffer);

  if((gcw = gj_gtk_gcw_find_by_widget(GTK_WIDGET(widget))) == NULL)
    {
      g_warning("%s: failed to find window based on button widget", __FUNCTION__);
      return FALSE;
    }
  
  if(length < 1)
    {
      /* if enter was pressed... warn user that we can not do this */
      if(event != NULL && event->keyval == GDK_Return && (event->state & GDK_SHIFT_MASK) == FALSE)
	g_warning("%s: can't send NULL or 0 length message", __FUNCTION__);
      
      return FALSE;
    }

  gtk_text_buffer_get_bounds(GTK_TEXT_BUFFER(buffer), &iter_start, &iter_end);
  text = gtk_text_buffer_get_text(GTK_TEXT_BUFFER(buffer), &iter_start, &iter_end, TRUE);
   
  /* send message and clear text box */
  if(event->keyval == GDK_Return && (event->state & GDK_SHIFT_MASK) == FALSE)
    gj_gtk_gcw_send(gcw, text);
    
  /* cleanup */
  if(text != NULL) g_free(text);
    
  return FALSE;
}

gboolean on_group_chat_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data)
{
  GroupChatWindow gcw = NULL;

  if(widget == NULL)
    {
      g_warning("%s: widget was NULL", __FUNCTION__);
      return FALSE;
    }

  if((gcw = gj_gtk_gcw_find_by_widget(widget)) == NULL)
    {
      g_warning("%s: could not find GroupChatWindow from widget:0x%.8x", __FUNCTION__, widget);
      return FALSE;
    }
    
  gj_gtk_gcw_disconnect(gcw);
  return TRUE;
}

#ifdef __cplusplus
}
#endif
