/* 
 * $RCSfile: gj_gtk_group_chat_join.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/08/19 22:54:00 $
 * 
 * $Revision: 1.3 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <string.h>
#include <stdlib.h>

#include <gtk/gtk.h>
#include <glade/glade.h>

#include "common.h"

#include "gj_main.h"
#include "gj_roster.h"
#include "gj_iq_requests.h"
#include "gj_config.h"

#include "gj_gtk_group_chat_join.h"

struct 
{
  GtkWidget *window;

  GtkWidget *entry_nickname;
  GtkWidget *entry_host;
  GtkWidget *entry_room;

  GtkWidget *button_ok;

} ctlGCJ;


gboolean gj_gtk_gcj_join();
gboolean gj_gtk_gcj_check();


/*
 * gui functions 
 */
void on_entry_gcj_nickname_changed(GtkEditable *editable, gpointer user_data);
void on_entry_gcj_host_changed(GtkEditable *editable, gpointer user_data);
void on_entry_gcj_room_changed(GtkEditable *editable, gpointer user_data);

void on_group_chat_join_response(GtkDialog *dialog, gint response, gpointer user_data);


/*
 * window functions
 */

gboolean gj_gtk_gcj_load()
{
  GladeXML *xml = NULL;

  if(ctlGCJ.window != NULL)
    {
      gtk_window_present(GTK_WINDOW(ctlGCJ.window));
      return TRUE;
    }

  if((xml = glade_xml_new(gnome_jabber_get_glade_file(),
			  "group_chat_join",
			  NULL)) == NULL)
    {
      g_warning("gj_gtk_gcj_load: couldn't open glade file at root 'group_chat_join'");
      return FALSE;
    }
  
  ctlGCJ.window = glade_xml_get_widget(xml,"group_chat_join");

  ctlGCJ.entry_nickname = glade_xml_get_widget(xml,"entry_gcj_nickname");
  ctlGCJ.entry_host = glade_xml_get_widget(xml,"entry_gcj_host");
  ctlGCJ.entry_room = glade_xml_get_widget(xml,"entry_gcj_room");

  ctlGCJ.button_ok = glade_xml_get_widget(xml,"button_gcj_ok");
  
  gdk_window_set_decorations (GDK_WINDOW(ctlGCJ.window->window), 
			      GDK_DECOR_BORDER | GDK_DECOR_TITLE);

#ifndef WIN32
  glade_xml_signal_autoconnect(xml);
#else
  g_signal_connect ((gpointer) ctlGCJ.window, "response",
		    G_CALLBACK (on_group_chat_join_response),
		    NULL);
  g_signal_connect ((gpointer) ctlGCJ.entry_nickname, "changed",
		    G_CALLBACK (on_entry_gcj_nickname_changed),
		    NULL);
  g_signal_connect ((gpointer) ctlGCJ.entry_host, "changed",
		    G_CALLBACK (on_entry_gcj_host_changed),
		    NULL);
  g_signal_connect ((gpointer) ctlGCJ.entry_room, "changed",
		    G_CALLBACK (on_entry_gcj_room_changed),
		    NULL);
#endif

  /* set up window */
  gtk_entry_set_text(GTK_ENTRY(ctlGCJ.entry_nickname), gj_config_get_cs_nickname());
/*   gtk_entry_set_text(GTK_ENTRY(ctlGCJ.entry_host), "conference.jabber.org"); */
/*   gtk_entry_set_text(GTK_ENTRY(ctlGCJ.entry_room), "jdev"); */

  /* clean up */
  if(xml != NULL)
    {
      g_object_unref(G_OBJECT(xml));
      xml = NULL;
    }

  return TRUE;
}

gboolean gj_gtk_gcj_unload()
{
  if(ctlGCJ.window == NULL)
    return FALSE;

  gtk_widget_destroy(GTK_WIDGET(ctlGCJ.window));

  ctlGCJ.window = NULL;
  ctlGCJ.entry_nickname = NULL;
  ctlGCJ.entry_host = NULL;
  ctlGCJ.entry_room = NULL;
  ctlGCJ.button_ok = NULL;

  return TRUE;
}

gboolean gj_gtk_gcj_join()
{
  G_CONST_RETURN gchar *nickname = NULL;
  G_CONST_RETURN gchar *host = NULL;
  G_CONST_RETURN gchar *room = NULL;

  nickname = gtk_entry_get_text(GTK_ENTRY(ctlGCJ.entry_nickname));
  host = gtk_entry_get_text(GTK_ENTRY(ctlGCJ.entry_host));
  room = gtk_entry_get_text(GTK_ENTRY(ctlGCJ.entry_room));

  if(gj_gtk_gcj_check() == FALSE)
    return FALSE;

  return gj_gtk_gcw_join(host, room, nickname);
}

gboolean gj_gtk_gcj_check()
{
  G_CONST_RETURN gchar *nickname = NULL;
  G_CONST_RETURN gchar *host = NULL;
  G_CONST_RETURN gchar *room = NULL;

  nickname = gtk_entry_get_text(GTK_ENTRY(ctlGCJ.entry_nickname));
  host = gtk_entry_get_text(GTK_ENTRY(ctlGCJ.entry_host));
  room = gtk_entry_get_text(GTK_ENTRY(ctlGCJ.entry_room));

  if((nickname == NULL || g_utf8_strlen(nickname, -1) < 1) ||
     (host == NULL || g_utf8_strlen(host, -1) < 1) ||
     (room == NULL || g_utf8_strlen(room, -1) < 1))
    {
      /* disable ok */
      gtk_widget_set_sensitive(GTK_WIDGET(ctlGCJ.button_ok), FALSE);
      return FALSE;
    }

  gtk_widget_set_sensitive(GTK_WIDGET(ctlGCJ.button_ok), TRUE);
  return TRUE;
}


/*
 * GUI events
 */
void on_entry_gcj_nickname_changed(GtkEditable *editable, gpointer user_data)
{
  gj_gtk_gcj_check();
}

void on_entry_gcj_host_changed(GtkEditable *editable, gpointer user_data)
{
  gj_gtk_gcj_check();
}

void on_entry_gcj_room_changed(GtkEditable *editable, gpointer user_data)
{
  gj_gtk_gcj_check();
}

void on_group_chat_join_response(GtkDialog *dialog, gint response, gpointer user_data)
{
  switch(response) 
    {
    case GTK_RESPONSE_OK:
      gj_gtk_gcj_join();

    case GTK_RESPONSE_HELP:
    case GTK_RESPONSE_NONE:
    case GTK_RESPONSE_REJECT:
    case GTK_RESPONSE_ACCEPT:
    case GTK_RESPONSE_DELETE_EVENT:
    case GTK_RESPONSE_CANCEL:
    case GTK_RESPONSE_CLOSE:
    case GTK_RESPONSE_YES:
    case GTK_RESPONSE_NO:
    case GTK_RESPONSE_APPLY:
      break;
    }

  gj_gtk_gcj_unload();
}

#ifdef __cplusplus
}
#endif
