/* 
 * $RCSfile: gj_gtk_group_message.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/07/31 16:43:54 $
 * 
 * $Revision: 1.9 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <string.h>
#include <stdlib.h>

#include <gtk/gtk.h>
#include <glade/glade.h>

#include "common.h"

#include "gj_main.h"
#include "gj_roster.h"
#include "gj_message.h"

#include "gj_gtk_group_message.h"
#include "gj_gtk_msg.h"


/* columns */
enum
  {
    COL_GROUP_MESSAGE_JID,
    COL_GROUP_MESSAGE_NAME,
    COL_GROUP_MESSAGE_ENABLED,
    COL_GROUP_MESSAGE_COUNT
  };


struct 
{
  GtkWidget *window;

  GtkWidget *entry;
  GtkWidget *textview;
  GtkWidget *treeview;

  GtkWidget *button_send;
  GtkWidget *button_close;

} ctlGM;

struct
{
  gchar *group;

} optGM;


/*
 * member functions
 */
     
/* gets */
gchar *gj_gtk_gm_get_group();

/* sets */
gboolean gj_gtk_gm_set_group(const gchar *group);

/*
 * model functions
 */
gboolean gj_gtk_gm_model_setup();
gboolean gj_gtk_gm_model_populate_columns(GtkTreeView *view);
gboolean gj_gtk_gm_model_populate_from_roster();

gboolean gj_gtk_gm_model_add_contact(rosterItem ri);

/*
 *  window functions 
 */
gboolean gj_gtk_gm_setup(const gchar *group);
gboolean gj_gtk_gm_send();


/*
 * gui functions 
 */
void on_treeview_gm_cell_toggled(GtkCellRendererToggle *cell, gchar *path_string, gpointer data);

void on_textbuffer_gm_changed(GtkTextBuffer *textbuffer, gpointer user_data);

void on_button_gm_send_clicked(GtkButton *button, gpointer user_data);
void on_button_gm_close_clicked(GtkButton *button, gpointer user_data);

gboolean on_group_message_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data);

/*
 * member functions
 */

/* gets */
gchar *gj_gtk_gm_get_group()
{
  return optGM.group;
}

/* sets */
gboolean gj_gtk_gm_set_group(const gchar *group)
{
  if(group == NULL)
    {
      g_warning("gj_gtk_gm_set_group: group was NULL");
      return FALSE;
    }

  if(optGM.group != NULL)
    {
      g_free(optGM.group);
      optGM.group = NULL;
    }

  optGM.group = g_strdup(group);
  return TRUE;
}

/*
 * window functions
 */
gboolean gj_gtk_gm_load(const gchar *group)
{
  GladeXML *xml = NULL;

  if(ctlGM.window != NULL)
    {
      gtk_window_present(GTK_WINDOW(ctlGM.window));
      return TRUE;
    }

  if(group == NULL || g_utf8_strlen(group, -1) < 1)
    {
      g_warning("gj_gtk_gm_load: group was NULL");
      return FALSE;
    }

  if((xml = glade_xml_new(gnome_jabber_get_glade_file(),
			  "group_message",
			  NULL)) == NULL)
    {
      g_warning("gj_gtk_gm_load: couldn't open glade file at root 'group_message'");
      return FALSE;
    }
  
  ctlGM.window = glade_xml_get_widget(xml,"group_message");

  ctlGM.entry = glade_xml_get_widget(xml,"entry_gm");
  ctlGM.textview = glade_xml_get_widget(xml,"textview_gm");
  ctlGM.treeview = glade_xml_get_widget(xml,"treeview_gm");

  ctlGM.button_send = glade_xml_get_widget(xml,"button_gm_send");
  ctlGM.button_close = glade_xml_get_widget(xml,"button_gm_close");
  
  gdk_window_set_decorations (GDK_WINDOW(ctlGM.window->window), 
			      GDK_DECOR_BORDER | GDK_DECOR_TITLE | GDK_DECOR_MENU);

#ifndef WIN32
  glade_xml_signal_autoconnect(xml);
#else
  g_signal_connect ((gpointer) ctlGM.window, "delete_event",
		    G_CALLBACK (on_group_message_delete_event),
		    NULL);
  g_signal_connect ((gpointer) ctlGM.button_send, "clicked",
		    G_CALLBACK (on_button_gm_send_clicked),
		    NULL);
  g_signal_connect ((gpointer) ctlGM.button_close, "clicked",
		    G_CALLBACK (on_button_gm_close_clicked),
		    NULL);
#endif

  gj_gtk_gm_setup(group);

  /* clean up */
  if(xml != NULL)
    {
      g_object_unref(G_OBJECT(xml));
      xml = NULL;
    }

  return TRUE;
}

gboolean gj_gtk_gm_unload()
{
  if(ctlGM.window == NULL)
    return FALSE;

  gtk_widget_destroy(GTK_WIDGET(ctlGM.window));

  ctlGM.window = NULL;
  ctlGM.entry = NULL;
  ctlGM.textview = NULL;
  ctlGM.treeview = NULL;
  ctlGM.button_send = NULL;
  ctlGM.button_close = NULL;

  return TRUE;
}

gboolean gj_gtk_gm_setup(const gchar *group)
{
  GtkTextBuffer *buffer = NULL;

  gchar *text = NULL;

  if(optGM.group != NULL)
    {
      g_free(optGM.group);
      optGM.group = NULL;
    }

  optGM.group = g_strdup(group);

  /* set title */
  text = g_strdup_printf("%s %s", _("Sending Message to Group"), group);
  gtk_window_set_title(GTK_WINDOW(ctlGM.window), text);
  g_free(text);

  /* set buffer changed callback */
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(ctlGM.textview));

  g_signal_connect(G_OBJECT(buffer), "changed", 
		   G_CALLBACK(on_textbuffer_gm_changed), 
		   NULL);

  /* set up model */
  gj_gtk_gm_model_setup();
  gj_gtk_gm_model_populate_from_roster();

  return TRUE;
}

gboolean gj_gtk_gm_model_load(GList *roster)
{
  /* populate model */
  if(gj_gtk_gm_model_populate_from_roster(roster) == FALSE)
    {
      g_warning("gj_gtk_gm_model_load: failed to setup roster");
      return FALSE;
    }

  return TRUE;
}

gboolean gj_gtk_gm_model_unload()
{
  GtkTreeView *view = NULL;
  GtkListStore *store = NULL;

  if((view = GTK_TREE_VIEW(ctlGM.treeview)) == NULL)
    {
      g_warning("gj_gtk_gm_model_unload: view was NULL");
      return FALSE;
    }

  if((store = GTK_LIST_STORE(gtk_tree_view_get_model(view))) == NULL)
    {
      g_warning("gj_gtk_gm_model_unload: store was NULL");
      return FALSE;
    }

  /* clear roster */
  gtk_list_store_clear(store);
  
  return TRUE;
}

gboolean gj_gtk_gm_model_setup()
{
  GtkTreeView *view = NULL;
  GtkTreeModel *model = NULL;
  GtkListStore *store = NULL;
  GtkTreeSelection *selection = NULL;

  if((view = GTK_TREE_VIEW(ctlGM.treeview)) == NULL)
    {
      g_warning("gj_gtk_gm_model_setup: view was NULL");
      return FALSE;
    }

  if((selection = GTK_TREE_SELECTION(gtk_tree_view_get_selection(view))) == NULL)
    {
      g_warning("gj_gtk_gm_model_setup: selection was NULL");
      return FALSE;
    }

  /* store */
  store = gtk_list_store_new(COL_GROUP_MESSAGE_COUNT,
			     G_TYPE_STRING,   /* name */
			     G_TYPE_STRING,   /* jid */
			     G_TYPE_BOOLEAN); /* include */

			     
  /* model */
  model = GTK_TREE_MODEL(store);
  gtk_tree_view_set_model(view, model);

  /* selection */
  gtk_tree_selection_set_mode(selection, GTK_SELECTION_SINGLE);

  gj_gtk_gm_model_populate_columns(view);

  /* properties */
  gtk_tree_view_set_enable_search(view, TRUE);
  gtk_tree_view_set_search_column(view, COL_GROUP_MESSAGE_NAME); 
  gtk_tree_view_set_rules_hint(view, FALSE);
  gtk_tree_view_set_headers_visible(view, FALSE);
  gtk_tree_view_set_headers_clickable(view, FALSE);
  gtk_tree_view_columns_autosize(view);

  /* clean up */
  g_object_unref(G_OBJECT(store));

  return TRUE;
}

gboolean gj_gtk_gm_model_populate_columns(GtkTreeView *view)
{
  GtkTreeViewColumn *column = NULL;
  GtkCellRenderer *renderer = NULL;
  gint col_offset = 0;

  if(view == NULL)
    {
      g_warning("gj_gtk_gm_model_populate_columns: view was NULL");
      return FALSE;
    }

  /* COL_ROSTER_GROUP_ENABLED */
  renderer = gtk_cell_renderer_toggle_new();
  g_signal_connect(G_OBJECT(renderer), "toggled", G_CALLBACK(on_treeview_gm_cell_toggled), NULL);

  column = gtk_tree_view_column_new_with_attributes("",
						    renderer,
						    "active", COL_GROUP_MESSAGE_ENABLED,
						    NULL);

  /* set this column to a fixed sizing (of 50 pixels) */
/*   gtk_tree_view_column_set_sizing(GTK_TREE_VIEW_COLUMN(column), GTK_TREE_VIEW_COLUMN_FIXED); */
/*   gtk_tree_view_column_set_fixed_width(GTK_TREE_VIEW_COLUMN(column), 50); */
  gtk_tree_view_column_set_sort_column_id(column, COL_GROUP_MESSAGE_ENABLED);
  gtk_tree_view_append_column(view, column);

  /* COL_ROSTER_GROUP_NAME */
  renderer = gtk_cell_renderer_text_new();
  col_offset = gtk_tree_view_insert_column_with_attributes(view,
							   -1, "Group",
							   renderer, 
							   "text", COL_GROUP_MESSAGE_NAME,
							   NULL);
  
  /* set extra information */
  g_object_set_data(G_OBJECT(renderer), "column", (gint *)COL_GROUP_MESSAGE_NAME);

  column = gtk_tree_view_get_column(view, col_offset - 1);
  gtk_tree_view_column_set_sort_column_id(column, COL_GROUP_MESSAGE_NAME);

  return TRUE;
}

gboolean gj_gtk_gm_model_populate_from_roster()
{
  gint index = 0;
  gint count = 0;
  gchar *group = NULL;
  GList *list = NULL;
  
  group = gj_gtk_gm_get_group();
  list = gj_roster_get_all_items_by_group(group);
  
  for(index=0,count=0;index<g_list_length(list);index++)
    {
      gchar *jid = NULL;
      rosterItem ri = NULL;
      presence pres = NULL;

      jid = (gchar*) g_list_nth_data(list, index);
      if(jid == NULL) continue;

      ri = gj_roster_item_find_by_jid(jid);
      if(ri == NULL) continue;

      pres = gj_roster_item_get_presence(ri);
      if(pres == NULL || gj_presence_get_type(pres) != presenceTypeAvailable) continue; 

      gj_gtk_gm_model_add_contact(ri);
      count++;

      g_message("gj_gtk_gm_model_populate_from_roster: index:%2.2d using jid:'%s'",index, gj_roster_item_get_jid(ri)); 
    }

  g_message("gj_gtk_gm_model_populate_from_roster: %d contacts all together",count);
  return TRUE;
}

gboolean gj_gtk_gm_model_add_contact(rosterItem ri)
{
  GtkTreeView *view = NULL;
  GtkListStore *store = NULL;
  GtkTreeIter iter;

  gchar *name = NULL;

  if((view = GTK_TREE_VIEW(ctlGM.treeview)) == NULL)
    {
      g_warning("gj_gtk_gm_model_add_contact: view was NULL");
      return FALSE;
    }

  if((store = GTK_LIST_STORE(gtk_tree_view_get_model(view))) == NULL)
    {
      g_warning("gj_gtk_gm_model_add_contact: store was NULL");
      return FALSE;
    }

  if(ri == NULL)
    {
      g_warning("gj_gtk_gm_model_add_contact: rosterItem was NULL"); 
      return FALSE;
    }

  /* check details */
  if((name = gj_roster_item_get_name(ri)) == NULL)
    {
      name = gj_roster_item_get_jid(ri);
      g_message("gj_gtk_gm_model_add_contact: roster item had no name, using jid:'%s', as name", gj_roster_item_get_jid(ri));
    }

  /* add to tree */
  gtk_list_store_append(store, &iter);
  gtk_list_store_set(store, &iter,
		     COL_GROUP_MESSAGE_NAME, name,
		     COL_GROUP_MESSAGE_JID, gj_roster_item_get_jid(ri),
		     COL_GROUP_MESSAGE_ENABLED, TRUE,
		     -1);
  

  return TRUE;
}

gboolean gj_gtk_gm_send()
{
  GtkTreeView *treeview = NULL;
  GtkTreeModel *model = NULL;
  GtkTreeIter iter;
  gboolean valid = FALSE;

  gchar *name = NULL;
  gchar *jid = NULL;
  gboolean enabled = FALSE;

  GtkTextView *textview = NULL;
  GtkTextBuffer *buffer = NULL;
  GtkTextIter iter_start, iter_end;

  G_CONST_RETURN gchar *subject = NULL;
  gchar *body = NULL;

  if((treeview = GTK_TREE_VIEW(ctlGM.treeview)) == NULL)
    {
      g_warning("gj_gtk_gm_send: treeview was NULL");
      return FALSE;
    }

  if((model = GTK_TREE_MODEL(gtk_tree_view_get_model(treeview))) == NULL)
    {
      g_warning("gj_gtk_gm_send: model was NULL");
      return FALSE;
    } 

  if((textview = GTK_TEXT_VIEW(ctlGM.textview)) == NULL)
    {
      g_warning("gj_gtk_gm_send: textview was NULL");
      return FALSE;
    }

  if((buffer = GTK_TEXT_BUFFER(gtk_text_view_get_buffer(textview))) == NULL)
    {
      g_warning("gj_gtk_gm_send: buffer was NULL");
      return FALSE;
    }

  gtk_text_buffer_get_bounds(buffer, &iter_start, &iter_end);
  body = gtk_text_buffer_get_text(buffer, &iter_start, &iter_end, FALSE);
  subject = gtk_entry_get_text(GTK_ENTRY(ctlGM.entry));

  if(body == NULL || g_utf8_strlen(body, -1) < 1)
    {
      if(body != NULL) g_free(body);
      g_warning("gj_gtk_gm_send: body was NULL or length was < 1");
      return FALSE;
    }
  
  /* Get the first iter in the list */
  valid = gtk_tree_model_get_iter_first(model, &iter);

  for(;valid == TRUE; valid = gtk_tree_model_iter_next(model, &iter))
    {
      gtk_tree_model_get(model, &iter, 
			 COL_GROUP_MESSAGE_JID, &jid,
			 COL_GROUP_MESSAGE_NAME, &name,
			 COL_GROUP_MESSAGE_ENABLED, &enabled,
			 -1);
		  
      if(jid == NULL || enabled == FALSE) 
	{
	  if(jid != NULL) g_free(jid);
	  if(name != NULL) g_free(name);
	  continue;
	}
      
      gj_message_send(messageTypeNormal, jid, (gchar*)subject, body, NULL);
    }

  return TRUE;
}

/*
 * GUI events
 */
void on_treeview_gm_cell_toggled(GtkCellRendererToggle *cell, gchar *path_string, gpointer data)
{
  GtkTreeView *view = NULL;
  GtkTreeModel *model = NULL;
  GtkListStore *store = NULL;
  GtkTreePath *path = NULL;
  GtkTreeIter iter;

  gboolean enabled = FALSE;

  if((view = GTK_TREE_VIEW(ctlGM.treeview)) == NULL)
    {
      g_warning("gj_gtk_gm_cell_toggled: view was NULL");
      return;
    }
  
  if((model = GTK_TREE_MODEL(gtk_tree_view_get_model(view))) == NULL)
    {
      g_warning("gj_gtk_gm_cell_toggled: model was NULL");
      return;
    }

  if((store = GTK_LIST_STORE(gtk_tree_view_get_model(view))) == NULL)
    {
      g_warning("gj_gtk_gm_cell_toggled: store was NULL");
      return;
    }

  path = gtk_tree_path_new_from_string(path_string);

  /* get toggled iter */
  gtk_tree_model_get_iter(model, &iter, path);
  gtk_tree_model_get(model, &iter, 
		     COL_GROUP_MESSAGE_ENABLED, &enabled, 
		     -1);

  /* do something */
  enabled ^= 1;

  /* set new value */
  gtk_list_store_set(store, &iter, 
		     COL_GROUP_MESSAGE_ENABLED, enabled, 
		     -1);

  /* clean up */
  gtk_tree_path_free(path);

  g_message("on_treeview_gm_cell_toggled: item is now %s", enabled?"enabled":"disabled");
}

void on_textbuffer_gm_changed(GtkTextBuffer *textbuffer, gpointer user_data)
{
  GtkTextIter iter_start;
  GtkTextIter iter_end;

  gint length = 0;

  /* check WHOLE buffer for urls */
  gtk_text_buffer_get_bounds(textbuffer, &iter_start, &iter_end);
  length = gtk_text_buffer_get_char_count(textbuffer);

  if(length > 0)
    gtk_widget_set_sensitive(GTK_WIDGET(ctlGM.button_send), TRUE);
  else
    gtk_widget_set_sensitive(GTK_WIDGET(ctlGM.button_send), FALSE);
}

void on_button_gm_send_clicked(GtkButton *button, gpointer user_data)
{
  gj_gtk_gm_send();
  gj_gtk_gm_unload();
}

void on_button_gm_close_clicked(GtkButton *button, gpointer user_data)
{
  gj_gtk_gm_unload();
}

gboolean on_group_message_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data)
{
  gj_gtk_gm_unload();

  return TRUE;
}

#ifdef __cplusplus
}
#endif
