/* 
 * $RCSfile: gj_gtk_headline.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/08/23 11:17:26 $
 * 
 * $Revision: 1.18 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include <glib.h>
#include <gtk/gtk.h>
#include <glade/glade.h>

#include "common.h"

#include "gj_main.h"
#include "gj_roster.h"
#include "gj_presence.h"
#include "gj_message.h"
#include "gj_translate.h"
#include "gj_event.h"
#include "gj_support.h"
#include "gj_config.h"

#include "gj_gtk_roster.h"
#include "gj_gtk_headline.h" 
#include "gj_gtk_my_preferences.h"
#include "gj_gtk_support.h"

struct t_controls
{
  GtkWidget *window;

  GtkWidget *textview;

  GtkWidget *button_clear;
  GtkWidget *button_close;

};

struct t_HeadlineWindow
{
  gint ri_id;

  /*   gint x; */
  /*   gint y; */
  /*   gint w; */
  /*   gint h; */

  struct t_controls controls;
};


GList *headline_window_list = NULL;


gboolean gj_gtk_hw_load(HeadlineWindow hw);
gboolean gj_gtk_hw_unload(HeadlineWindow hw);
gboolean gj_gtk_hw_unload_by_widget(GtkWidget *widget);

gboolean gj_gtk_hw_hide(HeadlineWindow hw);
gboolean gj_gtk_hw_hide_by_widget(GtkWidget *widget);

void gj_gtk_hw_raise();

/* list functions */
gboolean gj_gtk_hw_list_add(HeadlineWindow hw);
gboolean gj_gtk_hw_list_del(HeadlineWindow hw);
GList *gj_gtk_hw_list_get();

/* window functions */
HeadlineWindow gj_gtk_hw_new(rosterItem ri);
gboolean gj_gtk_hw_free(HeadlineWindow hw);
void gj_gtk_hw_destroy(HeadlineWindow hw);

gboolean gj_gtk_hw_setup(HeadlineWindow hw);

gboolean gj_gtk_hw_add(HeadlineWindow hw, message msg);
gboolean gj_gtk_hw_clear(HeadlineWindow hw);

/* gui callbacks */
void on_textbuffer_hw_changed(GtkTextBuffer *textbuffer, gpointer user_data);
gboolean on_textview_hw_motion_notify_event(GtkWidget *widget, GdkEventMotion *event, gpointer user_data);
gboolean on_textview_hw_find_space(gunichar ch, gpointer user_data);
gboolean on_textview_hw_button_press_event(GtkWidget *widget, GdkEventButton *event, gpointer user_data);

void on_button_hw_clear_clicked(GtkButton *button, gpointer user_data);
void on_button_hw_close_clicked(GtkButton *button, gpointer user_data);

gboolean on_headline_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data);

/*
 * code
 */

/* list handling */
gboolean gj_gtk_hw_list_add(HeadlineWindow hw)
{
  if(hw == NULL)
    {
      g_warning("%s: headline window was NULL", __FUNCTION__);
      return FALSE;
    }

  headline_window_list = g_list_append(headline_window_list,hw);
  return TRUE;
}

gboolean gj_gtk_hw_list_del(HeadlineWindow hw)
{
  if(hw == NULL)
    {
      g_warning("%s: headline window was NULL", __FUNCTION__);
      return FALSE;
    }

  headline_window_list = g_list_remove(headline_window_list,hw);
  return TRUE;
}

GList *gj_gtk_hw_list_get()
{
  return headline_window_list;
}

HeadlineWindow gj_gtk_hw_new(rosterItem ri)
{
  HeadlineWindow hw = NULL;

  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return NULL;
    }

  if((hw = (HeadlineWindow) g_new(struct t_HeadlineWindow,1)) != NULL)
    {
      hw->ri_id = gj_roster_item_get_id(ri);

      hw->controls.window = NULL;
      hw->controls.textview = NULL;
      hw->controls.button_clear = NULL;
      hw->controls.button_close = NULL;

      gj_gtk_hw_list_add(hw);
    }

  return hw;
}

gboolean gj_gtk_hw_free(HeadlineWindow hw)
{
  if(hw == NULL)
    {
      g_warning("%s: headline window was NULL", __FUNCTION__);
      return FALSE;
    }

  /* free from list */
  gj_gtk_hw_list_del(hw);
  
  /* destroy members */
  gj_gtk_hw_destroy(hw);

  /* free struct */
  g_free(hw);
  
  return TRUE;
}

void gj_gtk_hw_destroy(HeadlineWindow hw)
{
  if(hw == NULL)
    {
      g_warning("%s: headline window was NULL", __FUNCTION__);
      return;
    }

  hw->ri_id = -1;

  hw->controls.window = NULL;
  hw->controls.textview = NULL;
  hw->controls.button_clear = NULL;
  hw->controls.button_close = NULL;
}

HeadlineWindow gj_gtk_hw_find_by_ri_id(gint id)
{
  gint index = 0;

  if(id < 0)
    {
      g_warning("%s: id:%d was < 0", __FUNCTION__, id);
      return NULL;
    }

  for(index=0;index<g_list_length(headline_window_list);index++)
    {
      HeadlineWindow hw = g_list_nth_data(headline_window_list, index);

      if(hw == NULL) continue;
      if(hw->ri_id == id) return hw;
    }

  return NULL;  
}

HeadlineWindow gj_gtk_hw_find_by_widget(GtkWidget *widget)
{
  gint index = 0;
  GtkWidget *toplevel = NULL;

  if(widget == NULL)
    {
      g_warning("%s: widget was NULL", __FUNCTION__);
      return NULL;
    }

  /* get top level widget */
  toplevel = gtk_widget_get_toplevel(widget);

  if(toplevel == NULL)
    {
      g_warning("%s: top level widget was NULL", __FUNCTION__);
      return NULL;
    }

  for(index=0;index<g_list_length(headline_window_list);index++)
    {
      HeadlineWindow hw = g_list_nth_data(headline_window_list, index);

      if(hw == NULL) continue;
      if(hw->controls.window == toplevel) return hw;
    }

  return NULL;  
}

/*
 * member functions 
 */

/* sets */
gboolean gj_gtk_hw_set_ri_id(HeadlineWindow hw, gint id)
{
  if(hw == NULL)
    {
      g_warning("%s: headline window was NULL", __FUNCTION__);
      return FALSE;
    }

  if(id < 0)
    {
      g_warning("%s: id:%d was < 0", __FUNCTION__, id);
      return FALSE;
    }

  hw->ri_id = id;

  return TRUE;
}

/* gets */
gint gj_gtk_hw_get_ri_id(HeadlineWindow hw)
{
  if(hw == NULL)
    {
      g_warning("%s: headline window was NULL", __FUNCTION__);
      return -1;
    }

  return hw->ri_id;
}

/*
 * window functions
 */
gboolean gj_gtk_hw_show(HeadlineWindow hw, gboolean show)
{
  if(hw == NULL)
    {
      g_warning("%s: headline window was NULL", __FUNCTION__);
      return FALSE;
    }
  
  if(show == TRUE)
    gtk_widget_show(hw->controls.window);
  else
    gtk_widget_hide(hw->controls.window);

  return TRUE;
}

void gj_gtk_hw_raise(HeadlineWindow hw)
{
  if(hw == NULL)
    {
      g_warning("%s: headline window was NULL", __FUNCTION__);
      return;
    }
  
  if(hw->controls.window == NULL)
    {
      g_warning("%s: window was NULL", __FUNCTION__);
      return;
    }
  
  gtk_window_present(GTK_WINDOW(hw->controls.window));
}

gboolean gj_gtk_hw_load(HeadlineWindow hw)
{
  GladeXML *xml = NULL;
  GdkPixbuf *pb = NULL;
  
  if(hw == NULL)
    {
      g_warning("%s: headline window was NULL", __FUNCTION__);
      return FALSE;
    }

  if(hw->controls.window != NULL)
    {
      gtk_window_present(GTK_WINDOW(hw->controls.window));
      return TRUE;
    }

  if((xml = glade_xml_new(gnome_jabber_get_glade_file(),
			  "headline",
			  NULL)) == NULL)
    {
      g_warning("%s: couldn't open glade file at root 'headline'", __FUNCTION__);
      return FALSE;
    }
  
  hw->controls.window = glade_xml_get_widget(xml,"headline");

  hw->controls.textview = glade_xml_get_widget(xml, "textview_hw");

  hw->controls.button_clear = glade_xml_get_widget(xml, "button_hw_clear");
  hw->controls.button_close = glade_xml_get_widget(xml, "button_hw_close");
  
  gdk_window_set_decorations (GDK_WINDOW(hw->controls.window->window), 
			      GDK_DECOR_BORDER | GDK_DECOR_TITLE | GDK_DECOR_MENU);

#ifndef WIN32
  glade_xml_signal_autoconnect(xml);
#else
  gtk_widget_set_events(GTK_WIDGET(hw->controls.textview),
			GDK_EXPOSURE_MASK | 
			GDK_POINTER_MOTION_MASK | 
			GDK_POINTER_MOTION_HINT_MASK | 
			GDK_BUTTON_RELEASE_MASK | 
			GDK_STRUCTURE_MASK | 
			GDK_PROPERTY_CHANGE_MASK);

  g_signal_connect ((gpointer) hw->controls.window, "delete_event",
		    G_CALLBACK (on_headline_delete_event),
		    NULL);
  g_signal_connect ((gpointer) hw->controls.textview, "button_press_event",
		    G_CALLBACK (on_textview_hw_button_press_event),
		    NULL);
  g_signal_connect ((gpointer) hw->controls.textview, "motion_notify_event",
		    G_CALLBACK (on_textview_hw_motion_notify_event),
		    NULL);
  g_signal_connect ((gpointer) hw->controls.button_clear, "clicked",
		    G_CALLBACK (on_button_hw_clear_clicked),
		    NULL);
  g_signal_connect ((gpointer) hw->controls.button_close, "clicked",
		    G_CALLBACK (on_button_hw_close_clicked),
		    NULL);
#endif

  /* set icon */
  pb = gj_gtk_support_pixbuf_new_from_file(FILE_MESSAGE_HEADLINE);
  gtk_window_set_icon(GTK_WINDOW(hw->controls.window), pb);
  g_object_unref(G_OBJECT(pb));

  /* set up model */
  gj_gtk_hw_setup(hw);

  /* 
   * cleanup 
   */
  if(xml != NULL)
    {
      g_object_unref(G_OBJECT(xml));
      xml = NULL;
    }

  return TRUE;
}

gboolean gj_gtk_hw_load_by_ri(rosterItem ri)
{
  HeadlineWindow hw = NULL;

  if(ri == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }

  if((hw = gj_gtk_hw_find_by_ri_id(gj_roster_item_get_id(ri))) == NULL)
    {
      g_message("%s: could not get roster item by id:%d (jid:'%s')",
		__FUNCTION__, gj_roster_item_get_id(ri), gj_roster_item_get_jid(ri));
      return FALSE;     
    }

  gj_gtk_hw_load(hw);
  return TRUE;    
}

gboolean gj_gtk_hw_unload(HeadlineWindow hw)
{
  if(hw == NULL)
    {
      g_warning("%s: headline window was NULL", __FUNCTION__);
      return FALSE;
    }

  /* if no more msgs reset icon */
  if(hw->controls.window != NULL)
    {
      gtk_widget_destroy(GTK_WIDGET(hw->controls.window));
      hw->controls.window = NULL;
    }

  gj_gtk_hw_free(hw);
      
  g_message("%s: unloaded...", __FUNCTION__);
  
  return TRUE;
}

gboolean gj_gtk_hw_unload_by_widget(GtkWidget *widget)
{
  HeadlineWindow hw = NULL;

  if(widget == NULL)
    {
      g_warning("%s: widget was NULL", __FUNCTION__);
      return FALSE;
    }

  if((hw = gj_gtk_hw_find_by_widget(widget)) == NULL)
    {
      g_warning("%s: could not find hw from widget", __FUNCTION__);
      return FALSE;
    }

  gj_gtk_hw_unload(hw);
  return TRUE;
}

gboolean gj_gtk_hw_hide(HeadlineWindow hw)
{
  if(hw == NULL)
    {
      g_warning("%s: headline window was NULL", __FUNCTION__);
      return FALSE;
    }


  gtk_widget_hide(GTK_WIDGET(hw->controls.window));
  g_message("%s: hidden...", __FUNCTION__);
  
  return TRUE;
}

gboolean gj_gtk_hw_hide_by_widget(GtkWidget *widget)
{
  HeadlineWindow hw = NULL;

  if(widget == NULL)
    {
      g_warning("%s: widget was NULL", __FUNCTION__);
      return FALSE;
    }

  if((hw = gj_gtk_hw_find_by_widget(widget)) == NULL)
    {
      g_warning("%s: could not find headline window from widget", __FUNCTION__);
      return FALSE;
    }

  gj_gtk_hw_hide(hw);
  return TRUE;
}

gboolean gj_gtk_hw_setup(HeadlineWindow hw)
{
  GtkTextBuffer *buffer = NULL;
  gchar *title = NULL;
  
  rosterItem ri = NULL;

  if(hw == NULL)
    {
      g_warning("%s: headline window was NULL", __FUNCTION__);
      return FALSE;
    }

  if((ri = gj_roster_item_find_by_id(gj_gtk_hw_get_ri_id(hw))) == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }

  if((buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(hw->controls.textview))) == NULL)
    {
      g_warning("%s: buffer was NULL", __FUNCTION__);
      return FALSE;
    }
  
  /* set function callbacks */
  g_signal_connect(G_OBJECT(buffer), "changed", 
		   G_CALLBACK(on_textbuffer_hw_changed), 
		   NULL);

  /* set styles */
  gtk_text_buffer_create_tag(buffer, "bold", 
			     "weight", PANGO_WEIGHT_BOLD, 
			     NULL);
  gtk_text_buffer_create_tag(buffer, "italic", 
			     "style", PANGO_STYLE_ITALIC, 
			     NULL);  
  gtk_text_buffer_create_tag(buffer, "blue_foreground", 
			     "foreground", "darkblue", 
			     NULL);  
  gtk_text_buffer_create_tag(buffer, "red_foreground", 
			     "foreground", "darkred", 
			     NULL);  
  gtk_text_buffer_create_tag(buffer, "black_foreground", 
			     "foreground", "black", 
			     NULL); 
  gtk_text_buffer_create_tag(buffer, "url", 
			     "foreground", "blue", 
			     NULL); 
  gtk_text_buffer_create_tag(buffer, "url_desc", 
			     /* 			     "weight", PANGO_WEIGHT_BOLD,  */
			     "underline", PANGO_UNDERLINE_SINGLE,
			     NULL); 
  gtk_text_buffer_create_tag(buffer, "left_margin_small",
			     "left_margin", 5,
			     NULL);
  gtk_text_buffer_create_tag(buffer, "left_margin_large",
			     "left_margin", 30,
			     NULL);
  gtk_text_buffer_create_tag(buffer, "not_editable",
			     "editable", FALSE, 
			     NULL);
  gtk_text_buffer_create_tag(buffer, "underline",
			     "underline", PANGO_UNDERLINE_SINGLE, 
			     NULL);
  gtk_text_buffer_create_tag(buffer, "tiny_gap_after_line",
			     "pixels_below_lines", 5, 
			     NULL);
  gtk_text_buffer_create_tag(buffer, "small_gap_after_line",
			     "pixels_below_lines", 15, 
			     NULL);
  gtk_text_buffer_create_tag(buffer, "big_gap_after_line",
			     "pixels_below_lines", 30, 
			     NULL);
  gtk_text_buffer_create_tag(buffer, "big_gap_before_line",
			     "pixels_above_lines", 30, 
			     NULL);
  gtk_text_buffer_create_tag(buffer, "large",
			     "scale", PANGO_SCALE_LARGE, 
			     NULL);
  
  /* set title */
  title = g_strdup_printf("%s [%s]", _("Headlines for"), gj_roster_item_get_jid(ri));
  gtk_window_set_title(GTK_WINDOW(hw->controls.window), title);

  /* clean up */
  if(title != NULL) g_free(title);

  return TRUE;
}

gboolean gj_gtk_hw_add(HeadlineWindow hw, message msg)
{
  GtkTextView *textview = NULL;
  GtkTextBuffer *buffer = NULL;
  GtkTextIter iter;

  const gchar *time = NULL;
  const gchar *message = NULL;
    
  time = gj_support_get_time_and_date_from_timet(gj_message_get_time(msg));
  message = gj_message_get_body(msg);

  if(hw == NULL)
    {
      g_warning("%s: headline window was NULL", __FUNCTION__);
      return FALSE;
    }

  if((textview = GTK_TEXT_VIEW(hw->controls.textview)) == NULL)
    {
      g_warning("%s: textview was NULL", __FUNCTION__);
      return FALSE;
    }      
    
  if((buffer = gtk_text_view_get_buffer(textview)) == NULL)
    {
      g_warning("%s: buffer was NULL", __FUNCTION__);
      return FALSE;
    }    

  if(time == NULL)
    {
      g_warning("%s: time was NULL", __FUNCTION__);
      return FALSE;
    }

  if(message == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return FALSE;
    }

  gtk_text_buffer_get_end_iter(buffer, &iter);

  gtk_text_buffer_insert_with_tags_by_name(buffer, &iter, 
					   time, -1,
					   "bold",
					   "black_foreground",
					   "left_margin_small",
					   "tiny_gap_after_line",
					   NULL);

  gtk_text_buffer_insert_with_tags_by_name(buffer, &iter, 
					   "\n", -1,
					   "black_foreground",
					   NULL);

  gtk_text_buffer_insert_with_tags_by_name(buffer, &iter, 
					   gj_message_get_body(msg), -1,
					   "red_foreground",
					   "left_margin_large",
					   NULL);

  if(gj_message_get_x_oob_url(msg) != NULL && gj_message_get_x_oob_description(msg) != NULL)
    {
      /* show at end of msg */
      gtk_text_buffer_remove_all_tags(buffer, &iter, &iter);
      gtk_text_buffer_insert_with_tags_by_name(buffer, &iter, 
					       "\n", -1,
					       "black_foreground",
					       "left_margin_large",
					       NULL);

      /* show url */      
      gtk_text_buffer_insert_with_tags_by_name(buffer, &iter, 
					       gj_message_get_x_oob_description(msg), -1,
					       "left_margin_large",
					       "url_desc",
					       NULL);

      gtk_text_buffer_remove_all_tags(buffer, &iter, &iter);
      gtk_text_buffer_insert_with_tags_by_name(buffer, &iter, 
					       "\n", -1,
					       "black_foreground",
					       NULL);

      gtk_text_buffer_insert_with_tags_by_name(buffer, &iter, 
					       gj_message_get_x_oob_url(msg), -1,
 					       "left_margin_large",
					       "url",
					       NULL);
    }

  gtk_text_buffer_remove_all_tags(buffer, &iter, &iter);
  gtk_text_buffer_insert_with_tags_by_name(buffer, &iter, 
					   "\n\n\n", -1,
					   "black_foreground",
					   NULL);

  return TRUE;
}

gboolean gj_gtk_hw_clear(HeadlineWindow hw)
{
  GtkTextBuffer *buffer = NULL;
  GtkTextIter iter_start;
  GtkTextIter iter_end;

  if(hw == NULL)
    {
      g_warning("%s: headline window was NULL", __FUNCTION__);
      return FALSE;
    }

  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(hw->controls.textview));
  gtk_text_buffer_get_bounds(GTK_TEXT_BUFFER(buffer), &iter_start, &iter_end);
  
  /* clear buffer */
  gtk_text_buffer_delete(GTK_TEXT_BUFFER(buffer), &iter_start, &iter_end);
  gtk_text_buffer_delete_mark_by_name(GTK_TEXT_BUFFER(buffer), "url");

  return TRUE;
}

HeadlineWindow gj_gtk_hw_notification_new(gint event_id, message msg)
{
  HeadlineWindow hw = NULL;
  event ev = NULL;
  rosterItem ri = NULL;

  if(msg == NULL)
    {
      g_warning("%s: message was NULL", __FUNCTION__);
      return NULL;
    }

  if((ev = gj_event_find_by_id(event_id)) == NULL)
    {
      g_warning("%s: event was NULL", __FUNCTION__);
      return NULL;
    }

  if((ri = gj_roster_item_find_by_id(gj_event_get_ri_id(ev))) == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return NULL;
    }

  if((hw = gj_gtk_hw_find_by_ri_id(gj_roster_item_get_id(ri))) == NULL)
    {
      if((hw = gj_gtk_hw_new(ri)) == NULL)
	{
	  g_warning("%s: could not create new hw from jid:'%s'", __FUNCTION__, gj_roster_item_get_jid(ri));
	  return NULL;
	}

      gj_gtk_hw_load(hw);
    }
  else
    {
      /* should we show it? */
      if(gj_config_get_mpf_raise_window_on_new_message() == TRUE)
	gj_gtk_hw_raise(hw);
    }

  gj_gtk_hw_add(hw, msg);

  return hw;
}

/*
 * GUI events
 */
void on_textbuffer_hw_changed(GtkTextBuffer *textbuffer, gpointer user_data)
{
  GtkTextMark *mark = NULL;
  GtkTextIter iter_space;
  GtkTextIter iter_word_start;
  GtkTextIter iter_word_end;
  GtkTextIter iter_start;
  GtkTextIter iter_end;
  GtkTextTagTable *tag_table = NULL;
  GtkTextTag *tag_url = NULL;

  gint length = 0;

  /* check WHOLE buffer for urls */
  gtk_text_buffer_get_bounds(textbuffer, &iter_start, &iter_end);
  length = gtk_text_buffer_get_char_count(textbuffer);

  if((mark = gtk_text_buffer_get_mark(textbuffer, "url")) != NULL)
    {
      gtk_text_buffer_get_iter_at_mark(textbuffer, &iter_start, mark);
      /*       gtk_text_mark_set_visible(mark, TRUE);  */
    }

  if((tag_table = gtk_text_buffer_get_tag_table(textbuffer)) == NULL)
    {
      g_warning("%s: could not get text buffer tag table", __FUNCTION__);
      return;
    }

  if((tag_url = gtk_text_tag_table_lookup(tag_table, "url")) == NULL)
    {
      g_warning("%s: could not find tag:'url'", __FUNCTION__);
      return;
    }

  /* save last iter */
  iter_word_start = iter_start;
  iter_word_end = iter_start;
  iter_space = iter_start;

  while(TRUE)
    {
      gboolean found = FALSE;
      gchar *buf = NULL;

      /* search for next space */
      found = gtk_text_iter_forward_find_char(&iter_space, on_textview_hw_find_space, NULL, NULL);

      if(found == FALSE)
	break;
      
      /* set word end to the space */
      iter_word_start = iter_word_end;
      iter_word_end = iter_space;

      /* move word end iter forward 1 */
      gtk_text_iter_forward_char(&iter_word_start);

      if(gtk_text_iter_has_tag(&iter_word_start, tag_url) == TRUE) 
	continue;

      /* get */
      buf = gtk_text_iter_get_text(&iter_word_start, &iter_word_end);

      if(buf != NULL) 
	{
	  if(gj_support_is_uri(buf) != uriTypeUnknown) 
	    {
	      /* highlight as uri */
	      gtk_text_buffer_apply_tag_by_name(textbuffer, "url", &iter_word_start, &iter_word_end);
	    } 
	  else
	    {
	      /* none highlighted */
	      gtk_text_buffer_remove_tag_by_name(textbuffer, "url", &iter_word_start, &iter_word_end);
	    }

	  g_free(buf);
	}

      /* set mark, then we can check next time from 
	 the last mark and it is more efficient */
      if((mark = gtk_text_buffer_get_mark(textbuffer, "url")) == NULL)
	{
	  if((mark = gtk_text_buffer_create_mark(textbuffer, "url", &iter_space, FALSE)) == NULL)
	    {
	      g_warning("%s: could not create mark!", __FUNCTION__);
	      return;
	    }

	  /* no need to move it, just continue */
	  continue;
	}

      /* move mark to latest iter */
      gtk_text_buffer_move_mark_by_name(textbuffer, "url", &iter_space);
    }
}

gboolean on_textview_hw_motion_notify_event(GtkWidget *widget, GdkEventMotion *event, gpointer user_data)
{
  GdkModifierType state = 0;
  GdkWindow *window = NULL;

  GtkTextView *textview = NULL;
  GtkTextBuffer *buffer = NULL;
  GtkTextIter iter, iter_start, iter_end;
  GtkTextWindowType text_window_type = 0;

  gboolean found_start = FALSE;
  gboolean found_end = FALSE;

  gint pointer_x = 0;
  gint pointer_y = 0;
  gint x = 0;
  gint y = 0;

  gchar *buf = NULL;

  HeadlineWindow hw = NULL;

  if(event->is_hint)
    window = gdk_window_get_pointer(event->window, &pointer_x, &pointer_y, &state);
  else
    {
      pointer_x = event->x;
      pointer_y = event->y;
      state = event->state;
    }

  if((hw = gj_gtk_hw_find_by_widget(GTK_WIDGET(widget))) == NULL)
    {
      g_warning("%s: could not find headline window from widget", __FUNCTION__);
      return FALSE;
    }
  
  if((textview = GTK_TEXT_VIEW(widget)) == NULL)
    {
      g_warning("%s: textview was NULL", __FUNCTION__);
      return FALSE;
    }

  if((buffer = GTK_TEXT_BUFFER(gtk_text_view_get_buffer(textview))) == NULL)
    {
      g_warning("%s: buffer was NULL", __FUNCTION__);
      return FALSE;
    }

  text_window_type = gtk_text_view_get_window_type(textview, event->window);
  gtk_text_view_window_to_buffer_coords(textview, text_window_type, pointer_x, pointer_y, &x, &y);
  gtk_text_view_get_iter_at_location(textview, &iter, x, y);
 
  if(g_unichar_isspace(gtk_text_iter_get_char(&iter)) == TRUE)
    {
      gdk_window_set_cursor(event->window, NULL);
      return FALSE;
    }

  iter_start = iter;
  iter_end = iter;
  
  /* search for next space */
  found_start = gtk_text_iter_backward_find_char(&iter_start, on_textview_hw_find_space, NULL, NULL);
  found_end = gtk_text_iter_forward_find_char(&iter_end, on_textview_hw_find_space, NULL, NULL);
  
  if(found_start == FALSE && found_end == FALSE)
    {
      gdk_window_set_cursor(event->window, NULL);
      return FALSE;
    }
  
  if(gtk_text_iter_is_start(&iter_start) == FALSE)
    gtk_text_iter_forward_char(&iter_start);
  
  if((buf = gtk_text_buffer_get_text(buffer, &iter_start, &iter_end, FALSE)) == NULL)
    {
      g_warning("%s: buffer was NULL", __FUNCTION__);
      return FALSE;
    }
  
  if(gj_support_is_uri(buf) != uriTypeUnknown)
    {
      /* width wise it works... */
      GdkCursor *cursor = NULL;
      cursor = gdk_cursor_new(GDK_HAND2);
      gdk_window_set_cursor (event->window, cursor);
      
      /* clean up */
      gdk_cursor_unref(cursor);
    }  
  else
    {
      /* return to parent window's cursor */
      gdk_window_set_cursor(event->window, NULL);
    }

      
  /* clean up */
  if(buf != NULL) g_free(buf);

  return FALSE;
}

gboolean on_textview_hw_find_space(gunichar ch, gpointer user_data)
{
  if(g_unichar_isspace(ch) == TRUE)
    return TRUE;

  return FALSE;
}

gboolean on_textview_hw_button_press_event(GtkWidget *widget, GdkEventButton *event, gpointer user_data)
{
  GdkModifierType state;
  GtkTextBuffer *buffer = NULL;
  GtkTextIter iter, iter_start, iter_end;
  GtkTextWindowType text_window_type = 0;

  gboolean found_start = FALSE;
  gboolean found_end = FALSE;

  gint pointer_x = 0;
  gint pointer_y = 0;
  gint x = 0;
  gint y = 0;

  gchar *buf = NULL;
    
  /* if not mouse 1 return */
  if(event->button != 1)
    return FALSE;

  gdk_window_get_pointer(event->window, &pointer_x, &pointer_y, &state);
    
  if((buffer = GTK_TEXT_BUFFER(gtk_text_view_get_buffer(GTK_TEXT_VIEW(widget)))) == NULL)
    {
      g_warning("%s: could not get buffer.", __FUNCTION__);
      return FALSE;
    }
    
  text_window_type = gtk_text_view_get_window_type(GTK_TEXT_VIEW(widget), event->window);
  gtk_text_view_window_to_buffer_coords(GTK_TEXT_VIEW(widget), text_window_type, pointer_x, pointer_y, &x, &y);
  gtk_text_view_get_iter_at_location(GTK_TEXT_VIEW(widget), &iter, x, y);
 
  iter_start = iter;
  iter_end = iter;
    
  /* search for next space */
  found_start = gtk_text_iter_backward_find_char(&iter_start, on_textview_hw_find_space, NULL, NULL);
  found_end = gtk_text_iter_forward_find_char(&iter_end, on_textview_hw_find_space, NULL, NULL);
    
  if(found_start == FALSE && found_end == FALSE)
    return FALSE;
    
  if(gtk_text_iter_is_start(&iter_start) == FALSE)
    gtk_text_iter_forward_char(&iter_start);
    
  if((buf = gtk_text_buffer_get_text(buffer, &iter_start, &iter_end, FALSE)) == NULL)
    {
      g_warning("%s: buffer was NULL", __FUNCTION__);
      return FALSE;
    }
  
  if(gj_support_is_uri(buf) != uriTypeUnknown)
    {
      gj_gtk_support_open_url(buf);
    }  
    
  /* clean up */
  if(buf != NULL) g_free(buf);
   
  return FALSE;
}

void on_button_hw_clear_clicked(GtkButton *button, gpointer user_data)
{
  HeadlineWindow hw = NULL;
  
  if((hw = gj_gtk_hw_find_by_widget(GTK_WIDGET(button))) == NULL)
    {
      g_warning("%s: could not find headline window from button widget", __FUNCTION__);
      return;
    }
 
  gj_gtk_hw_clear(hw);
}

void on_button_hw_close_clicked(GtkButton *button, gpointer user_data)
{
  gj_gtk_hw_hide_by_widget(GTK_WIDGET(button));
}

gboolean on_headline_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data)
{
  gj_gtk_hw_hide_by_widget(GTK_WIDGET(widget));

  return TRUE;
}


#ifdef __cplusplus
}
#endif
