/* 
 * $RCSfile: gj_history.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/07/28 14:46:35 $
 * 
 * $Revision: 1.13 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <string.h>
#include <time.h> 

#include <glib.h>

#include "common.h"

#include "gj_history.h"
#include "gj_support.h"
#include "gj_roster.h"
#include "gj_config.h"
#include "gj_main.h"

/* structs */
struct t_history
{
  gint ri_id;
  
  gchar *filename;
  GIOChannel *channel;
};

gboolean gj_history_list_add(history hist);
gboolean gj_history_list_del(history hist);

void gj_history_destroy(history hist);

gchar *gj_history_check_filename(const gchar *original_filename);

gchar *history_path = NULL;
GList *history_list = NULL;

/*
 * code
 */
gboolean gj_history_init()
{
  gchar *saved_path = NULL;
  saved_path = gj_config_get_mpf_history_location();
  
  if(saved_path == NULL || g_utf8_strlen(saved_path, -1) < 1)
    {
      gchar *path = NULL;

      path = g_build_path(G_DIR_SEPARATOR_S, gnome_jabber_get_config_dir(), PATH_HISTORY, NULL);
      gj_config_set_mpf_history_location(path);

      history_path = g_strdup(path);

      g_free(path);
    }
  else
    {
      /* use saved path */
      history_path = g_strdup(saved_path);
    }

  return TRUE;
}

gboolean gj_history_term()
{
  gint index = 0;
    
  /* list */
  if(history_list == NULL)
    return TRUE;
    
  /* free all items on the list */
  for(index=0;index<g_list_length(history_list);index++)
    {
      history hist = NULL;
	
      if((hist = (history) g_list_nth_data(history_list,index)) != NULL)
	gj_history_free(hist);
    }
    
  g_list_free(history_list);
  history_list = NULL;

  if(history_path != NULL)
    {
      g_free(history_path);
      history_path = NULL;
    }
    
  return TRUE;
}

gboolean gj_history_list_add(history hist)
{
  if(hist == NULL)
    {
      g_warning("%s: history was NULL", __FUNCTION__);
      return FALSE;
    }

  history_list = g_list_append(history_list,hist);

  return TRUE;
}

gboolean gj_history_list_del(history hist)
{
  if(hist == NULL)
    {
      g_warning("%s: history was NULL", __FUNCTION__);
      return FALSE;
    }

  history_list = g_list_remove(history_list,hist);
  return TRUE;
}

history gj_history_new(rosterItem ri)
{
  history hist = NULL;

  if(ri == NULL) 
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return NULL;
    }

  if((hist = g_new(struct t_history,1)))
    {
      hist->ri_id = gj_roster_item_get_id(ri);

      hist->filename = NULL;
      hist->channel = NULL;
    }

  history_list = g_list_append(history_list, hist);

  return hist;
}

gboolean gj_history_free(history hist)
{
  if(hist == NULL)
    {
      g_warning("%s: history was NULL", __FUNCTION__);
      return FALSE;
    }

  /* free mem */
  gj_history_list_del(hist);  

  /* destroy members */
  gj_history_destroy(hist);

  /* free structure */
  g_free(hist);

  return TRUE;
}

void gj_history_destroy(history hist)
{
  if(hist == NULL)
    {
      g_warning("%s: history was NULL", __FUNCTION__);
      return;
    }

  hist->ri_id = -1;

  if(hist->filename != NULL)
    {
      g_free(hist->filename);
      hist->filename = NULL;
    }

  if(hist->channel != NULL)
    {
      g_io_channel_shutdown(hist->channel,TRUE,NULL);
      hist->channel = NULL;
    }
}

/*
 * utils
 */
history gj_history_find_by_ri_id(gint id)
{
  gint index = 0;

  if(id < 0)
    {
      g_warning("%s: id:%d was < 0", __FUNCTION__, id);
      return NULL;
    }

  /* loop through all history_list on contacts history list to look for unique historyID */
  for(index=0;index<g_list_length(history_list);index++)
    {
      history hist = g_list_nth_data(history_list, index);
      
      if(hist == NULL) continue;
      if(hist->ri_id == id) return hist;
    }
 
  return NULL;
}

/*
 * member data
 */ 
gint gj_history_get_ri_id(history hist)
{
  if(hist == NULL)
    {
      g_warning("%s: history was NULL", __FUNCTION__);
      return -1;
    }

  return hist->ri_id;
}

gboolean gj_history_set_filename(history hist, gchar *filename)
{
  if(hist == NULL)
    {
      g_warning("%s: history was NULL", __FUNCTION__);
      return FALSE;
    }

  if(filename == NULL)
    {
      g_warning("%s: roster item was NULL", __FUNCTION__);
      return FALSE;
    }

  if(hist->filename != NULL)
    {
      g_free(hist->filename);
      hist->filename = NULL;
    }

  hist->filename = g_strdup(filename);

  return TRUE;
}

gchar *gj_history_get_filename(history hist)
{
  if(hist == NULL)
    {
      g_warning("%s: history was NULL", __FUNCTION__);
      return NULL;
    }

  return hist->filename;
}

GIOChannel *gj_history_get_channel(history hist)
{
  if(hist == NULL)
    {
      g_warning("%s: history was NULL", __FUNCTION__);
      return NULL;
    }

  return hist->channel;
}

/*
 * debugging 
 */
void gj_history_print_all()
{
  gint index = 0;

  g_message("******* Roster Item History List Start ********");  

  for(index=0;index<g_list_length(history_list);index++)
    {
      history hist = NULL;
      hist = g_list_nth_data(history_list,index);

      if(hist == NULL) continue;
      gj_history_print(hist);
    }

  g_message("******* Roster Item History List End **********");  

}

void gj_history_print(history hist)
{
  if(hist == NULL)
    return;

  g_message("******* Roster Item History ********");
  g_message("*******");
  g_message("******* roster item id:%d", gj_history_get_ri_id(hist));
  g_message("******* filename:'%s'", gj_history_get_filename(hist));
  g_message("*******");
  g_message("******* history: 0x%.8x", (gint)gj_history_get_channel(hist));
  g_message("*******");
  g_message("******* Roster Item History ********");
}

gchar *gj_history_check_filename(const gchar *original_filename)
{
  gchar *s1 = NULL;
  gchar *s2 = NULL;
  gchar *s3 = NULL;
  gchar *filename = NULL;
  
  s1 = g_strdup(original_filename);
  
  s2 = gj_support_find_in_str_and_replace("@", s1, "-");
  s3 = gj_support_find_in_str_and_replace("%", s2, "-"); 

  filename = g_strdup(s3);

  /* clean up */
  if(s1 != NULL) g_free(s1);
  if(s2 != NULL) g_free(s2);
  if(s3 != NULL) g_free(s3);

  return filename;
}

/*
 * HISTORY FUNCTIONS 
 *
 * - WIDELY USED BY OUTSIDE THIS MODULE
 *
 */
gboolean gj_history_start(history hist)
{
  guint bytes = 0;
  gboolean new_file = FALSE;
  gchar *app_start = NULL;

  rosterItem ri = NULL;

  if(hist == NULL)
    {
      g_warning("%s: history was NULL", __FUNCTION__);
      return FALSE;
    }

  if((ri = gj_roster_item_find_by_id(gj_history_get_ri_id(hist))) == NULL)
  {
      g_warning("%s: could not find by id:%d", __FUNCTION__, gj_history_get_ri_id(hist));
      return FALSE;
  }

  if(gj_history_get_filename(hist) == NULL)
    {
      gchar *filename = NULL;
      gchar *path = NULL;
      gchar *filename_unchecked = NULL;
      gchar *filename_checked = NULL;

      filename_unchecked = g_strdup_printf("%s%s%s.html", 
					   gj_roster_item_get_agent(ri)?gj_roster_item_get_agent(ri):"", 
					   gj_roster_item_get_agent(ri)?" - ":"",
					   gj_roster_item_get_name(ri));

      filename_checked = gj_history_check_filename(filename_unchecked);

      /* set path */
      path = gj_config_get_mpf_history_location();
      g_message("%s: using history path:'%s'", __FUNCTION__, path);
      
      filename = g_build_filename(path, G_DIR_SEPARATOR_S, filename_checked, NULL);
      g_message("%s: for jid:'%s', using filename:'%s'", __FUNCTION__, gj_roster_item_get_jid(ri), filename);
      
      if(filename != NULL)
	{
	  gj_history_set_filename(hist, filename);
	  g_free(filename);
	}
      else
	g_warning("%s: could not create string filename for history file, jid:'%s'",
		  __FUNCTION__, gj_roster_item_get_jid(ri));

      /* clean up */
      if(filename_unchecked != NULL) g_free(filename_unchecked);
      if(filename_checked != NULL) g_free(filename_checked);
    }

  /* does the file exist */
  if(g_file_test(gj_history_get_filename(hist), G_FILE_TEST_EXISTS) == FALSE)
  {
      g_message("%s: filename:'%s' does not exist, creating...", __FUNCTION__, gj_history_get_filename(hist));
      new_file = TRUE;
  }

  /* this opens a file or creates it */
  hist->channel = g_io_channel_new_file(gj_history_get_filename(hist),
					"a+",
					NULL);

  /* if this is a new file, add the header information */
  if(new_file == TRUE)
  {
    gchar *first_line = NULL;
    first_line = g_strdup_printf("<html><head><meta http-equiv=\"Content-Type\" content=\"text/html; charset=UTF-8\" /><link rel=\"stylesheet\" type=\"text/css\" href=\"%s\"/></head></html>\n",
				 FILE_FULL_CSS);

    g_io_channel_write_chars(gj_history_get_channel(hist),
			     first_line,
			     -1,
			     &bytes,
			     NULL); 

    if(first_line != NULL) g_free(first_line);
  }

  /* write the fact this is a start up procedure */
  app_start = g_strdup_printf("<html><body><div class=\"message\"><hr><span class=\"date\">%s</span> <span class=\"title\">Application Started</span><br/></div></body></html>\n", gj_support_get_timestamp_nice());

  g_io_channel_write_chars(gj_history_get_channel(hist),
			   app_start,
			   -1,
			   &bytes,
			   NULL); 

  if(app_start != NULL) g_free(app_start);

  /* flush to make sure tailing is up to date */
  g_io_channel_flush(gj_history_get_channel(hist), NULL);

  /* check it was created */
  if(hist->channel == NULL)
    {
      g_warning("%s: channel failed to append file:'%s'", __FUNCTION__, gj_history_get_filename(hist));
      return FALSE;
    }

  return TRUE;
}

gboolean gj_history_write_sent_by_ri_id(gint ri_id, gchar *date, gchar *subject, gchar *message)
{
  history hist = NULL;

  if(ri_id < 0)
    {
      g_warning("%s: roster item id:%d was < 0", __FUNCTION__, ri_id);
      return FALSE;
    }

  if((hist = gj_history_find_by_ri_id(ri_id)) == NULL)
    {
      g_warning("%s: could not find by roster item id:%d", __FUNCTION__, ri_id);
      return FALSE;
    }  

  return gj_history_write_sent(hist, date, subject, message);
}

gboolean gj_history_write_sent(history hist, gchar *date, gchar *subject, gchar *message)
{
  guint bytes = 0;
  gchar *formatted = NULL;

  if(gj_config_get_mpf_use_history() == FALSE)
      return TRUE;

  if(hist == NULL)
    {
      g_warning("%s: jid was NULL", __FUNCTION__);
      return FALSE;
    }

  if(gj_history_get_channel(hist) == NULL)
    {
      g_warning("%s: channel not yet initalised for ri_id:%d", __FUNCTION__, gj_history_get_ri_id(hist));
      return FALSE;
    }

  formatted = g_strdup_printf("<html><body><div class=\"message\"><span class=\"date\">%s</span> <span class=\"body\">%s</span> <span class=\"from\">%s<br/></span></div></body></html>\n", date, "[sent]", message);

  /* write to file */
  if(g_utf8_validate(formatted, -1, NULL) == FALSE)
    {
      fprintf(stderr, "[CRITICAL ERROR] !!!! INVALID UTF8 - WRITING HISTORY\n");
      fflush(stderr);
      return FALSE;
    }

  g_io_channel_write_chars(gj_history_get_channel(hist),
			   formatted,
			   -1,
			   &bytes,
			   NULL);  

  g_io_channel_flush(gj_history_get_channel(hist), NULL);

  g_message("%s: wrote %d bytes of %d characters to history (channel:0x%.8x)", 
	    __FUNCTION__, (gint)bytes, (gint)g_utf8_strlen(formatted, -1), (gint)gj_history_get_channel(hist)); 

  /* clean up */
  if(formatted != NULL) g_free(formatted);

  return TRUE;
}

gboolean gj_history_write_received_by_ri_id(gint ri_id, gchar *date, gchar *subject, gchar *message)
{
  history hist = NULL;

  if(ri_id < 0)
    {
      g_warning("%s: roster item id:%d was < 0", __FUNCTION__, ri_id);
      return FALSE;
    }

  if((hist = gj_history_find_by_ri_id(ri_id)) == NULL)
    {
      g_warning("%s: could not find by roster item id:%d", __FUNCTION__, ri_id);
      return FALSE;
    }  

  return gj_history_write_received(hist, date, subject, message);
}

gboolean gj_history_write_received(history hist, gchar *date, gchar *subject, gchar *message)
{
  guint bytes = 0;
  gchar *formatted = NULL;

  if(gj_config_get_mpf_use_history() == FALSE)
      return TRUE;

  if(hist == NULL)
    {
      g_warning("%s: history was NULL", __FUNCTION__);
      return FALSE;
    }

  if(gj_history_get_channel(hist) == NULL)
    {
      g_warning("%s: channel not yet initalised for roster item id:%d", __FUNCTION__, gj_history_get_ri_id(hist));
      return FALSE;
    }

  formatted = g_strdup_printf("<html><body><div class=\"message\"><span class=\"date\">%s</span> <span class=\"body\">%s</span> <span class=\"to\">%s<br/></span></div></body></html>\n", date, "[recv]", message);

  g_io_channel_write_chars(gj_history_get_channel(hist),
			   formatted,
			   -1,
			   &bytes,
			   NULL);  

  g_io_channel_flush(gj_history_get_channel(hist), NULL);

  g_message("%s: wrote %d bytes of %d characters to history (channel:0x%.8x)", 
	    __FUNCTION__, (gint)bytes, (gint)g_utf8_strlen(formatted, -1), (gint)gj_history_get_channel(hist)); 

  /* clean up */
  if(formatted != NULL) g_free(formatted);

  return TRUE;
}

#ifdef __cplusplus
}
#endif
