/* 
 * $RCSfile: gj_iq.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/07/28 14:46:35 $
 * 
 * $Revision: 1.8 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <string.h> 

#include <glib.h>

#include "gj_iq.h"
#include "gj_iq_responses.h"
#include "gj_parser.h"
#include "gj_roster.h"
#include "gj_me.h"
#include "gj_support.h"

/* iq list */
gboolean gj_iq_list_add(infoquery iq);
gboolean gj_iq_list_del(infoquery iq);

/* iq */
void gj_iq_destroy(infoquery iq);

/* checks */
gboolean gj_iq_check_params(infoquery iq);

/* structs */
struct t_infoquery
{
  guint eventID;
  guint eventIDRelated;

  infoqueryType type;

  infoqueryCallback cb;

  void *userdata;

  gchar *id;
  gchar *from;
  gchar *to;

  gchar *associated_jid;

  xmlNodePtr xml;
};

GList *infoquery_list;

/*
 * InfoQuery List functions
 */
gboolean gj_iq_list_add(infoquery iq)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    }

  infoquery_list = g_list_append(infoquery_list,iq);
  return TRUE;
}

gboolean gj_iq_list_del(infoquery iq)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    }

  infoquery_list = g_list_remove(infoquery_list,iq);
  return TRUE;
}

/* 
 * InfoQuery functions 
 */
infoquery gj_iq_new(infoqueryType type)
{
  infoquery iq = NULL;
  
  if(type < 1 || type > infoqueryTypeEnd)
    {
      g_warning("%s: infoquery type:%d was out of bounds", __FUNCTION__, type);
      return NULL;
    }


  if((iq = (infoquery) g_new(struct t_infoquery,1)) != NULL)
    {
      iq->eventID = gj_support_get_unique_id();
      iq->eventIDRelated = 0;
      
      iq->type = type;

      iq->xml = NULL;

      iq->cb = NULL;

      iq->userdata = NULL;

      iq->id = NULL;
      iq->to = NULL;
      iq->from = NULL;

      iq->associated_jid = NULL;

      gj_iq_list_add(iq);
    }
  
  return iq;  
}

void gj_iq_free(infoquery iq)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return;
    }

  /* free from list */
  gj_iq_list_del(iq);

  /* free members */
  gj_iq_destroy(iq);

  /* free mem */
  g_free(iq);  
}

void gj_iq_free_by_event_id(guint id)
{
  infoquery iq = NULL;
  
  if((iq = gj_iq_find_by_event_id(id)) == NULL)
    {
      g_warning("%s: could not find infoquery from event id:%d", __FUNCTION__, id);
      return;
    }
  
  gj_iq_free(iq);
}

void gj_iq_free_all_related_by_event_id(guint event_id)
{
  infoquery iq = NULL;
  infoquery iq_related = NULL;
  
  guint related_id = 0;
  
  if((iq = gj_iq_find_by_event_id(event_id)) == NULL)
    {
      g_warning("%s: could not find infoquery from event id:%d", __FUNCTION__, event_id);
      return;
    }
  else
    {
      /* get eventIDRelated from event */
      related_id = gj_iq_get_related_id(iq);
      gj_iq_free(iq);
    }
  
  if((iq_related = gj_iq_find_by_event_id(related_id)) == NULL)
    {
      g_warning("%s: could not find infoquery from related id:%d", __FUNCTION__, event_id);
      return;
    }
  else
    gj_iq_free(iq_related);
}

void gj_iq_free_all_related_by_related_id(guint related_id)
{
  infoquery iq = NULL;
  infoquery iq_related = NULL;
  
  guint event_id = 0;
  
  if((iq_related = gj_iq_find_by_related_id(related_id)) == NULL)
    {
      g_warning("%s: could not find infoquery from event id:%d", __FUNCTION__, related_id);
      return;
    }
  else
    {
      /* get eventIDRelated from event */
      event_id = gj_iq_get_event_id(iq_related);
      gj_iq_free(iq_related);
    }
  
  
  if((iq = gj_iq_find_by_related_id(event_id)) == NULL)
    {
      g_warning("%s: could not find infoquery from related id:%d", __FUNCTION__, related_id);
      return;
    }
  else
    gj_iq_free(iq);
}

void gj_iq_destroy(infoquery iq)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return;
    }

  iq->eventID = 0;
  iq->eventIDRelated = 0;

#if 0  
  iq->eventID = 0; /* get next */
  iq->responseID = 0;
#endif

  if(iq->xml != NULL)
    {
      /* freeing NODE crashes ??? */
      /*xmlFreeNode(iq->xml);  */
      iq->xml = NULL;
    }

  iq->type = 0;
  
  iq->cb = NULL;
 
  iq->userdata = NULL;

  if(iq->id != NULL)
    {
      g_free(iq->id);
      iq->id = NULL;
    }

  if(iq->to != NULL)
    {
      g_free(iq->to);
      iq->to = NULL;
    }

  if(iq->from != NULL)
    {
      g_free(iq->from);
      iq->from = NULL;
    }

  if(iq->associated_jid != NULL)
    {
      g_free(iq->associated_jid);
      iq->associated_jid = NULL;
    }
}


void gj_iq_print_all()
{
  infoquery iq = NULL;
  gint index = 0;

  g_message("++++++++++++++++++++++ InfoQuery - Print All, Begin. +++++++++++++++++++++");

  for(index=0;index<g_list_length(infoquery_list);index++)
    {
      iq = (infoquery) g_list_nth_data(infoquery_list,index);
      gj_iq_print(iq);
    }

  g_message("++++++++++++++++++++++ InfoQuery - Print All, End.   +++++++++++++++++++++");
  
  return;  
}

void gj_iq_print(infoquery iq)
{
  if(iq == NULL)
    return;

  g_message("+++++++++++++++++++++++++++ InfoQuery +++++++++++++++++++++++++");
  g_message("++++");
  g_message("++++ id:'%s'",iq->id);
  g_message("++++ from:'%s'",iq->from);
  g_message("++++ to:'%s'",iq->to);
  g_message("++++ type:%d",iq->type);
  g_message("++++");
  g_message("++++ associated jid:'%s'",iq->associated_jid);
  g_message("++++");
  g_message("++++ eventID:%d",iq->eventID);
  g_message("++++ eventIDRelated:%d",iq->eventIDRelated);
  g_message("++++");
  g_message("++++ response cb:0x%.8x",(gint)iq->cb);
  g_message("++++ userdata:0x%.8x",(gint)iq->userdata);
  g_message("++++");
  g_message("++++ xml:0x%.8x",(gint)iq->xml);
  g_message("+++++++++++++++++++++++++++ InfoQuery +++++++++++++++++++++");
}

/* 
 * external functions 
 */
gboolean gj_iq_init()
{
  /* list is created on first append */
  return TRUE;
}

gboolean gj_iq_term()
{
  gint index = 0;

  if(infoquery_list == NULL)
    return TRUE;

  /* free all items on the list */
  for(index=0;index<g_list_length(infoquery_list);index++)
    {
      infoquery iq = NULL;
      
      if((iq = (infoquery) g_list_nth_data(infoquery_list,index)) != NULL)
	gj_iq_free(iq);
    }

  g_list_free(infoquery_list);
  infoquery_list = NULL;

  return TRUE;
}

/* 
 * member functions 
 */

/* sets */
gboolean gj_iq_set_userdata(infoquery iq, void *userdata)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    }

  if(userdata == NULL)
    {
      g_warning("%s: userdata was NULL", __FUNCTION__);
      return FALSE;
    }

  iq->userdata = userdata;

  return TRUE;
}

gboolean gj_iq_set_event_id(infoquery iq, guint id)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    }

  if(id < 1)
    {
      g_warning("%s: event id:%d was < 1", __FUNCTION__, id);
      return FALSE;
    }

  iq->eventID = id;
  return TRUE;
}

gboolean gj_iq_set_related_id(infoquery iq, guint id)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    }

  if(id < 1)
    {
      g_warning("%s: related id:%d was < 1", __FUNCTION__, id);
      return FALSE;
    }

  iq->eventIDRelated = id;
  return TRUE;
}

gboolean gj_iq_set_id(infoquery iq, gchar *id)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    }

  if(id == NULL)
    {
      g_warning("%s: id was NULL", __FUNCTION__);
      return FALSE;
    }

  if(iq->id != NULL)
    {
      g_free(iq->id);
      iq->id = NULL;
    }
  
  iq->id = g_strdup(id);
  return TRUE;
}

gboolean gj_iq_set_to(infoquery iq, gchar *to)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    }

  if(to == NULL)
    {
      g_warning("%s: to was NULL", __FUNCTION__);
      return FALSE;
    }

  if(iq->to != NULL)
    {
      g_free(iq->to);
      iq->to = NULL;
    }
  
  iq->to = g_strdup(to);
  return TRUE;
}

gboolean gj_iq_set_from(infoquery iq, gchar *from)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    }

  if(from == NULL)
    {
      g_warning("%s: from was NULL", __FUNCTION__);
      return FALSE;
    }

  if(iq->from != NULL)
    {
      g_free(iq->from);
      iq->from = NULL;
    }
  
  iq->from = g_strdup(from);
  return TRUE;
}

gboolean gj_iq_set_type(infoquery iq, infoqueryType type)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    }

  if(type < 1 || type > infoqueryTypeEnd)
    {
      g_warning("%s: type:%d was out of bounds", __FUNCTION__, type);
      return FALSE;
    }

  iq->type = type;
  return TRUE;
}

gboolean gj_iq_set_xml(infoquery iq, xmlNodePtr node)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    }

  if(node == NULL)
    {
      g_warning("%s: xml node was NULL", __FUNCTION__);
      return FALSE;
    } 

  iq->xml = node;
  return TRUE;
}

gboolean gj_iq_set_callback(infoquery iq,  infoqueryCallback cb)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    }

  if(cb == NULL)
    {
      g_warning("%s: callback was NULL", __FUNCTION__);
      return FALSE;
    }

  iq->cb = cb;
  return TRUE;
}

gboolean gj_iq_set_associated_jid(infoquery iq, gchar *jid)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    }

  if(jid == NULL)
    {
      g_warning("%s: jid was NULL", __FUNCTION__);
      return FALSE;
    }

  if(iq->from != NULL)
    {
      g_free(iq->associated_jid);
      iq->associated_jid = NULL;
    }
  
  iq->associated_jid = g_strdup(jid);
  return TRUE;
}

/* gets */
void * gj_iq_get_userdata(infoquery iq)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    }

  return iq->userdata;
}

gchar *gj_iq_get_id(infoquery iq)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return NULL;
    }

  return iq->id;
}

gchar *gj_iq_get_to(infoquery iq)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return NULL;
    }

  return iq->to;
}

gchar *gj_iq_get_from(infoquery iq)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return NULL;
    }

  return iq->from;
}

infoqueryType gj_iq_get_type(infoquery iq)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return 0;
    }

  return iq->type;
}

xmlNodePtr gj_iq_get_xml(infoquery iq)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    }

  return iq->xml;
}

void *gj_iq_get_callback(infoquery iq)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    }

  return (void*)iq->cb;
}

guint gj_iq_get_event_id(infoquery iq)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return -1;
    }

  return iq->eventID;
}

guint gj_iq_get_related_id(infoquery iq)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return -1;
    }

  return iq->eventIDRelated;
}

gchar *gj_iq_get_associated_jid(infoquery iq)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return NULL;
    }

  return iq->associated_jid;
}


/* 
 * compile iq 
 */
gboolean gj_iq_check_params(infoquery iq)
{
  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    } 

  /* if iq is type GET, requires ID, To (optional) 
     if iq is type SET, requires ID, To (optional)
     if iq is type RESULT, requires ID, To (optional), From
     if iq is type ERROR, requires ID, To (optional), From
  */

  if(gj_iq_get_type(iq) != infoqueryTypeGet && 
     gj_iq_get_type(iq) != infoqueryTypeSet && 
     gj_iq_get_type(iq) != infoqueryTypeResult && 
     gj_iq_get_type(iq) != infoqueryTypeError)
    {
      g_warning("%s: infoquery type:%d unrecognised", __FUNCTION__, gj_iq_get_type(iq));
      return FALSE;      
    }

  return TRUE;
}

gboolean gj_iq_build(infoquery iq, xmlNodePtr node)
{
  xmlNodePtr parent = NULL;

  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    } 
  
  if(node == NULL)
    {
      g_warning("%s: xml node was NULL", __FUNCTION__);
      return FALSE;
    }
  
  if(gj_iq_check_params(iq) == FALSE)
    return FALSE;
  
  /* set root node */
  parent = xmlNewNode(NULL, (guchar*)"iq");
  
  /* id */
  if(gj_iq_get_id(iq) == NULL)
    {
      gchar *id = NULL;
      
      if((id = g_strdup_printf("%d",gj_iq_get_event_id(iq))) != NULL)
	{
	  gj_iq_set_id(iq,id);
	  g_free(id);
	}
    }

  xmlSetProp(parent, (guchar*)"id", (guchar*)gj_iq_get_id(iq));
  
  /* type */
  if(iq->type == infoqueryTypeGet)
    xmlSetProp(parent, (guchar*)"type", (guchar*)"get");
  if(iq->type == infoqueryTypeSet)
    xmlSetProp(parent, (guchar*)"type", (guchar*)"set");
  if(iq->type == infoqueryTypeResult)
    xmlSetProp(parent, (guchar*)"type", (guchar*)"result");
  if(iq->type == infoqueryTypeError)
    xmlSetProp(parent, (guchar*)"type", (guchar*)"error");      

  /* 
   * set props 
   */
  
  /* to */
  if(gj_iq_get_to(iq) != NULL)
    xmlSetProp(parent, (guchar*)"to", (guchar*)gj_iq_get_to(iq));
    
  /* from */
  if(gj_iq_get_from(iq) != NULL)
    xmlSetProp(parent, (guchar*)"from", (guchar*)gj_iq_get_from(iq));

  /* add node */
  xmlAddChild(parent, node);

  /* set iq query */
  gj_iq_set_xml(iq, parent);

  return TRUE;
}

gboolean gj_iq_build_from_list(infoquery iq, GList *nodes)
{
   xmlNodePtr parent = NULL;
   
   guint count = 0;

   if(iq == NULL)
     {
       g_warning("%s: xml node was NULL", __FUNCTION__);
       return FALSE;
     } 
   
   if(nodes == NULL)
     {
       g_warning("%s: list of xml nodes was NULL", __FUNCTION__);
       return FALSE;
     }
   
   if(gj_iq_check_params(iq) == FALSE)
     return FALSE;
   
   /* set root node */
   parent = xmlNewNode(NULL, (guchar*)"iq");
   
   /* id */
   if(gj_iq_get_id(iq) == NULL)
     {
       gchar *id = NULL;
       
       if((id = g_strdup_printf("%d",gj_iq_get_event_id(iq))) != NULL)
	 {
	   gj_iq_set_id(iq, id);
	   g_free(id);
	 }
     }
   
   xmlSetProp(parent, (guchar*)"id", (guchar*)gj_iq_get_id(iq));
   
   /* type */
   if(iq->type == infoqueryTypeGet)
     xmlSetProp(parent, (guchar*)"type", (guchar*)"get");
   if(iq->type == infoqueryTypeSet)
     xmlSetProp(parent, (guchar*)"type", (guchar*)"set");
   if(iq->type == infoqueryTypeResult)
     xmlSetProp(parent, (guchar*)"type", (guchar*)"result");
   if(iq->type == infoqueryTypeError)
     xmlSetProp(parent, (guchar*)"type", (guchar*)"error");      
   
   /* 
    * set props 
    */
   
   while(count < g_list_length(nodes)) 
     {
       xmlNodePtr node = g_list_nth_data(nodes,count);
       
       if(node == NULL)
	 continue;
       else
	 count++;

       /* add node */
       xmlAddChild(parent,node);
     }
   
   if(count == 0)
     {
       g_warning("%s: there were 0 top level nodes", __FUNCTION__);
       return FALSE;
     }
   
   /* to */
   if(gj_iq_get_to(iq) != NULL)
     xmlSetProp(parent, (guchar*)"to", (guchar*)gj_iq_get_to(iq));
   
   /* from */
   if(gj_iq_get_from(iq) != NULL)
     xmlSetProp(parent, (guchar*)"from", (guchar*)gj_iq_get_from(iq));
   
   /* set iq query */
   gj_iq_set_xml(iq, parent);
   
   return TRUE;
}

gboolean gj_iq_send(infoquery iq)
{
  xmlNodePtr node = NULL;

  if(iq == NULL)
    {
      g_warning("%s: infoquery was NULL", __FUNCTION__);
      return FALSE;
    } 

  if((node = gj_iq_get_xml(iq)) == NULL)
    {
      g_warning("%s: xml node was NULL", __FUNCTION__);
      return FALSE;
    }

  return gj_parser_push_output(node);
}

infoquery gj_iq_find_by_event_id(guint id)
{
  gint index = 0;

  if(id < 0)
    {
      g_warning("%s: id:%d was < 1", __FUNCTION__, id);
      return NULL;
    }

  for(index=0;index<g_list_length(infoquery_list);index++)
    {
      infoquery iq = g_list_nth_data(infoquery_list, index);

      if(iq == NULL) continue;
      if(iq->eventID == id) return iq;
    }

  return NULL;
}

infoquery gj_iq_find_by_related_id(guint id)
{
  gint index = 0;

  if(id < 0)
    {
      g_warning("%s: id:%d was < 1", __FUNCTION__, id);
      return NULL;
    }

  for(index=0;index<g_list_length(infoquery_list);index++)
    {
      infoquery iq = g_list_nth_data(infoquery_list, index);

      if(iq == NULL) continue;
      if(iq->eventIDRelated == id) return iq;
    }

  return NULL;
}

infoquery gj_iq_find_by_id(gchar *id)
{
  gint index = 0;

  if(id == NULL)
    {
      g_warning("%s: id was NULL", __FUNCTION__);
      return NULL;
    }

  for(index=0;index<g_list_length(infoquery_list);index++)
    {
      infoquery iq = g_list_nth_data(infoquery_list, index);

      if(iq == NULL) continue;
      if(iq->id != NULL && strncmp(id, iq->id, g_utf8_strlen(id, -1)) == 0) return iq;
    }

  return NULL;
}


/*
 * ENTRY POINT called by parser
 */
void gj_iq_receive(connection c, xmlNodePtr node)
{
  xmlChar *tmp = NULL;

  infoquery iq_new = NULL;
  infoquery iq_related = NULL;
  infoqueryType type = infoqueryTypeError;

  if(node == NULL)
    {
      g_warning("%s: xml node was NULL", __FUNCTION__);
      return;
    }

  if((tmp = gj_parser_find_by_attr_and_ret(node, (guchar*)"id")) != NULL && xmlStrlen(tmp) > 0)
    {
      /* see if there already is an infoquery with this id, 
	 if so, this is possibly the response */ 
      iq_related = gj_iq_find_by_id((gchar*)tmp); 
    }

  if((tmp = gj_parser_find_by_attr_and_ret(node, (guchar*)"type")) != NULL && xmlStrlen(tmp) > 0)
      {
	if(xmlStrncmp(tmp, (guchar*)"get", xmlStrlen(tmp)) == 0)
	  type = infoqueryTypeGet;
	else if(xmlStrncmp(tmp, (guchar*)"set", xmlStrlen(tmp)) == 0)
	  type = infoqueryTypeSet;
	else if(xmlStrncmp(tmp, (guchar*)"result", xmlStrlen(tmp)) == 0)
	  type = infoqueryTypeResult;
	else if(xmlStrncmp(tmp, (guchar*)"error", xmlStrlen(tmp)) == 0)
	  type = infoqueryTypeError;
	else
	  {
	    g_warning("%s: could not determin xml type from:'%s', thowing away....", __FUNCTION__, tmp);
	    return;
	  }
      }
  
  if((iq_new = gj_iq_new(type)) == NULL)
    {
      g_warning("%s: could not create new infoquery", __FUNCTION__);
      return;
    }
  
  /* set xml for iq */
  gj_iq_set_xml(iq_new, node);
  
  /* set characteristics */
  if((tmp = gj_parser_find_by_attr_and_ret(node, (guchar*)"id")) != NULL && xmlStrlen(tmp) > 0)
    gj_iq_set_id(iq_new, (gchar*)tmp);
  
  if((tmp = gj_parser_find_by_attr_and_ret(node, (guchar*)"to")) != NULL && xmlStrlen(tmp) > 0)
    gj_iq_set_to(iq_new, (gchar*)tmp);
  
  if((tmp = gj_parser_find_by_attr_and_ret(node, (guchar*)"from")) != NULL && xmlStrlen(tmp) > 0)
    gj_iq_set_from(iq_new, (gchar*)tmp);
  
  /* 
     see if this iq relates to a previous event, 
     and if so call the callback 
  */
  if(iq_related != NULL)
    {
      infoqueryCallback cb = NULL;
      
      if((cb = (infoqueryCallback) gj_iq_get_callback(iq_related)) == NULL)
	{
	  g_message("%s: infoquery has no callback to original query (id:%d), freeing...", 
		    __FUNCTION__, (gint)gj_iq_get_id(iq_related));
	  
	  /* free mem */
	  gj_iq_free(iq_new);  
	  gj_iq_free(iq_related); 
	  
	  return;
	}
      
      /* set responseID for new response infoquery */
      gj_iq_set_related_id(iq_related, gj_iq_get_event_id(iq_new));
      gj_iq_set_related_id(iq_new, gj_iq_get_event_id(iq_related));
      
      /* call callback function */
      (cb)(gj_iq_get_event_id(iq_related));
      
      
#if 0
      /* free iq */
      gj_iq_free(iq_related);
#endif
    }
  else
    {
      /* not related to a previous event, must be new from the server */
      gj_iq_request_new(iq_new);
    }
}

#ifdef __cplusplus
}
#endif
