/* 
 * $RCSfile: gj_message.h,v $
 *
 * $Author: ginxd $
 * $Date: 2003/08/23 11:17:26 $
 * 
 * $Revision: 1.9 $
 *
 */

#ifndef __gj_message_h
#define __gj_message_h

#include <time.h> 

#include "gj_connection.h"

/*
 * Message
 */

typedef enum messageType {
    messageTypeNormal = 1,
    messageTypeChat,
    messageTypeGroupChat,
    messageTypeHeadline,
    messageTypeError,
    messageTypeXEvent,
    
    messageTypeEnd
} messageType;
 

#define messageXEventTypeOffline      1 << 1
#define messageXEventTypeDelivered    1 << 2
#define messageXEventTypeDisplayed    1 << 3 
#define messageXEventTypeComposing    1 << 4
#define messageXEventTypeCancelled    1 << 5


typedef struct t_message *message;

typedef void (*messageFromCallback)(message msg);
typedef void (*messageToCallback)(gint requestID);


gboolean gj_message_init();
gboolean gj_message_term();

/* item functions */
void gj_message_free(message msg);

/* find functions */
message gj_message_find_by_to_jid(gchar *jid);

/*
 * member functions
 */

/* sets */
gboolean gj_message_set_type(message msg, messageType type);
gboolean gj_message_set_id(message msg, gchar *id);

gboolean gj_message_set_to_jid(message msg, gchar *jid);
gboolean gj_message_set_to_resource(message msg, gchar *resource);

gboolean gj_message_set_from_jid(message msg, gchar *jid);
gboolean gj_message_set_from_resource(message msg, gchar *resource);

gboolean gj_message_set_subject(message msg, gchar *subject);
gboolean gj_message_set_body(message msg, gchar *body);

gboolean gj_message_set_thread(message msg, gchar *thread);

gboolean gj_message_set_x_event_type(message msg, gint type);
gboolean gj_message_set_x_event_message_id(message msg, gchar *message_id);

gboolean gj_message_set_x_delay_stamp(message msg, gchar *stamp);

gboolean gj_message_set_xoob_url(message msg, gchar *url);
gboolean gj_message_set_xoob_description(message msg, gchar *description);

gboolean gj_message_set_error_code(message msg, gchar *error_code);
gboolean gj_message_set_error_reason(message msg, gchar *error_reason);

/* gets */
messageType gj_message_get_type(message msg);
gchar *gj_message_get_id(message msg);

gint gj_message_get_request_id(message msg);

gchar *gj_message_get_to_jid(message msg);
gchar *gj_message_get_to_resource(message msg);

gchar *gj_message_get_from_jid(message msg);
gchar *gj_message_get_from_resource(message msg);

gchar *gj_message_get_subject(message msg);
gchar *gj_message_get_body(message msg);

gchar *gj_message_get_thread(message msg);

gint gj_message_get_x_event_type(message msg);
gchar *gj_message_get_x_event_message_id(message msg);

gchar *gj_message_get_x_delay_stamp(message msg);

gchar *gj_message_get_x_oob_url(message msg);
gchar *gj_message_get_x_oob_description(message msg);

gchar *gj_message_get_error_code(message msg);
gchar *gj_message_get_error_reason(message msg);

time_t gj_message_get_time(message msg);

/*
 * message functions to tcp / parser layer 
 */
gboolean gj_message_send(messageType type, gchar *to, gchar *subject, gchar *body, gchar *thread);
gboolean gj_message_send_x_event(gint type, gchar *to, gchar *id);
gchar *gj_message_send_with_x_event_request(messageType type, gchar *to, gchar *subject, gchar *body, gchar *thread, gint xevents);

void gj_message_receive(connection c, xmlNodePtr node);

/*
 * callbacks
 */
gboolean gj_message_set_cb_to(messageToCallback cb);
gboolean gj_message_set_cb_from(messageFromCallback cb);


#endif
