/* 
 * $RCSfile: gj_presence.h,v $
 *
 * $Author: ginxd $
 * $Date: 2003/08/01 20:40:40 $
 * 
 * $Revision: 1.7 $
 *
 */

#ifndef __gj_presence_h
#define __gj_presence_h

#include <libxml/tree.h>

#include "gj_connection.h"

/*
 * Presence
 */
typedef enum presenceType {
  presenceTypeAvailable = 1,   /* online */  
  presenceTypeUnAvailable,     /* offline */
  presenceTypeProbe,           /* ask another user's status */
  presenceTypeSubscribe,       /* request to subscribe to presence for user */
  presenceTypeUnSubscribe,     /* request to unsubscribe to presence for user */
  presenceTypeSubscribed,      /* result to subscribe */ 
  presenceTypeUnSubscribed,    /* result to unsubscribe */
  presenceTypeError,           /* error */
  
  presenceTypeEnd              /* end of types */
} presenceType;

typedef enum presenceShow {
  presenceShowAway = 1,    /* temporarily away */
  presenceShowChat,        /* similar to normal except in a conversation */
  presenceShowDND,         /* do not disturb */ 
  presenceShowNormal,      /* default */
  presenceShowXA,          /* eXtended Away, almost unavailable */
  
  presenceShowEnd          /* end of types */
} presenceShow;

typedef struct t_presence *presence;

typedef void (*presenceFromCallback)(presence pres);
typedef void (*presenceToCallback)(gint requestID);

typedef struct t_presenceTo *presenceTo; 

gboolean gj_presence_init();
gboolean gj_presence_term();

/* item functions */
presence gj_presence_new(gint requestID, presenceType type, presenceShow show);
void gj_presence_free(presence pres);
presence gj_presence_copy(presence pres);

/* find functions */
presence gj_presence_find_by_to_jid(gchar *jid);

/*
 * member functions
 */

/* sets */
gboolean gj_presence_set_type(presence pres, presenceType type);
gboolean gj_presence_set_show(presence pres, presenceShow show);

gboolean gj_presence_set_status(presence pres, gchar *status);
gboolean gj_presence_set_priority(presence pres, gint8 priority);

gboolean gj_presence_set_to_jid(presence pres, gchar *jid);
gboolean gj_presence_set_to_resource(presence pres, gchar *resource);

gboolean gj_presence_set_from_jid(presence pres, gchar *jid);
gboolean gj_presence_set_from_resource(presence pres, gchar *resource);

gboolean gj_presence_set_error_code(presence pres, gchar *error_code);
gboolean gj_presence_set_error_reason(presence pres, gchar *error_reason);

/* gets */
gint gj_presence_get_request_id(presence pres);

presenceType gj_presence_get_type(presence pres);
presenceShow gj_presence_get_show(presence pres);

gchar *gj_presence_get_status(presence pres);
gint8 gj_presence_get_priority(presence pres);

gchar *gj_presence_get_to_jid(presence pres);
gchar *gj_presence_get_to_resource(presence pres);

gchar *gj_presence_get_from_jid(presence pres);
gchar *gj_presence_get_from_resource(presence pres);

gchar *gj_presence_get_error_code(presence pres);
gchar *gj_presence_get_error_reason(presence pres);

/*
 * entry into module... sending/receiving presence.
 */
gboolean gj_presence_set_to(gchar *jid, presenceType type,presenceShow show, gchar *status, gint priority);

gboolean gj_presence_send(presenceType type,presenceShow show, gchar *status, gint8 priority);
gboolean gj_presence_send_to(gchar *jid, presenceType type, presenceShow show, gchar *status, gint8 priority);

void gj_presence_receive(connection c, xmlNodePtr node);

/* callbacks */
gboolean gj_presence_set_cb_to(presenceToCallback cb);
gboolean gj_presence_set_cb_from(presenceFromCallback cb);

#endif

