/*
 * $RCSfile: gj_support.c,v $
 *
 * $Author: ginxd $
 * $Date: 2003/07/28 14:46:35 $
 *
 * $Revision: 1.8 $
 *
 */

#ifdef __cplusplus
extern "C" {
#if 0
}
#endif
#endif

#include <string.h>
#include <ctype.h>
#include <time.h> 

#include <glib.h>

#include <gnet.h>

#include "gj_support.h"

int gj_support_get_unique_id()
{
  static int id = 69;
  return id++;
}

const char *gj_support_get_time_zone()
{
#ifndef WIN32
  static char buf[256];
  time_t tnow = time(0);
  struct tm *tmLocal = localtime(&tnow);
 
  if(tmLocal == NULL)
    return "time zone unknown";
  else
    return tmLocal->tm_zone; 
#else
  return "time zone unknown on win32";
#endif
}

const char *gj_support_get_timestamp()
{
  static char buf[256];
  time_t tnow = time(0);
  struct tm *tmLocal = localtime(&tnow);

  if(strftime(buf, 256, "%H:%M:%S", tmLocal) == 0)
    return "time unknown";
  else
    return buf;
}

const char *gj_support_get_timestamp_long()
{
  static char buf[256];
  time_t tnow = time(0);
  struct tm *tmLocal = localtime(&tnow);

  if(strftime(buf, 256, "%H%M%S", tmLocal) == 0)
    return "time unknown";
  else
    return buf;
}

const char *gj_support_get_timestamp_from_timet(time_t now)
{
  static char buf[256];
  struct tm *tmLocal = localtime(&now);

  if(strftime(buf, 256, "%H:%M:%S", tmLocal) == 0)
    return "time unknown";
  else
    return buf;
}

const char *gj_support_get_timestamp_full()
{
  static char buf[256];
  time_t tnow = time(0);
  struct tm *tmLocal = localtime(&tnow);

  if(strftime(buf, 256, "%Y%m%d%H%M%S", tmLocal) == 0)
    return "time/date unknown";
  else
    return buf;
}

const char *gj_support_get_timestamp_nice()
{
  static char buf[256];
  time_t tnow = time(0);
  struct tm *tmLocal = localtime(&tnow);

  if(strftime(buf, 256, "%a, %b %d, %H:%M:%S", tmLocal) == 0)
    return "time/date unknown";
  else
    return buf;
}

const char *gj_support_get_timestamp_nice_from_timet(time_t now)
{
  static char buf[256];
  struct tm *tmLocal = localtime(&now);

  if(strftime(buf, 256, "%a, %b %d, %H:%M:%S", tmLocal) == 0)
    return "time/date unknown";
  else
    return buf;
}

const char *gj_support_get_datestamp()
{
  static char buf[256];
  time_t tnow = time(0);
  struct tm *tmLocal = localtime(&tnow);

  if(strftime(buf, 256, "%a, %b %d", tmLocal) == 0)
    return "date unknown";
  else
    return buf;
}

const char *gj_support_get_datestamp_long()
{
  static char buf[256];
  time_t tnow = time(0);
  struct tm *tmLocal = localtime(&tnow);

  if(strftime(buf, 256, "%Y%m%d", tmLocal) == 0)
    return "date unknown";
  else
    return buf;
}

const char *gj_support_get_datestamp_from_timet(time_t now)
{
  static char buf[256];
  struct tm *tmLocal = localtime(&now);

  if(strftime(buf, 256, "%a, %b %d", tmLocal) == 0)
    return "date unknown";
  else
    return buf;
} 

const char *gj_support_get_time_and_date_from_timet(time_t now)
{
  static char buf[256];
  struct tm *tmLocal = localtime(&now);

  if(strftime(buf, 256, "%A, %B %d, %H:%M:%S", tmLocal) == 0)
    return "date/time unknown";
  else
    return buf;
}

gboolean gj_support_find_in_str(const gchar *needle, const gchar *haystack)
{
    gchar *result = NULL;

    if(needle == NULL)
    {
	g_warning("%s: needle is NULL", __FUNCTION__);
	return FALSE;
    }

    if(haystack == NULL)
    {
	g_warning("%s: haystack is NULL", __FUNCTION__);
	return FALSE;
    }

    if((result = g_strrstr(haystack, needle)) == NULL)
	return FALSE;
    
    return TRUE;
}

gchar *gj_support_find_in_str_and_remove(const gchar *needle, const gchar *haystack)
{
    gchar **vstripped = NULL;
    gchar *result = NULL;

    gint i = 0;
    gboolean end_of_array = FALSE;

    if(needle == NULL)
    {
	g_warning("%s: needle is NULL", __FUNCTION__);
	return NULL;
    }

    if(haystack == NULL)
    {
	g_warning("%s: haystack is NULL", __FUNCTION__);
	return NULL;
    }

    /* remove all except numbers */
    vstripped = g_strsplit(haystack, needle, -1);

    if(vstripped == NULL)
    {
	g_warning("%s: could not split text with delimiter:'%s' - NULL returned..", __FUNCTION__, needle);
	return NULL;
    }
   
    while(end_of_array == FALSE)
    {
	if(vstripped[i] == NULL)
	{
	    end_of_array = TRUE;
	    break;
	}

	if(result == NULL)
	    result = g_strdup(vstripped[i]);
	else
	    result = g_strconcat(result, vstripped[i], NULL);

	i++;
    }

    /* clean up */
    if(vstripped != NULL) g_strfreev(vstripped);

    return result;
}

gchar *gj_support_find_in_str_and_replace(const gchar *needle, const gchar *haystack, const gchar *replacement)
{
    gchar **vstripped = NULL;
    gchar *result = NULL;

    gint i = 0;
    gboolean end_of_array = FALSE;

    if(needle == NULL)
    {
	g_warning("%s: needle is NULL", __FUNCTION__);
	return NULL;
    }

    if(haystack == NULL)
    {
	g_warning("%s: haystack is NULL", __FUNCTION__);
	return NULL;
    }

    if(replacement == NULL)
    {
	g_warning("%s: replacement is NULL", __FUNCTION__);
	return NULL;
    }

    /* remove all except numbers */
    vstripped = g_strsplit(haystack, needle, -1);

    if(vstripped == NULL)
    {
	g_warning("%s: could not split text with delimiter:'%s' - NULL returned..", __FUNCTION__, needle);
	return NULL;
    }
   
    while(end_of_array == FALSE)
    {
	if(vstripped[i] == NULL)
	{
	    end_of_array = TRUE;
	    break;
	}

	if(result == NULL)
	    result = g_strdup(vstripped[i]);
	else
	    result = g_strconcat(result, replacement, vstripped[i], NULL);

	i++;
    }

    /* clean up */
    if(vstripped != NULL) g_strfreev(vstripped);

    return result;
}

/*
 * utils 
 */
gboolean gj_support_is_word_separator(gchar c)
{
/* 	return !isalpha(c) && c != '\''; */
  return(isspace(c));
}

uriType gj_support_is_uri(gchar *word)
{
  gint len = 0;
  if(!word) return uriTypeUnknown;
  len = g_utf8_strlen(word, -1);
  
  if(!strncasecmp(word, "irc://", 6))
    return uriTypeURL;
  
  if(!strncasecmp(word, "irc.", 4))
    return uriTypeURL;
  
  if(!strncasecmp(word, "ftp.", 4))
    return uriTypeURL;
  
  if(!strncasecmp(word, "ftp:", 4))
    return uriTypeURL;
  
  if(!strncasecmp(word, "www.", 4))
    return uriTypeURL;
  
  if(!strncasecmp(word, "http:", 5))
    return uriTypeURL;
  
  if(!strncasecmp(word, "https:", 6))
    return uriTypeURL;
  
  if(!strncasecmp(word + len - 4, ".org", 4))
    return uriTypeHost;
  
  if(!strncasecmp(word + len - 4, ".net", 4))
    return uriTypeHost;
  
  if(!strncasecmp(word + len - 4, ".com", 4))
    return uriTypeHost;
  
  if(!strncasecmp(word + len - 4, ".edu", 4))
    return uriTypeHost;
  
  return uriTypeUnknown;
}

gchar *gj_support_get_sha_by_session_id_and_password(gchar *session_id, gchar *password)
{
  GSHA *sha = NULL;
  gchar *digest_input = NULL;
  gchar *digest_output = NULL;
  
  if(session_id == NULL)
    {
      g_warning("%s: session ID was NULL", __FUNCTION__);
      return NULL;
    }
  
  if(password == NULL)
    {
      g_warning("%s: password was NULL", __FUNCTION__);
      return NULL;
    }
  
  digest_input = g_strconcat(session_id, password, NULL);
  g_message("%s: SHA input is:'%s'", __FUNCTION__, digest_input);

  sha = gnet_sha_new((gchar*)digest_input, strlen(digest_input));
  digest_output = gnet_sha_get_string(sha);
  g_message("%s: SHA output is:'%s'", __FUNCTION__, digest_output);
  
  /* clean up */
  if(digest_input != NULL) g_free(digest_input);
  gnet_sha_delete(sha);
  
  return digest_output;
}

#ifdef __cplusplus
}
#endif
